#pragma once

#include <LibGfx/BitmapSequence.h>
#include <LibGfx/ColorSpace.h>
#include <AK/Error.h>
#include <AK/MemoryStream.h>
#include <AK/OwnPtr.h>
#include <AK/Result.h>
#include <AK/Utf8View.h>
#include <LibIPC/Connection.h>
#include <LibIPC/Decoder.h>
#include <LibIPC/Encoder.h>
#include <LibIPC/File.h>
#include <LibIPC/Message.h>
#include <LibIPC/Stub.h>

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdefaulted-function-deleted"
#endif

namespace Messages::ImageDecoderClient {

enum class MessageID : i32 {
    DidDecodeImage = 1,
    DidFailToDecodeImage = 2,
};

class DidDecodeImage final : public IPC::Message {
public:
    DidDecodeImage(DidDecodeImage const&) = default;
    DidDecodeImage(DidDecodeImage&&) = default;
    DidDecodeImage& operator=(DidDecodeImage const&) = default;
    DidDecodeImage(i64 image_id, bool is_animated, u32 loop_count, Gfx::BitmapSequence bitmaps, Vector<u32> durations, Gfx::FloatPoint scale, Gfx::ColorSpace color_profile) : m_image_id(move(image_id)), m_is_animated(move(is_animated)), m_loop_count(move(loop_count)), m_bitmaps(move(bitmaps)), m_durations(move(durations)), m_scale(move(scale)), m_color_profile(move(color_profile)) {}

    virtual ~DidDecodeImage() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 945471142;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidDecodeImage; }
    static i32 static_message_id() { return (int)MessageID::DidDecodeImage; }
    virtual const char* message_name() const override { return "ImageDecoderClient::DidDecodeImage"; }

    static ErrorOr<NonnullOwnPtr<DidDecodeImage>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto image_id = TRY((decoder.decode<i64>()));
        auto is_animated = TRY((decoder.decode<bool>()));
        auto loop_count = TRY((decoder.decode<u32>()));
        auto bitmaps = TRY((decoder.decode<Gfx::BitmapSequence>()));
        auto durations = TRY((decoder.decode<Vector<u32>>()));
        auto scale = TRY((decoder.decode<Gfx::FloatPoint>()));
        auto color_profile = TRY((decoder.decode<Gfx::ColorSpace>()));
        return make<DidDecodeImage>(move(image_id), move(is_animated), move(loop_count), move(bitmaps), move(durations), move(scale), move(color_profile));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i64 image_id, bool is_animated, u32 loop_count, Gfx::BitmapSequence const& bitmaps, ReadonlySpan<u32> durations, Gfx::FloatPoint scale, Gfx::ColorSpace const& color_profile)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidDecodeImage));
        TRY(stream.encode(image_id));
        TRY(stream.encode(is_animated));
        TRY(stream.encode(loop_count));
        TRY(stream.encode(bitmaps));
        TRY(stream.encode(durations));
        TRY(stream.encode(scale));
        TRY(stream.encode(color_profile));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_image_id, m_is_animated, m_loop_count, m_bitmaps, m_durations, m_scale, m_color_profile);
    }

    i64 image_id() const { return m_image_id; }

    bool is_animated() const { return m_is_animated; }

    u32 loop_count() const { return m_loop_count; }

    const Gfx::BitmapSequence& bitmaps() const { return m_bitmaps; }
    Gfx::BitmapSequence take_bitmaps() { return move(m_bitmaps); }

    const Vector<u32>& durations() const { return m_durations; }
    Vector<u32> take_durations() { return move(m_durations); }

    Gfx::FloatPoint scale() const { return m_scale; }

    const Gfx::ColorSpace& color_profile() const { return m_color_profile; }
    Gfx::ColorSpace take_color_profile() { return move(m_color_profile); }

private:
    i64 m_image_id;
    bool m_is_animated;
    u32 m_loop_count;
    Gfx::BitmapSequence m_bitmaps;
    Vector<u32> m_durations;
    Gfx::FloatPoint m_scale;
    Gfx::ColorSpace m_color_profile;
};

class DidFailToDecodeImage final : public IPC::Message {
public:
    DidFailToDecodeImage(DidFailToDecodeImage const&) = default;
    DidFailToDecodeImage(DidFailToDecodeImage&&) = default;
    DidFailToDecodeImage& operator=(DidFailToDecodeImage const&) = default;
    DidFailToDecodeImage(i64 image_id, String error_message) : m_image_id(move(image_id)), m_error_message(move(error_message)) {}

    virtual ~DidFailToDecodeImage() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 945471142;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidFailToDecodeImage; }
    static i32 static_message_id() { return (int)MessageID::DidFailToDecodeImage; }
    virtual const char* message_name() const override { return "ImageDecoderClient::DidFailToDecodeImage"; }

    static ErrorOr<NonnullOwnPtr<DidFailToDecodeImage>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto image_id = TRY((decoder.decode<i64>()));
        auto error_message = TRY((decoder.decode<String>()));
        return make<DidFailToDecodeImage>(move(image_id), move(error_message));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i64 image_id, StringView error_message)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidFailToDecodeImage));
        TRY(stream.encode(image_id));
        TRY(stream.encode(error_message));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_image_id, m_error_message);
    }

    i64 image_id() const { return m_image_id; }

    const String& error_message() const { return m_error_message; }
    String take_error_message() { return move(m_error_message); }

private:
    i64 m_image_id;
    String m_error_message;
};

} // namespace Messages::ImageDecoderClient

template<typename LocalEndpoint, typename PeerEndpoint>
class ImageDecoderClientProxy {
public:
    // Used to disambiguate the constructor call.
    struct Tag { };

    ImageDecoderClientProxy(IPC::Connection<LocalEndpoint, PeerEndpoint>& connection, Tag)
        : m_connection(connection)
    { }

    void async_did_decode_image(i64 image_id, bool is_animated, u32 loop_count, Gfx::BitmapSequence const& bitmaps, ReadonlySpan<u32> durations, Gfx::FloatPoint scale, Gfx::ColorSpace const& color_profile) {
        auto message_buffer = MUST(Messages::ImageDecoderClient::DidDecodeImage::static_encode(image_id, is_animated, loop_count, move(bitmaps), durations, scale, move(color_profile)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_fail_to_decode_image(i64 image_id, StringView error_message) {
        VERIFY(Utf8View { error_message }.validate());
        auto message_buffer = MUST(Messages::ImageDecoderClient::DidFailToDecodeImage::static_encode(image_id, error_message));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_fail_to_decode_image(i64 image_id, String const& error_message) {
        auto message_buffer = MUST(Messages::ImageDecoderClient::DidFailToDecodeImage::static_encode(image_id, error_message));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

private:
    IPC::Connection<LocalEndpoint, PeerEndpoint>& m_connection;
};

template<typename LocalEndpoint, typename PeerEndpoint>
class ImageDecoderClientProxy;
class ImageDecoderClientStub;

class ImageDecoderClientEndpoint {
public:
    template<typename LocalEndpoint>
    using Proxy = ImageDecoderClientProxy<LocalEndpoint, ImageDecoderClientEndpoint>;
    using Stub = ImageDecoderClientStub;

    static u32 static_magic() { return 945471142; }

    static ErrorOr<NonnullOwnPtr<IPC::Message>> decode_message(ReadonlyBytes buffer, [[maybe_unused]] Queue<IPC::File>& files)
    {
        FixedMemoryStream stream { buffer };
        auto message_endpoint_magic = TRY(stream.read_value<u32>());

        if (message_endpoint_magic != 945471142) {
            return Error::from_string_literal("Endpoint magic number mismatch, not my message!");
        }

        auto message_id = TRY(stream.read_value<i32>());


        switch (message_id) {

        case (int)Messages::ImageDecoderClient::MessageID::DidDecodeImage:
            return TRY(Messages::ImageDecoderClient::DidDecodeImage::decode(stream, files));
        case (int)Messages::ImageDecoderClient::MessageID::DidFailToDecodeImage:
            return TRY(Messages::ImageDecoderClient::DidFailToDecodeImage::decode(stream, files));
        default:
            return Error::from_string_literal("Failed to decode ImageDecoderClient message");
        }

        VERIFY_NOT_REACHED();
    }

};

class ImageDecoderClientStub : public IPC::Stub {
public:
    ImageDecoderClientStub() { }
    virtual ~ImageDecoderClientStub() override { }

    virtual u32 magic() const override { return 945471142; }
    virtual ByteString name() const override { return "ImageDecoderClient"; }

    virtual ErrorOr<OwnPtr<IPC::MessageBuffer>> handle(NonnullOwnPtr<IPC::Message> message) override
    {
        switch (message->message_id()) {

        case (int)Messages::ImageDecoderClient::MessageID::DidDecodeImage: {
            [[maybe_unused]] auto& request = static_cast<Messages::ImageDecoderClient::DidDecodeImage&>(*message);
            did_decode_image(request.image_id(), request.is_animated(), request.loop_count(), request.take_bitmaps(), request.take_durations(), request.scale(), request.take_color_profile());
            return nullptr;
        }
        case (int)Messages::ImageDecoderClient::MessageID::DidFailToDecodeImage: {
            [[maybe_unused]] auto& request = static_cast<Messages::ImageDecoderClient::DidFailToDecodeImage&>(*message);
            did_fail_to_decode_image(request.image_id(), request.take_error_message());
            return nullptr;
        }
        default:
            return Error::from_string_literal("Unknown message ID for ImageDecoderClient endpoint");
        }
    }

    virtual void did_decode_image(i64 image_id, bool is_animated, u32 loop_count, Gfx::BitmapSequence bitmaps, Vector<u32> durations, Gfx::FloatPoint scale, Gfx::ColorSpace color_profile) = 0;
    virtual void did_fail_to_decode_image(i64 image_id, String error_message) = 0;
};

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic pop
#endif
