#pragma once

#include <LibCore/AnonymousBuffer.h>
#include <AK/Error.h>
#include <AK/MemoryStream.h>
#include <AK/OwnPtr.h>
#include <AK/Result.h>
#include <AK/Utf8View.h>
#include <LibIPC/Connection.h>
#include <LibIPC/Decoder.h>
#include <LibIPC/Encoder.h>
#include <LibIPC/File.h>
#include <LibIPC/Message.h>
#include <LibIPC/Stub.h>

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdefaulted-function-deleted"
#endif

namespace Messages::ImageDecoderServer {

enum class MessageID : i32 {
    InitTransport = 1,
    InitTransportResponse = 2,
    DecodeImage = 3,
    DecodeImageResponse = 4,
    CancelDecoding = 5,
    ConnectNewClients = 6,
    ConnectNewClientsResponse = 7,
};

class InitTransportResponse final : public IPC::Message {
public:
    InitTransportResponse(InitTransportResponse const&) = default;
    InitTransportResponse(InitTransportResponse&&) = default;
    InitTransportResponse& operator=(InitTransportResponse const&) = default;
    InitTransportResponse(int peer_pid) : m_peer_pid(move(peer_pid)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, int>)
    InitTransportResponse(WrappedReturnType&& value)
        : m_peer_pid(forward<WrappedReturnType>(value))
    {
    }

    virtual ~InitTransportResponse() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 3964467294;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::InitTransportResponse; }
    static i32 static_message_id() { return (int)MessageID::InitTransportResponse; }
    virtual const char* message_name() const override { return "ImageDecoderServer::InitTransportResponse"; }

    static ErrorOr<NonnullOwnPtr<InitTransportResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto peer_pid = TRY((decoder.decode<int>()));
        return make<InitTransportResponse>(move(peer_pid));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(int peer_pid)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::InitTransportResponse));
        TRY(stream.encode(peer_pid));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_peer_pid);
    }

    int peer_pid() const { return m_peer_pid; }

private:
    int m_peer_pid;
};

class InitTransport final : public IPC::Message {
public:
   typedef class InitTransportResponse ResponseType;

    InitTransport(InitTransport const&) = default;
    InitTransport(InitTransport&&) = default;
    InitTransport& operator=(InitTransport const&) = default;
    InitTransport(int peer_pid) : m_peer_pid(move(peer_pid)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, int>)
    InitTransport(WrappedReturnType&& value)
        : m_peer_pid(forward<WrappedReturnType>(value))
    {
    }

    virtual ~InitTransport() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 3964467294;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::InitTransport; }
    static i32 static_message_id() { return (int)MessageID::InitTransport; }
    virtual const char* message_name() const override { return "ImageDecoderServer::InitTransport"; }

    static ErrorOr<NonnullOwnPtr<InitTransport>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto peer_pid = TRY((decoder.decode<int>()));
        return make<InitTransport>(move(peer_pid));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(int peer_pid)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::InitTransport));
        TRY(stream.encode(peer_pid));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_peer_pid);
    }

    int peer_pid() const { return m_peer_pid; }

private:
    int m_peer_pid;
};

class DecodeImageResponse final : public IPC::Message {
public:
    DecodeImageResponse(DecodeImageResponse const&) = default;
    DecodeImageResponse(DecodeImageResponse&&) = default;
    DecodeImageResponse& operator=(DecodeImageResponse const&) = default;
    DecodeImageResponse(i64 image_id) : m_image_id(move(image_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, i64>)
    DecodeImageResponse(WrappedReturnType&& value)
        : m_image_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DecodeImageResponse() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 3964467294;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DecodeImageResponse; }
    static i32 static_message_id() { return (int)MessageID::DecodeImageResponse; }
    virtual const char* message_name() const override { return "ImageDecoderServer::DecodeImageResponse"; }

    static ErrorOr<NonnullOwnPtr<DecodeImageResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto image_id = TRY((decoder.decode<i64>()));
        return make<DecodeImageResponse>(move(image_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i64 image_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DecodeImageResponse));
        TRY(stream.encode(image_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_image_id);
    }

    i64 image_id() const { return m_image_id; }

private:
    i64 m_image_id;
};

class DecodeImage final : public IPC::Message {
public:
   typedef class DecodeImageResponse ResponseType;

    DecodeImage(DecodeImage const&) = default;
    DecodeImage(DecodeImage&&) = default;
    DecodeImage& operator=(DecodeImage const&) = default;
    DecodeImage(Core::AnonymousBuffer data, Optional<Gfx::IntSize> ideal_size, Optional<ByteString> mime_type) : m_data(move(data)), m_ideal_size(move(ideal_size)), m_mime_type(move(mime_type)) {}

    virtual ~DecodeImage() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 3964467294;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DecodeImage; }
    static i32 static_message_id() { return (int)MessageID::DecodeImage; }
    virtual const char* message_name() const override { return "ImageDecoderServer::DecodeImage"; }

    static ErrorOr<NonnullOwnPtr<DecodeImage>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto data = TRY((decoder.decode<Core::AnonymousBuffer>()));
        auto ideal_size = TRY((decoder.decode<Optional<Gfx::IntSize>>()));
        auto mime_type = TRY((decoder.decode<Optional<ByteString>>()));
        return make<DecodeImage>(move(data), move(ideal_size), move(mime_type));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(Core::AnonymousBuffer const& data, Optional<Gfx::IntSize> const& ideal_size, Optional<ByteString> const& mime_type)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DecodeImage));
        TRY(stream.encode(data));
        TRY(stream.encode(ideal_size));
        TRY(stream.encode(mime_type));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_data, m_ideal_size, m_mime_type);
    }

    const Core::AnonymousBuffer& data() const { return m_data; }
    Core::AnonymousBuffer take_data() { return move(m_data); }

    const Optional<Gfx::IntSize>& ideal_size() const { return m_ideal_size; }
    Optional<Gfx::IntSize> take_ideal_size() { return move(m_ideal_size); }

    const Optional<ByteString>& mime_type() const { return m_mime_type; }
    Optional<ByteString> take_mime_type() { return move(m_mime_type); }

private:
    Core::AnonymousBuffer m_data;
    Optional<Gfx::IntSize> m_ideal_size;
    Optional<ByteString> m_mime_type;
};

class CancelDecoding final : public IPC::Message {
public:
    CancelDecoding(CancelDecoding const&) = default;
    CancelDecoding(CancelDecoding&&) = default;
    CancelDecoding& operator=(CancelDecoding const&) = default;
    CancelDecoding(i64 image_id) : m_image_id(move(image_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, i64>)
    CancelDecoding(WrappedReturnType&& value)
        : m_image_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~CancelDecoding() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 3964467294;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::CancelDecoding; }
    static i32 static_message_id() { return (int)MessageID::CancelDecoding; }
    virtual const char* message_name() const override { return "ImageDecoderServer::CancelDecoding"; }

    static ErrorOr<NonnullOwnPtr<CancelDecoding>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto image_id = TRY((decoder.decode<i64>()));
        return make<CancelDecoding>(move(image_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i64 image_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::CancelDecoding));
        TRY(stream.encode(image_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_image_id);
    }

    i64 image_id() const { return m_image_id; }

private:
    i64 m_image_id;
};

class ConnectNewClientsResponse final : public IPC::Message {
public:
    ConnectNewClientsResponse(ConnectNewClientsResponse const&) = default;
    ConnectNewClientsResponse(ConnectNewClientsResponse&&) = default;
    ConnectNewClientsResponse& operator=(ConnectNewClientsResponse const&) = default;
    ConnectNewClientsResponse(Vector<IPC::File> sockets) : m_sockets(move(sockets)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, Vector<IPC::File>>)
    ConnectNewClientsResponse(WrappedReturnType&& value)
        : m_sockets(forward<WrappedReturnType>(value))
    {
    }

    virtual ~ConnectNewClientsResponse() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 3964467294;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::ConnectNewClientsResponse; }
    static i32 static_message_id() { return (int)MessageID::ConnectNewClientsResponse; }
    virtual const char* message_name() const override { return "ImageDecoderServer::ConnectNewClientsResponse"; }

    static ErrorOr<NonnullOwnPtr<ConnectNewClientsResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto sockets = TRY((decoder.decode<Vector<IPC::File>>()));
        return make<ConnectNewClientsResponse>(move(sockets));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(ReadonlySpan<IPC::File> sockets)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::ConnectNewClientsResponse));
        TRY(stream.encode(sockets));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_sockets);
    }

    const Vector<IPC::File>& sockets() const { return m_sockets; }
    Vector<IPC::File> take_sockets() { return move(m_sockets); }

private:
    Vector<IPC::File> m_sockets;
};

class ConnectNewClients final : public IPC::Message {
public:
   typedef class ConnectNewClientsResponse ResponseType;

    ConnectNewClients(ConnectNewClients const&) = default;
    ConnectNewClients(ConnectNewClients&&) = default;
    ConnectNewClients& operator=(ConnectNewClients const&) = default;
    ConnectNewClients(size_t count) : m_count(move(count)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, size_t>)
    ConnectNewClients(WrappedReturnType&& value)
        : m_count(forward<WrappedReturnType>(value))
    {
    }

    virtual ~ConnectNewClients() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 3964467294;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::ConnectNewClients; }
    static i32 static_message_id() { return (int)MessageID::ConnectNewClients; }
    virtual const char* message_name() const override { return "ImageDecoderServer::ConnectNewClients"; }

    static ErrorOr<NonnullOwnPtr<ConnectNewClients>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto count = TRY((decoder.decode<size_t>()));
        return make<ConnectNewClients>(move(count));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(size_t count)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::ConnectNewClients));
        TRY(stream.encode(count));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_count);
    }

    size_t count() const { return m_count; }

private:
    size_t m_count;
};

} // namespace Messages::ImageDecoderServer

template<typename LocalEndpoint, typename PeerEndpoint>
class ImageDecoderServerProxy {
public:
    // Used to disambiguate the constructor call.
    struct Tag { };

    ImageDecoderServerProxy(IPC::Connection<LocalEndpoint, PeerEndpoint>& connection, Tag)
        : m_connection(connection)
    { }

    int init_transport(int peer_pid) {
        return m_connection.template send_sync<Messages::ImageDecoderServer::InitTransport>(peer_pid)->peer_pid();
    }

    void async_init_transport(int peer_pid) {
        auto message_buffer = MUST(Messages::ImageDecoderServer::InitTransport::static_encode(peer_pid));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    IPC::IPCErrorOr<int> try_init_transport(int peer_pid) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::ImageDecoderServer::InitTransport>(peer_pid);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    i64 decode_image(Core::AnonymousBuffer data, Optional<Gfx::IntSize> ideal_size, Optional<ByteString> mime_type) {
        return m_connection.template send_sync<Messages::ImageDecoderServer::DecodeImage>(move(data), move(ideal_size), move(mime_type))->image_id();
    }

    void async_decode_image(Core::AnonymousBuffer const& data, Optional<Gfx::IntSize> const& ideal_size, Optional<ByteString> const& mime_type) {
        auto message_buffer = MUST(Messages::ImageDecoderServer::DecodeImage::static_encode(move(data), move(ideal_size), move(mime_type)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    IPC::IPCErrorOr<i64> try_decode_image(Core::AnonymousBuffer data, Optional<Gfx::IntSize> ideal_size, Optional<ByteString> mime_type) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::ImageDecoderServer::DecodeImage>(move(data), move(ideal_size), move(mime_type));
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    void async_cancel_decoding(i64 image_id) {
        auto message_buffer = MUST(Messages::ImageDecoderServer::CancelDecoding::static_encode(image_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    Vector<IPC::File> connect_new_clients(size_t count) {
        return m_connection.template send_sync<Messages::ImageDecoderServer::ConnectNewClients>(count)->take_sockets();
    }

    void async_connect_new_clients(size_t count) {
        auto message_buffer = MUST(Messages::ImageDecoderServer::ConnectNewClients::static_encode(count));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    IPC::IPCErrorOr<Vector<IPC::File>> try_connect_new_clients(size_t count) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::ImageDecoderServer::ConnectNewClients>(count);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

private:
    IPC::Connection<LocalEndpoint, PeerEndpoint>& m_connection;
};

template<typename LocalEndpoint, typename PeerEndpoint>
class ImageDecoderServerProxy;
class ImageDecoderServerStub;

class ImageDecoderServerEndpoint {
public:
    template<typename LocalEndpoint>
    using Proxy = ImageDecoderServerProxy<LocalEndpoint, ImageDecoderServerEndpoint>;
    using Stub = ImageDecoderServerStub;

    static u32 static_magic() { return 3964467294; }

    static ErrorOr<NonnullOwnPtr<IPC::Message>> decode_message(ReadonlyBytes buffer, [[maybe_unused]] Queue<IPC::File>& files)
    {
        FixedMemoryStream stream { buffer };
        auto message_endpoint_magic = TRY(stream.read_value<u32>());

        if (message_endpoint_magic != 3964467294) {
            return Error::from_string_literal("Endpoint magic number mismatch, not my message!");
        }

        auto message_id = TRY(stream.read_value<i32>());


        switch (message_id) {

        case (int)Messages::ImageDecoderServer::MessageID::InitTransport:
            return TRY(Messages::ImageDecoderServer::InitTransport::decode(stream, files));
        case (int)Messages::ImageDecoderServer::MessageID::InitTransportResponse:
            return TRY(Messages::ImageDecoderServer::InitTransportResponse::decode(stream, files));
        case (int)Messages::ImageDecoderServer::MessageID::DecodeImage:
            return TRY(Messages::ImageDecoderServer::DecodeImage::decode(stream, files));
        case (int)Messages::ImageDecoderServer::MessageID::DecodeImageResponse:
            return TRY(Messages::ImageDecoderServer::DecodeImageResponse::decode(stream, files));
        case (int)Messages::ImageDecoderServer::MessageID::CancelDecoding:
            return TRY(Messages::ImageDecoderServer::CancelDecoding::decode(stream, files));
        case (int)Messages::ImageDecoderServer::MessageID::ConnectNewClients:
            return TRY(Messages::ImageDecoderServer::ConnectNewClients::decode(stream, files));
        case (int)Messages::ImageDecoderServer::MessageID::ConnectNewClientsResponse:
            return TRY(Messages::ImageDecoderServer::ConnectNewClientsResponse::decode(stream, files));
        default:
            return Error::from_string_literal("Failed to decode ImageDecoderServer message");
        }

        VERIFY_NOT_REACHED();
    }

};

class ImageDecoderServerStub : public IPC::Stub {
public:
    ImageDecoderServerStub() { }
    virtual ~ImageDecoderServerStub() override { }

    virtual u32 magic() const override { return 3964467294; }
    virtual ByteString name() const override { return "ImageDecoderServer"; }

    virtual ErrorOr<OwnPtr<IPC::MessageBuffer>> handle(NonnullOwnPtr<IPC::Message> message) override
    {
        switch (message->message_id()) {

        case (int)Messages::ImageDecoderServer::MessageID::InitTransport: {
            [[maybe_unused]] auto& request = static_cast<Messages::ImageDecoderServer::InitTransport&>(*message);
            auto response = init_transport(request.peer_pid());
            return make<IPC::MessageBuffer>(TRY(response.encode()));
        }
        case (int)Messages::ImageDecoderServer::MessageID::DecodeImage: {
            [[maybe_unused]] auto& request = static_cast<Messages::ImageDecoderServer::DecodeImage&>(*message);
            auto response = decode_image(request.take_data(), request.take_ideal_size(), request.take_mime_type());
            return make<IPC::MessageBuffer>(TRY(response.encode()));
        }
        case (int)Messages::ImageDecoderServer::MessageID::CancelDecoding: {
            [[maybe_unused]] auto& request = static_cast<Messages::ImageDecoderServer::CancelDecoding&>(*message);
            cancel_decoding(request.image_id());
            return nullptr;
        }
        case (int)Messages::ImageDecoderServer::MessageID::ConnectNewClients: {
            [[maybe_unused]] auto& request = static_cast<Messages::ImageDecoderServer::ConnectNewClients&>(*message);
            auto response = connect_new_clients(request.count());
            return make<IPC::MessageBuffer>(TRY(response.encode()));
        }
        default:
            return Error::from_string_literal("Unknown message ID for ImageDecoderServer endpoint");
        }
    }

    virtual Messages::ImageDecoderServer::InitTransportResponse init_transport(int peer_pid) = 0;
    virtual Messages::ImageDecoderServer::DecodeImageResponse decode_image(Core::AnonymousBuffer data, Optional<Gfx::IntSize> ideal_size, Optional<ByteString> mime_type) = 0;
    virtual void cancel_decoding(i64 image_id) = 0;
    virtual Messages::ImageDecoderServer::ConnectNewClientsResponse connect_new_clients(size_t count) = 0;
};

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic pop
#endif
