#pragma once

#include <LibHTTP/HeaderMap.h>
#include <LibRequests/NetworkErrorEnum.h>
#include <LibRequests/RequestTimingInfo.h>
#include <LibURL/URL.h>
#include <AK/Error.h>
#include <AK/MemoryStream.h>
#include <AK/OwnPtr.h>
#include <AK/Result.h>
#include <AK/Utf8View.h>
#include <LibIPC/Connection.h>
#include <LibIPC/Decoder.h>
#include <LibIPC/Encoder.h>
#include <LibIPC/File.h>
#include <LibIPC/Message.h>
#include <LibIPC/Stub.h>

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdefaulted-function-deleted"
#endif

namespace Messages::RequestClient {

enum class MessageID : i32 {
    RequestStarted = 1,
    RequestFinished = 2,
    HeadersBecameAvailable = 3,
    WebsocketConnected = 4,
    WebsocketReceived = 5,
    WebsocketErrored = 6,
    WebsocketClosed = 7,
    WebsocketReadyStateChanged = 8,
    WebsocketSubprotocol = 9,
    WebsocketCertificateRequested = 10,
    CertificateRequested = 11,
};

class RequestStarted final : public IPC::Message {
public:
    RequestStarted(RequestStarted const&) = default;
    RequestStarted(RequestStarted&&) = default;
    RequestStarted& operator=(RequestStarted const&) = default;
    RequestStarted(i32 request_id, IPC::File fd) : m_request_id(move(request_id)), m_fd(move(fd)) {}

    virtual ~RequestStarted() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 1502652576;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::RequestStarted; }
    static i32 static_message_id() { return (int)MessageID::RequestStarted; }
    virtual const char* message_name() const override { return "RequestClient::RequestStarted"; }

    static ErrorOr<NonnullOwnPtr<RequestStarted>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto request_id = TRY((decoder.decode<i32>()));
        auto fd = TRY((decoder.decode<IPC::File>()));
        return make<RequestStarted>(move(request_id), move(fd));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i32 request_id, IPC::File const& fd)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::RequestStarted));
        TRY(stream.encode(request_id));
        TRY(stream.encode(fd));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_request_id, m_fd);
    }

    i32 request_id() const { return m_request_id; }

    const IPC::File& fd() const { return m_fd; }
    IPC::File take_fd() { return move(m_fd); }

private:
    i32 m_request_id;
    IPC::File m_fd;
};

class RequestFinished final : public IPC::Message {
public:
    RequestFinished(RequestFinished const&) = default;
    RequestFinished(RequestFinished&&) = default;
    RequestFinished& operator=(RequestFinished const&) = default;
    RequestFinished(i32 request_id, u64 total_size, Requests::RequestTimingInfo timing_info, Optional<Requests::NetworkError> network_error) : m_request_id(move(request_id)), m_total_size(move(total_size)), m_timing_info(move(timing_info)), m_network_error(move(network_error)) {}

    virtual ~RequestFinished() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 1502652576;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::RequestFinished; }
    static i32 static_message_id() { return (int)MessageID::RequestFinished; }
    virtual const char* message_name() const override { return "RequestClient::RequestFinished"; }

    static ErrorOr<NonnullOwnPtr<RequestFinished>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto request_id = TRY((decoder.decode<i32>()));
        auto total_size = TRY((decoder.decode<u64>()));
        auto timing_info = TRY((decoder.decode<Requests::RequestTimingInfo>()));
        auto network_error = TRY((decoder.decode<Optional<Requests::NetworkError>>()));
        return make<RequestFinished>(move(request_id), move(total_size), move(timing_info), move(network_error));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i32 request_id, u64 total_size, Requests::RequestTimingInfo const& timing_info, Optional<Requests::NetworkError> const& network_error)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::RequestFinished));
        TRY(stream.encode(request_id));
        TRY(stream.encode(total_size));
        TRY(stream.encode(timing_info));
        TRY(stream.encode(network_error));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_request_id, m_total_size, m_timing_info, m_network_error);
    }

    i32 request_id() const { return m_request_id; }

    u64 total_size() const { return m_total_size; }

    const Requests::RequestTimingInfo& timing_info() const { return m_timing_info; }
    Requests::RequestTimingInfo take_timing_info() { return move(m_timing_info); }

    const Optional<Requests::NetworkError>& network_error() const { return m_network_error; }
    Optional<Requests::NetworkError> take_network_error() { return move(m_network_error); }

private:
    i32 m_request_id;
    u64 m_total_size;
    Requests::RequestTimingInfo m_timing_info;
    Optional<Requests::NetworkError> m_network_error;
};

class HeadersBecameAvailable final : public IPC::Message {
public:
    HeadersBecameAvailable(HeadersBecameAvailable const&) = default;
    HeadersBecameAvailable(HeadersBecameAvailable&&) = default;
    HeadersBecameAvailable& operator=(HeadersBecameAvailable const&) = default;
    HeadersBecameAvailable(i32 request_id, HTTP::HeaderMap response_headers, Optional<u32> status_code, Optional<String> reason_phrase) : m_request_id(move(request_id)), m_response_headers(move(response_headers)), m_status_code(move(status_code)), m_reason_phrase(move(reason_phrase)) {}

    virtual ~HeadersBecameAvailable() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 1502652576;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::HeadersBecameAvailable; }
    static i32 static_message_id() { return (int)MessageID::HeadersBecameAvailable; }
    virtual const char* message_name() const override { return "RequestClient::HeadersBecameAvailable"; }

    static ErrorOr<NonnullOwnPtr<HeadersBecameAvailable>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto request_id = TRY((decoder.decode<i32>()));
        auto response_headers = TRY((decoder.decode<HTTP::HeaderMap>()));
        auto status_code = TRY((decoder.decode<Optional<u32>>()));
        auto reason_phrase = TRY((decoder.decode<Optional<String>>()));
        return make<HeadersBecameAvailable>(move(request_id), move(response_headers), move(status_code), move(reason_phrase));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i32 request_id, HTTP::HeaderMap const& response_headers, Optional<u32> const& status_code, Optional<String> const& reason_phrase)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::HeadersBecameAvailable));
        TRY(stream.encode(request_id));
        TRY(stream.encode(response_headers));
        TRY(stream.encode(status_code));
        TRY(stream.encode(reason_phrase));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_request_id, m_response_headers, m_status_code, m_reason_phrase);
    }

    i32 request_id() const { return m_request_id; }

    const HTTP::HeaderMap& response_headers() const { return m_response_headers; }
    HTTP::HeaderMap take_response_headers() { return move(m_response_headers); }

    const Optional<u32>& status_code() const { return m_status_code; }
    Optional<u32> take_status_code() { return move(m_status_code); }

    const Optional<String>& reason_phrase() const { return m_reason_phrase; }
    Optional<String> take_reason_phrase() { return move(m_reason_phrase); }

private:
    i32 m_request_id;
    HTTP::HeaderMap m_response_headers;
    Optional<u32> m_status_code;
    Optional<String> m_reason_phrase;
};

class WebsocketConnected final : public IPC::Message {
public:
    WebsocketConnected(WebsocketConnected const&) = default;
    WebsocketConnected(WebsocketConnected&&) = default;
    WebsocketConnected& operator=(WebsocketConnected const&) = default;
    WebsocketConnected(i64 websocket_id) : m_websocket_id(move(websocket_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, i64>)
    WebsocketConnected(WrappedReturnType&& value)
        : m_websocket_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~WebsocketConnected() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 1502652576;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::WebsocketConnected; }
    static i32 static_message_id() { return (int)MessageID::WebsocketConnected; }
    virtual const char* message_name() const override { return "RequestClient::WebsocketConnected"; }

    static ErrorOr<NonnullOwnPtr<WebsocketConnected>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto websocket_id = TRY((decoder.decode<i64>()));
        return make<WebsocketConnected>(move(websocket_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i64 websocket_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::WebsocketConnected));
        TRY(stream.encode(websocket_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_websocket_id);
    }

    i64 websocket_id() const { return m_websocket_id; }

private:
    i64 m_websocket_id;
};

class WebsocketReceived final : public IPC::Message {
public:
    WebsocketReceived(WebsocketReceived const&) = default;
    WebsocketReceived(WebsocketReceived&&) = default;
    WebsocketReceived& operator=(WebsocketReceived const&) = default;
    WebsocketReceived(i64 websocket_id, bool is_text, ByteBuffer data) : m_websocket_id(move(websocket_id)), m_is_text(move(is_text)), m_data(move(data)) {}

    virtual ~WebsocketReceived() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 1502652576;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::WebsocketReceived; }
    static i32 static_message_id() { return (int)MessageID::WebsocketReceived; }
    virtual const char* message_name() const override { return "RequestClient::WebsocketReceived"; }

    static ErrorOr<NonnullOwnPtr<WebsocketReceived>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto websocket_id = TRY((decoder.decode<i64>()));
        auto is_text = TRY((decoder.decode<bool>()));
        auto data = TRY((decoder.decode<ByteBuffer>()));
        return make<WebsocketReceived>(move(websocket_id), move(is_text), move(data));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i64 websocket_id, bool is_text, ReadonlyBytes data)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::WebsocketReceived));
        TRY(stream.encode(websocket_id));
        TRY(stream.encode(is_text));
        TRY(stream.encode(data));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_websocket_id, m_is_text, m_data);
    }

    i64 websocket_id() const { return m_websocket_id; }

    bool is_text() const { return m_is_text; }

    const ByteBuffer& data() const { return m_data; }
    ByteBuffer take_data() { return move(m_data); }

private:
    i64 m_websocket_id;
    bool m_is_text;
    ByteBuffer m_data;
};

class WebsocketErrored final : public IPC::Message {
public:
    WebsocketErrored(WebsocketErrored const&) = default;
    WebsocketErrored(WebsocketErrored&&) = default;
    WebsocketErrored& operator=(WebsocketErrored const&) = default;
    WebsocketErrored(i64 websocket_id, i32 message) : m_websocket_id(move(websocket_id)), m_message(move(message)) {}

    virtual ~WebsocketErrored() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 1502652576;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::WebsocketErrored; }
    static i32 static_message_id() { return (int)MessageID::WebsocketErrored; }
    virtual const char* message_name() const override { return "RequestClient::WebsocketErrored"; }

    static ErrorOr<NonnullOwnPtr<WebsocketErrored>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto websocket_id = TRY((decoder.decode<i64>()));
        auto message = TRY((decoder.decode<i32>()));
        return make<WebsocketErrored>(move(websocket_id), move(message));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i64 websocket_id, i32 message)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::WebsocketErrored));
        TRY(stream.encode(websocket_id));
        TRY(stream.encode(message));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_websocket_id, m_message);
    }

    i64 websocket_id() const { return m_websocket_id; }

    i32 message() const { return m_message; }

private:
    i64 m_websocket_id;
    i32 m_message;
};

class WebsocketClosed final : public IPC::Message {
public:
    WebsocketClosed(WebsocketClosed const&) = default;
    WebsocketClosed(WebsocketClosed&&) = default;
    WebsocketClosed& operator=(WebsocketClosed const&) = default;
    WebsocketClosed(i64 websocket_id, u16 code, ByteString reason, bool clean) : m_websocket_id(move(websocket_id)), m_code(move(code)), m_reason(move(reason)), m_clean(move(clean)) {}

    virtual ~WebsocketClosed() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 1502652576;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::WebsocketClosed; }
    static i32 static_message_id() { return (int)MessageID::WebsocketClosed; }
    virtual const char* message_name() const override { return "RequestClient::WebsocketClosed"; }

    static ErrorOr<NonnullOwnPtr<WebsocketClosed>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto websocket_id = TRY((decoder.decode<i64>()));
        auto code = TRY((decoder.decode<u16>()));
        auto reason = TRY((decoder.decode<ByteString>()));
        auto clean = TRY((decoder.decode<bool>()));
        return make<WebsocketClosed>(move(websocket_id), move(code), move(reason), move(clean));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i64 websocket_id, u16 code, StringView reason, bool clean)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::WebsocketClosed));
        TRY(stream.encode(websocket_id));
        TRY(stream.encode(code));
        TRY(stream.encode(reason));
        TRY(stream.encode(clean));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_websocket_id, m_code, m_reason, m_clean);
    }

    i64 websocket_id() const { return m_websocket_id; }

    u16 code() const { return m_code; }

    const ByteString& reason() const { return m_reason; }
    ByteString take_reason() { return move(m_reason); }

    bool clean() const { return m_clean; }

private:
    i64 m_websocket_id;
    u16 m_code;
    ByteString m_reason;
    bool m_clean;
};

class WebsocketReadyStateChanged final : public IPC::Message {
public:
    WebsocketReadyStateChanged(WebsocketReadyStateChanged const&) = default;
    WebsocketReadyStateChanged(WebsocketReadyStateChanged&&) = default;
    WebsocketReadyStateChanged& operator=(WebsocketReadyStateChanged const&) = default;
    WebsocketReadyStateChanged(i64 websocket_id, u32 ready_state) : m_websocket_id(move(websocket_id)), m_ready_state(move(ready_state)) {}

    virtual ~WebsocketReadyStateChanged() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 1502652576;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::WebsocketReadyStateChanged; }
    static i32 static_message_id() { return (int)MessageID::WebsocketReadyStateChanged; }
    virtual const char* message_name() const override { return "RequestClient::WebsocketReadyStateChanged"; }

    static ErrorOr<NonnullOwnPtr<WebsocketReadyStateChanged>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto websocket_id = TRY((decoder.decode<i64>()));
        auto ready_state = TRY((decoder.decode<u32>()));
        return make<WebsocketReadyStateChanged>(move(websocket_id), move(ready_state));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i64 websocket_id, u32 ready_state)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::WebsocketReadyStateChanged));
        TRY(stream.encode(websocket_id));
        TRY(stream.encode(ready_state));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_websocket_id, m_ready_state);
    }

    i64 websocket_id() const { return m_websocket_id; }

    u32 ready_state() const { return m_ready_state; }

private:
    i64 m_websocket_id;
    u32 m_ready_state;
};

class WebsocketSubprotocol final : public IPC::Message {
public:
    WebsocketSubprotocol(WebsocketSubprotocol const&) = default;
    WebsocketSubprotocol(WebsocketSubprotocol&&) = default;
    WebsocketSubprotocol& operator=(WebsocketSubprotocol const&) = default;
    WebsocketSubprotocol(i64 websocket_id, ByteString subprotocol) : m_websocket_id(move(websocket_id)), m_subprotocol(move(subprotocol)) {}

    virtual ~WebsocketSubprotocol() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 1502652576;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::WebsocketSubprotocol; }
    static i32 static_message_id() { return (int)MessageID::WebsocketSubprotocol; }
    virtual const char* message_name() const override { return "RequestClient::WebsocketSubprotocol"; }

    static ErrorOr<NonnullOwnPtr<WebsocketSubprotocol>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto websocket_id = TRY((decoder.decode<i64>()));
        auto subprotocol = TRY((decoder.decode<ByteString>()));
        return make<WebsocketSubprotocol>(move(websocket_id), move(subprotocol));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i64 websocket_id, StringView subprotocol)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::WebsocketSubprotocol));
        TRY(stream.encode(websocket_id));
        TRY(stream.encode(subprotocol));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_websocket_id, m_subprotocol);
    }

    i64 websocket_id() const { return m_websocket_id; }

    const ByteString& subprotocol() const { return m_subprotocol; }
    ByteString take_subprotocol() { return move(m_subprotocol); }

private:
    i64 m_websocket_id;
    ByteString m_subprotocol;
};

class WebsocketCertificateRequested final : public IPC::Message {
public:
    WebsocketCertificateRequested(WebsocketCertificateRequested const&) = default;
    WebsocketCertificateRequested(WebsocketCertificateRequested&&) = default;
    WebsocketCertificateRequested& operator=(WebsocketCertificateRequested const&) = default;
    WebsocketCertificateRequested(i64 websocket_id) : m_websocket_id(move(websocket_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, i64>)
    WebsocketCertificateRequested(WrappedReturnType&& value)
        : m_websocket_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~WebsocketCertificateRequested() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 1502652576;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::WebsocketCertificateRequested; }
    static i32 static_message_id() { return (int)MessageID::WebsocketCertificateRequested; }
    virtual const char* message_name() const override { return "RequestClient::WebsocketCertificateRequested"; }

    static ErrorOr<NonnullOwnPtr<WebsocketCertificateRequested>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto websocket_id = TRY((decoder.decode<i64>()));
        return make<WebsocketCertificateRequested>(move(websocket_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i64 websocket_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::WebsocketCertificateRequested));
        TRY(stream.encode(websocket_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_websocket_id);
    }

    i64 websocket_id() const { return m_websocket_id; }

private:
    i64 m_websocket_id;
};

class CertificateRequested final : public IPC::Message {
public:
    CertificateRequested(CertificateRequested const&) = default;
    CertificateRequested(CertificateRequested&&) = default;
    CertificateRequested& operator=(CertificateRequested const&) = default;
    CertificateRequested(i32 request_id) : m_request_id(move(request_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, i32>)
    CertificateRequested(WrappedReturnType&& value)
        : m_request_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~CertificateRequested() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 1502652576;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::CertificateRequested; }
    static i32 static_message_id() { return (int)MessageID::CertificateRequested; }
    virtual const char* message_name() const override { return "RequestClient::CertificateRequested"; }

    static ErrorOr<NonnullOwnPtr<CertificateRequested>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto request_id = TRY((decoder.decode<i32>()));
        return make<CertificateRequested>(move(request_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(i32 request_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::CertificateRequested));
        TRY(stream.encode(request_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_request_id);
    }

    i32 request_id() const { return m_request_id; }

private:
    i32 m_request_id;
};

} // namespace Messages::RequestClient

template<typename LocalEndpoint, typename PeerEndpoint>
class RequestClientProxy {
public:
    // Used to disambiguate the constructor call.
    struct Tag { };

    RequestClientProxy(IPC::Connection<LocalEndpoint, PeerEndpoint>& connection, Tag)
        : m_connection(connection)
    { }

    void async_request_started(i32 request_id, IPC::File const& fd) {
        auto message_buffer = MUST(Messages::RequestClient::RequestStarted::static_encode(request_id, move(fd)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_request_finished(i32 request_id, u64 total_size, Requests::RequestTimingInfo const& timing_info, Optional<Requests::NetworkError> const& network_error) {
        auto message_buffer = MUST(Messages::RequestClient::RequestFinished::static_encode(request_id, total_size, move(timing_info), move(network_error)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_headers_became_available(i32 request_id, HTTP::HeaderMap const& response_headers, Optional<u32> const& status_code, Optional<String> const& reason_phrase) {
        auto message_buffer = MUST(Messages::RequestClient::HeadersBecameAvailable::static_encode(request_id, move(response_headers), move(status_code), move(reason_phrase)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_websocket_connected(i64 websocket_id) {
        auto message_buffer = MUST(Messages::RequestClient::WebsocketConnected::static_encode(websocket_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_websocket_received(i64 websocket_id, bool is_text, ReadonlyBytes data) {
        auto message_buffer = MUST(Messages::RequestClient::WebsocketReceived::static_encode(websocket_id, is_text, data));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_websocket_errored(i64 websocket_id, i32 message) {
        auto message_buffer = MUST(Messages::RequestClient::WebsocketErrored::static_encode(websocket_id, message));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_websocket_closed(i64 websocket_id, u16 code, StringView reason, bool clean) {
        auto message_buffer = MUST(Messages::RequestClient::WebsocketClosed::static_encode(websocket_id, code, reason, clean));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_websocket_ready_state_changed(i64 websocket_id, u32 ready_state) {
        auto message_buffer = MUST(Messages::RequestClient::WebsocketReadyStateChanged::static_encode(websocket_id, ready_state));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_websocket_subprotocol(i64 websocket_id, StringView subprotocol) {
        auto message_buffer = MUST(Messages::RequestClient::WebsocketSubprotocol::static_encode(websocket_id, subprotocol));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_websocket_certificate_requested(i64 websocket_id) {
        auto message_buffer = MUST(Messages::RequestClient::WebsocketCertificateRequested::static_encode(websocket_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_certificate_requested(i32 request_id) {
        auto message_buffer = MUST(Messages::RequestClient::CertificateRequested::static_encode(request_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

private:
    IPC::Connection<LocalEndpoint, PeerEndpoint>& m_connection;
};

template<typename LocalEndpoint, typename PeerEndpoint>
class RequestClientProxy;
class RequestClientStub;

class RequestClientEndpoint {
public:
    template<typename LocalEndpoint>
    using Proxy = RequestClientProxy<LocalEndpoint, RequestClientEndpoint>;
    using Stub = RequestClientStub;

    static u32 static_magic() { return 1502652576; }

    static ErrorOr<NonnullOwnPtr<IPC::Message>> decode_message(ReadonlyBytes buffer, [[maybe_unused]] Queue<IPC::File>& files)
    {
        FixedMemoryStream stream { buffer };
        auto message_endpoint_magic = TRY(stream.read_value<u32>());

        if (message_endpoint_magic != 1502652576) {
            return Error::from_string_literal("Endpoint magic number mismatch, not my message!");
        }

        auto message_id = TRY(stream.read_value<i32>());


        switch (message_id) {

        case (int)Messages::RequestClient::MessageID::RequestStarted:
            return TRY(Messages::RequestClient::RequestStarted::decode(stream, files));
        case (int)Messages::RequestClient::MessageID::RequestFinished:
            return TRY(Messages::RequestClient::RequestFinished::decode(stream, files));
        case (int)Messages::RequestClient::MessageID::HeadersBecameAvailable:
            return TRY(Messages::RequestClient::HeadersBecameAvailable::decode(stream, files));
        case (int)Messages::RequestClient::MessageID::WebsocketConnected:
            return TRY(Messages::RequestClient::WebsocketConnected::decode(stream, files));
        case (int)Messages::RequestClient::MessageID::WebsocketReceived:
            return TRY(Messages::RequestClient::WebsocketReceived::decode(stream, files));
        case (int)Messages::RequestClient::MessageID::WebsocketErrored:
            return TRY(Messages::RequestClient::WebsocketErrored::decode(stream, files));
        case (int)Messages::RequestClient::MessageID::WebsocketClosed:
            return TRY(Messages::RequestClient::WebsocketClosed::decode(stream, files));
        case (int)Messages::RequestClient::MessageID::WebsocketReadyStateChanged:
            return TRY(Messages::RequestClient::WebsocketReadyStateChanged::decode(stream, files));
        case (int)Messages::RequestClient::MessageID::WebsocketSubprotocol:
            return TRY(Messages::RequestClient::WebsocketSubprotocol::decode(stream, files));
        case (int)Messages::RequestClient::MessageID::WebsocketCertificateRequested:
            return TRY(Messages::RequestClient::WebsocketCertificateRequested::decode(stream, files));
        case (int)Messages::RequestClient::MessageID::CertificateRequested:
            return TRY(Messages::RequestClient::CertificateRequested::decode(stream, files));
        default:
            return Error::from_string_literal("Failed to decode RequestClient message");
        }

        VERIFY_NOT_REACHED();
    }

};

class RequestClientStub : public IPC::Stub {
public:
    RequestClientStub() { }
    virtual ~RequestClientStub() override { }

    virtual u32 magic() const override { return 1502652576; }
    virtual ByteString name() const override { return "RequestClient"; }

    virtual ErrorOr<OwnPtr<IPC::MessageBuffer>> handle(NonnullOwnPtr<IPC::Message> message) override
    {
        switch (message->message_id()) {

        case (int)Messages::RequestClient::MessageID::RequestStarted: {
            [[maybe_unused]] auto& request = static_cast<Messages::RequestClient::RequestStarted&>(*message);
            request_started(request.request_id(), request.take_fd());
            return nullptr;
        }
        case (int)Messages::RequestClient::MessageID::RequestFinished: {
            [[maybe_unused]] auto& request = static_cast<Messages::RequestClient::RequestFinished&>(*message);
            request_finished(request.request_id(), request.total_size(), request.take_timing_info(), request.take_network_error());
            return nullptr;
        }
        case (int)Messages::RequestClient::MessageID::HeadersBecameAvailable: {
            [[maybe_unused]] auto& request = static_cast<Messages::RequestClient::HeadersBecameAvailable&>(*message);
            headers_became_available(request.request_id(), request.take_response_headers(), request.take_status_code(), request.take_reason_phrase());
            return nullptr;
        }
        case (int)Messages::RequestClient::MessageID::WebsocketConnected: {
            [[maybe_unused]] auto& request = static_cast<Messages::RequestClient::WebsocketConnected&>(*message);
            websocket_connected(request.websocket_id());
            return nullptr;
        }
        case (int)Messages::RequestClient::MessageID::WebsocketReceived: {
            [[maybe_unused]] auto& request = static_cast<Messages::RequestClient::WebsocketReceived&>(*message);
            websocket_received(request.websocket_id(), request.is_text(), request.take_data());
            return nullptr;
        }
        case (int)Messages::RequestClient::MessageID::WebsocketErrored: {
            [[maybe_unused]] auto& request = static_cast<Messages::RequestClient::WebsocketErrored&>(*message);
            websocket_errored(request.websocket_id(), request.message());
            return nullptr;
        }
        case (int)Messages::RequestClient::MessageID::WebsocketClosed: {
            [[maybe_unused]] auto& request = static_cast<Messages::RequestClient::WebsocketClosed&>(*message);
            websocket_closed(request.websocket_id(), request.code(), request.take_reason(), request.clean());
            return nullptr;
        }
        case (int)Messages::RequestClient::MessageID::WebsocketReadyStateChanged: {
            [[maybe_unused]] auto& request = static_cast<Messages::RequestClient::WebsocketReadyStateChanged&>(*message);
            websocket_ready_state_changed(request.websocket_id(), request.ready_state());
            return nullptr;
        }
        case (int)Messages::RequestClient::MessageID::WebsocketSubprotocol: {
            [[maybe_unused]] auto& request = static_cast<Messages::RequestClient::WebsocketSubprotocol&>(*message);
            websocket_subprotocol(request.websocket_id(), request.take_subprotocol());
            return nullptr;
        }
        case (int)Messages::RequestClient::MessageID::WebsocketCertificateRequested: {
            [[maybe_unused]] auto& request = static_cast<Messages::RequestClient::WebsocketCertificateRequested&>(*message);
            websocket_certificate_requested(request.websocket_id());
            return nullptr;
        }
        case (int)Messages::RequestClient::MessageID::CertificateRequested: {
            [[maybe_unused]] auto& request = static_cast<Messages::RequestClient::CertificateRequested&>(*message);
            certificate_requested(request.request_id());
            return nullptr;
        }
        default:
            return Error::from_string_literal("Unknown message ID for RequestClient endpoint");
        }
    }

    virtual void request_started(i32 request_id, IPC::File fd) = 0;
    virtual void request_finished(i32 request_id, u64 total_size, Requests::RequestTimingInfo timing_info, Optional<Requests::NetworkError> network_error) = 0;
    virtual void headers_became_available(i32 request_id, HTTP::HeaderMap response_headers, Optional<u32> status_code, Optional<String> reason_phrase) = 0;
    virtual void websocket_connected(i64 websocket_id) = 0;
    virtual void websocket_received(i64 websocket_id, bool is_text, ByteBuffer data) = 0;
    virtual void websocket_errored(i64 websocket_id, i32 message) = 0;
    virtual void websocket_closed(i64 websocket_id, u16 code, ByteString reason, bool clean) = 0;
    virtual void websocket_ready_state_changed(i64 websocket_id, u32 ready_state) = 0;
    virtual void websocket_subprotocol(i64 websocket_id, ByteString subprotocol) = 0;
    virtual void websocket_certificate_requested(i64 websocket_id) = 0;
    virtual void certificate_requested(i32 request_id) = 0;
};

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic pop
#endif
