#pragma once

#include <AK/Error.h>
#include <AK/MemoryStream.h>
#include <AK/OwnPtr.h>
#include <AK/Result.h>
#include <AK/Utf8View.h>
#include <LibIPC/Connection.h>
#include <LibIPC/Decoder.h>
#include <LibIPC/Encoder.h>
#include <LibIPC/File.h>
#include <LibIPC/Message.h>
#include <LibIPC/Stub.h>

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdefaulted-function-deleted"
#endif

namespace Messages::WebUIClient {

enum class MessageID : i32 {
    ReceivedMessage = 1,
};

class ReceivedMessage final : public IPC::Message {
public:
    ReceivedMessage(ReceivedMessage const&) = default;
    ReceivedMessage(ReceivedMessage&&) = default;
    ReceivedMessage& operator=(ReceivedMessage const&) = default;
    ReceivedMessage(String name, JsonValue data) : m_name(move(name)), m_data(move(data)) {}

    virtual ~ReceivedMessage() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 937913408;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::ReceivedMessage; }
    static i32 static_message_id() { return (int)MessageID::ReceivedMessage; }
    virtual const char* message_name() const override { return "WebUIClient::ReceivedMessage"; }

    static ErrorOr<NonnullOwnPtr<ReceivedMessage>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto name = TRY((decoder.decode<String>()));
        auto data = TRY((decoder.decode<JsonValue>()));
        return make<ReceivedMessage>(move(name), move(data));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(StringView name, JsonValue const& data)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::ReceivedMessage));
        TRY(stream.encode(name));
        TRY(stream.encode(data));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_name, m_data);
    }

    const String& name() const { return m_name; }
    String take_name() { return move(m_name); }

    const JsonValue& data() const { return m_data; }
    JsonValue take_data() { return move(m_data); }

private:
    String m_name;
    JsonValue m_data;
};

} // namespace Messages::WebUIClient

template<typename LocalEndpoint, typename PeerEndpoint>
class WebUIClientProxy {
public:
    // Used to disambiguate the constructor call.
    struct Tag { };

    WebUIClientProxy(IPC::Connection<LocalEndpoint, PeerEndpoint>& connection, Tag)
        : m_connection(connection)
    { }

    void async_received_message(StringView name, JsonValue const& data) {
        VERIFY(Utf8View { name }.validate());
        auto message_buffer = MUST(Messages::WebUIClient::ReceivedMessage::static_encode(name, move(data)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_received_message(String const& name, JsonValue const& data) {
        auto message_buffer = MUST(Messages::WebUIClient::ReceivedMessage::static_encode(name, move(data)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

private:
    IPC::Connection<LocalEndpoint, PeerEndpoint>& m_connection;
};

template<typename LocalEndpoint, typename PeerEndpoint>
class WebUIClientProxy;
class WebUIClientStub;

class WebUIClientEndpoint {
public:
    template<typename LocalEndpoint>
    using Proxy = WebUIClientProxy<LocalEndpoint, WebUIClientEndpoint>;
    using Stub = WebUIClientStub;

    static u32 static_magic() { return 937913408; }

    static ErrorOr<NonnullOwnPtr<IPC::Message>> decode_message(ReadonlyBytes buffer, [[maybe_unused]] Queue<IPC::File>& files)
    {
        FixedMemoryStream stream { buffer };
        auto message_endpoint_magic = TRY(stream.read_value<u32>());

        if (message_endpoint_magic != 937913408) {
            return Error::from_string_literal("Endpoint magic number mismatch, not my message!");
        }

        auto message_id = TRY(stream.read_value<i32>());


        switch (message_id) {

        case (int)Messages::WebUIClient::MessageID::ReceivedMessage:
            return TRY(Messages::WebUIClient::ReceivedMessage::decode(stream, files));
        default:
            return Error::from_string_literal("Failed to decode WebUIClient message");
        }

        VERIFY_NOT_REACHED();
    }

};

class WebUIClientStub : public IPC::Stub {
public:
    WebUIClientStub() { }
    virtual ~WebUIClientStub() override { }

    virtual u32 magic() const override { return 937913408; }
    virtual ByteString name() const override { return "WebUIClient"; }

    virtual ErrorOr<OwnPtr<IPC::MessageBuffer>> handle(NonnullOwnPtr<IPC::Message> message) override
    {
        switch (message->message_id()) {

        case (int)Messages::WebUIClient::MessageID::ReceivedMessage: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebUIClient::ReceivedMessage&>(*message);
            received_message(request.take_name(), request.take_data());
            return nullptr;
        }
        default:
            return Error::from_string_literal("Unknown message ID for WebUIClient endpoint");
        }
    }

    virtual void received_message(String name, JsonValue data) = 0;
};

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic pop
#endif
