#pragma once

#include <AK/Error.h>
#include <AK/MemoryStream.h>
#include <AK/OwnPtr.h>
#include <AK/Result.h>
#include <AK/Utf8View.h>
#include <LibIPC/Connection.h>
#include <LibIPC/Decoder.h>
#include <LibIPC/Encoder.h>
#include <LibIPC/File.h>
#include <LibIPC/Message.h>
#include <LibIPC/Stub.h>

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdefaulted-function-deleted"
#endif

namespace Messages::WebUIServer {

enum class MessageID : i32 {
    SendMessage = 1,
};

class SendMessage final : public IPC::Message {
public:
    SendMessage(SendMessage const&) = default;
    SendMessage(SendMessage&&) = default;
    SendMessage& operator=(SendMessage const&) = default;
    SendMessage(String name, JsonValue data) : m_name(move(name)), m_data(move(data)) {}

    virtual ~SendMessage() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 818868636;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::SendMessage; }
    static i32 static_message_id() { return (int)MessageID::SendMessage; }
    virtual const char* message_name() const override { return "WebUIServer::SendMessage"; }

    static ErrorOr<NonnullOwnPtr<SendMessage>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto name = TRY((decoder.decode<String>()));
        auto data = TRY((decoder.decode<JsonValue>()));
        return make<SendMessage>(move(name), move(data));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(StringView name, JsonValue const& data)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::SendMessage));
        TRY(stream.encode(name));
        TRY(stream.encode(data));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_name, m_data);
    }

    const String& name() const { return m_name; }
    String take_name() { return move(m_name); }

    const JsonValue& data() const { return m_data; }
    JsonValue take_data() { return move(m_data); }

private:
    String m_name;
    JsonValue m_data;
};

} // namespace Messages::WebUIServer

template<typename LocalEndpoint, typename PeerEndpoint>
class WebUIServerProxy {
public:
    // Used to disambiguate the constructor call.
    struct Tag { };

    WebUIServerProxy(IPC::Connection<LocalEndpoint, PeerEndpoint>& connection, Tag)
        : m_connection(connection)
    { }

    void async_send_message(StringView name, JsonValue const& data) {
        VERIFY(Utf8View { name }.validate());
        auto message_buffer = MUST(Messages::WebUIServer::SendMessage::static_encode(name, move(data)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_send_message(String const& name, JsonValue const& data) {
        auto message_buffer = MUST(Messages::WebUIServer::SendMessage::static_encode(name, move(data)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

private:
    IPC::Connection<LocalEndpoint, PeerEndpoint>& m_connection;
};

template<typename LocalEndpoint, typename PeerEndpoint>
class WebUIServerProxy;
class WebUIServerStub;

class WebUIServerEndpoint {
public:
    template<typename LocalEndpoint>
    using Proxy = WebUIServerProxy<LocalEndpoint, WebUIServerEndpoint>;
    using Stub = WebUIServerStub;

    static u32 static_magic() { return 818868636; }

    static ErrorOr<NonnullOwnPtr<IPC::Message>> decode_message(ReadonlyBytes buffer, [[maybe_unused]] Queue<IPC::File>& files)
    {
        FixedMemoryStream stream { buffer };
        auto message_endpoint_magic = TRY(stream.read_value<u32>());

        if (message_endpoint_magic != 818868636) {
            return Error::from_string_literal("Endpoint magic number mismatch, not my message!");
        }

        auto message_id = TRY(stream.read_value<i32>());


        switch (message_id) {

        case (int)Messages::WebUIServer::MessageID::SendMessage:
            return TRY(Messages::WebUIServer::SendMessage::decode(stream, files));
        default:
            return Error::from_string_literal("Failed to decode WebUIServer message");
        }

        VERIFY_NOT_REACHED();
    }

};

class WebUIServerStub : public IPC::Stub {
public:
    WebUIServerStub() { }
    virtual ~WebUIServerStub() override { }

    virtual u32 magic() const override { return 818868636; }
    virtual ByteString name() const override { return "WebUIServer"; }

    virtual ErrorOr<OwnPtr<IPC::MessageBuffer>> handle(NonnullOwnPtr<IPC::Message> message) override
    {
        switch (message->message_id()) {

        case (int)Messages::WebUIServer::MessageID::SendMessage: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebUIServer::SendMessage&>(*message);
            send_message(request.take_name(), request.take_data());
            return nullptr;
        }
        default:
            return Error::from_string_literal("Unknown message ID for WebUIServer endpoint");
        }
    }

    virtual void send_message(String name, JsonValue data) = 0;
};

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic pop
#endif
