#pragma once

#include <LibCore/AnonymousBuffer.h>
#include <LibGfx/Color.h>
#include <LibGfx/Cursor.h>
#include <LibGfx/ShareableBitmap.h>
#include <LibURL/URL.h>
#include <LibWeb/Cookie/Cookie.h>
#include <LibWeb/Cookie/ParsedCookie.h>
#include <LibWeb/CSS/Selector.h>
#include <LibWeb/CSS/StyleSheetIdentifier.h>
#include <LibWeb/HTML/ActivateTab.h>
#include <LibWeb/HTML/AudioPlayState.h>
#include <LibWeb/HTML/FileFilter.h>
#include <LibWeb/HTML/SelectedFile.h>
#include <LibWeb/HTML/SelectItem.h>
#include <LibWeb/HTML/WebViewHints.h>
#include <LibWeb/Page/EventResult.h>
#include <LibWeb/Page/Page.h>
#include <LibWebView/Attribute.h>
#include <LibWebView/ConsoleOutput.h>
#include <LibWebView/DOMNodeProperties.h>
#include <LibWebView/Mutation.h>
#include <LibWebView/PageInfo.h>
#include <LibWebView/ProcessHandle.h>
#include <AK/Error.h>
#include <AK/MemoryStream.h>
#include <AK/OwnPtr.h>
#include <AK/Result.h>
#include <AK/Utf8View.h>
#include <LibIPC/Connection.h>
#include <LibIPC/Decoder.h>
#include <LibIPC/Encoder.h>
#include <LibIPC/File.h>
#include <LibIPC/Message.h>
#include <LibIPC/Stub.h>

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdefaulted-function-deleted"
#endif

namespace Messages::WebContentClient {

enum class MessageID : i32 {
    DidRequestNewProcessForNavigation = 1,
    DidStartLoading = 2,
    DidFinishLoading = 3,
    DidRequestRefresh = 4,
    DidPaint = 5,
    DidRequestCursorChange = 6,
    DidChangeTitle = 7,
    DidChangeUrl = 8,
    DidRequestTooltipOverride = 9,
    DidStopTooltipOverride = 10,
    DidEnterTooltipArea = 11,
    DidLeaveTooltipArea = 12,
    DidHoverLink = 13,
    DidUnhoverLink = 14,
    DidClickLink = 15,
    DidMiddleClickLink = 16,
    DidRequestContextMenu = 17,
    DidRequestLinkContextMenu = 18,
    DidRequestImageContextMenu = 19,
    DidRequestMediaContextMenu = 20,
    DidRequestAlert = 21,
    DidRequestConfirm = 22,
    DidRequestPrompt = 23,
    DidRequestSetPromptText = 24,
    DidRequestAcceptDialog = 25,
    DidRequestDismissDialog = 26,
    DidGetSource = 27,
    DidInspectDomTree = 28,
    DidInspectDomNode = 29,
    DidInspectAccessibilityTree = 30,
    DidGetHoveredNodeId = 31,
    DidFinishEditingDomNode = 32,
    DidMutateDom = 33,
    DidGetDomNodeHtml = 34,
    DidListStyleSheets = 35,
    DidGetStyleSheetSource = 36,
    DidTakeScreenshot = 37,
    DidGetInternalPageInfo = 38,
    DidChangeFavicon = 39,
    DidRequestAllCookies = 40,
    DidRequestAllCookiesResponse = 41,
    DidRequestNamedCookie = 42,
    DidRequestNamedCookieResponse = 43,
    DidRequestCookie = 44,
    DidRequestCookieResponse = 45,
    DidSetCookie = 46,
    DidSetCookieResponse = 47,
    DidUpdateCookie = 48,
    DidExpireCookiesWithTimeOffset = 49,
    DidUpdateResourceCount = 50,
    DidRequestNewWebView = 51,
    DidRequestNewWebViewResponse = 52,
    DidRequestActivateTab = 53,
    DidCloseBrowsingContext = 54,
    DidRequestRestoreWindow = 55,
    DidRequestRepositionWindow = 56,
    DidRequestResizeWindow = 57,
    DidRequestMaximizeWindow = 58,
    DidRequestMinimizeWindow = 59,
    DidRequestFullscreenWindow = 60,
    DidRequestFile = 61,
    DidRequestColorPicker = 62,
    DidRequestFilePicker = 63,
    DidRequestSelectDropdown = 64,
    DidFinishHandlingInputEvent = 65,
    DidChangeThemeColor = 66,
    DidInsertClipboardEntry = 67,
    DidUpdateNavigationButtonsState = 68,
    DidAllocateBackingStores = 69,
    DidChangeAudioPlayState = 70,
    DidExecuteJsConsoleInput = 71,
    DidOutputJsConsoleMessage = 72,
    DidGetJsConsoleMessages = 73,
    DidFinishTest = 74,
    DidSetTestTimeout = 75,
    DidSetBrowserZoom = 76,
    DidFindInPage = 77,
    RequestWorkerAgent = 78,
    RequestWorkerAgentResponse = 79,
};

class DidRequestNewProcessForNavigation final : public IPC::Message {
public:
    DidRequestNewProcessForNavigation(DidRequestNewProcessForNavigation const&) = default;
    DidRequestNewProcessForNavigation(DidRequestNewProcessForNavigation&&) = default;
    DidRequestNewProcessForNavigation& operator=(DidRequestNewProcessForNavigation const&) = default;
    DidRequestNewProcessForNavigation(u64 page_id, URL::URL url) : m_page_id(move(page_id)), m_url(move(url)) {}

    virtual ~DidRequestNewProcessForNavigation() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestNewProcessForNavigation; }
    static i32 static_message_id() { return (int)MessageID::DidRequestNewProcessForNavigation; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestNewProcessForNavigation"; }

    static ErrorOr<NonnullOwnPtr<DidRequestNewProcessForNavigation>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto url = TRY((decoder.decode<URL::URL>()));
        return make<DidRequestNewProcessForNavigation>(move(page_id), move(url));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, URL::URL const& url)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestNewProcessForNavigation));
        TRY(stream.encode(page_id));
        TRY(stream.encode(url));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_url);
    }

    u64 page_id() const { return m_page_id; }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

private:
    u64 m_page_id;
    URL::URL m_url;
};

class DidStartLoading final : public IPC::Message {
public:
    DidStartLoading(DidStartLoading const&) = default;
    DidStartLoading(DidStartLoading&&) = default;
    DidStartLoading& operator=(DidStartLoading const&) = default;
    DidStartLoading(u64 page_id, URL::URL url, bool is_redirect) : m_page_id(move(page_id)), m_url(move(url)), m_is_redirect(move(is_redirect)) {}

    virtual ~DidStartLoading() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidStartLoading; }
    static i32 static_message_id() { return (int)MessageID::DidStartLoading; }
    virtual const char* message_name() const override { return "WebContentClient::DidStartLoading"; }

    static ErrorOr<NonnullOwnPtr<DidStartLoading>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto url = TRY((decoder.decode<URL::URL>()));
        auto is_redirect = TRY((decoder.decode<bool>()));
        return make<DidStartLoading>(move(page_id), move(url), move(is_redirect));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, URL::URL const& url, bool is_redirect)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidStartLoading));
        TRY(stream.encode(page_id));
        TRY(stream.encode(url));
        TRY(stream.encode(is_redirect));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_url, m_is_redirect);
    }

    u64 page_id() const { return m_page_id; }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    bool is_redirect() const { return m_is_redirect; }

private:
    u64 m_page_id;
    URL::URL m_url;
    bool m_is_redirect;
};

class DidFinishLoading final : public IPC::Message {
public:
    DidFinishLoading(DidFinishLoading const&) = default;
    DidFinishLoading(DidFinishLoading&&) = default;
    DidFinishLoading& operator=(DidFinishLoading const&) = default;
    DidFinishLoading(u64 page_id, URL::URL url) : m_page_id(move(page_id)), m_url(move(url)) {}

    virtual ~DidFinishLoading() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidFinishLoading; }
    static i32 static_message_id() { return (int)MessageID::DidFinishLoading; }
    virtual const char* message_name() const override { return "WebContentClient::DidFinishLoading"; }

    static ErrorOr<NonnullOwnPtr<DidFinishLoading>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto url = TRY((decoder.decode<URL::URL>()));
        return make<DidFinishLoading>(move(page_id), move(url));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, URL::URL const& url)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidFinishLoading));
        TRY(stream.encode(page_id));
        TRY(stream.encode(url));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_url);
    }

    u64 page_id() const { return m_page_id; }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

private:
    u64 m_page_id;
    URL::URL m_url;
};

class DidRequestRefresh final : public IPC::Message {
public:
    DidRequestRefresh(DidRequestRefresh const&) = default;
    DidRequestRefresh(DidRequestRefresh&&) = default;
    DidRequestRefresh& operator=(DidRequestRefresh const&) = default;
    DidRequestRefresh(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestRefresh(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestRefresh() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestRefresh; }
    static i32 static_message_id() { return (int)MessageID::DidRequestRefresh; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestRefresh"; }

    static ErrorOr<NonnullOwnPtr<DidRequestRefresh>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        return make<DidRequestRefresh>(move(page_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestRefresh));
        TRY(stream.encode(page_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id);
    }

    u64 page_id() const { return m_page_id; }

private:
    u64 m_page_id;
};

class DidPaint final : public IPC::Message {
public:
    DidPaint(DidPaint const&) = default;
    DidPaint(DidPaint&&) = default;
    DidPaint& operator=(DidPaint const&) = default;
    DidPaint(u64 page_id, Gfx::IntRect content_rect, i32 bitmap_id) : m_page_id(move(page_id)), m_content_rect(move(content_rect)), m_bitmap_id(move(bitmap_id)) {}

    virtual ~DidPaint() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidPaint; }
    static i32 static_message_id() { return (int)MessageID::DidPaint; }
    virtual const char* message_name() const override { return "WebContentClient::DidPaint"; }

    static ErrorOr<NonnullOwnPtr<DidPaint>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto content_rect = TRY((decoder.decode<Gfx::IntRect>()));
        auto bitmap_id = TRY((decoder.decode<i32>()));
        return make<DidPaint>(move(page_id), move(content_rect), move(bitmap_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Gfx::IntRect const& content_rect, i32 bitmap_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidPaint));
        TRY(stream.encode(page_id));
        TRY(stream.encode(content_rect));
        TRY(stream.encode(bitmap_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_content_rect, m_bitmap_id);
    }

    u64 page_id() const { return m_page_id; }

    const Gfx::IntRect& content_rect() const { return m_content_rect; }
    Gfx::IntRect take_content_rect() { return move(m_content_rect); }

    i32 bitmap_id() const { return m_bitmap_id; }

private:
    u64 m_page_id;
    Gfx::IntRect m_content_rect;
    i32 m_bitmap_id;
};

class DidRequestCursorChange final : public IPC::Message {
public:
    DidRequestCursorChange(DidRequestCursorChange const&) = default;
    DidRequestCursorChange(DidRequestCursorChange&&) = default;
    DidRequestCursorChange& operator=(DidRequestCursorChange const&) = default;
    DidRequestCursorChange(u64 page_id, Gfx::Cursor cursor) : m_page_id(move(page_id)), m_cursor(move(cursor)) {}

    virtual ~DidRequestCursorChange() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestCursorChange; }
    static i32 static_message_id() { return (int)MessageID::DidRequestCursorChange; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestCursorChange"; }

    static ErrorOr<NonnullOwnPtr<DidRequestCursorChange>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto cursor = TRY((decoder.decode<Gfx::Cursor>()));
        return make<DidRequestCursorChange>(move(page_id), move(cursor));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Gfx::Cursor const& cursor)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestCursorChange));
        TRY(stream.encode(page_id));
        TRY(stream.encode(cursor));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_cursor);
    }

    u64 page_id() const { return m_page_id; }

    const Gfx::Cursor& cursor() const { return m_cursor; }
    Gfx::Cursor take_cursor() { return move(m_cursor); }

private:
    u64 m_page_id;
    Gfx::Cursor m_cursor;
};

class DidChangeTitle final : public IPC::Message {
public:
    DidChangeTitle(DidChangeTitle const&) = default;
    DidChangeTitle(DidChangeTitle&&) = default;
    DidChangeTitle& operator=(DidChangeTitle const&) = default;
    DidChangeTitle(u64 page_id, ByteString title) : m_page_id(move(page_id)), m_title(move(title)) {}

    virtual ~DidChangeTitle() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidChangeTitle; }
    static i32 static_message_id() { return (int)MessageID::DidChangeTitle; }
    virtual const char* message_name() const override { return "WebContentClient::DidChangeTitle"; }

    static ErrorOr<NonnullOwnPtr<DidChangeTitle>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto title = TRY((decoder.decode<ByteString>()));
        return make<DidChangeTitle>(move(page_id), move(title));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, StringView title)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidChangeTitle));
        TRY(stream.encode(page_id));
        TRY(stream.encode(title));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_title);
    }

    u64 page_id() const { return m_page_id; }

    const ByteString& title() const { return m_title; }
    ByteString take_title() { return move(m_title); }

private:
    u64 m_page_id;
    ByteString m_title;
};

class DidChangeUrl final : public IPC::Message {
public:
    DidChangeUrl(DidChangeUrl const&) = default;
    DidChangeUrl(DidChangeUrl&&) = default;
    DidChangeUrl& operator=(DidChangeUrl const&) = default;
    DidChangeUrl(u64 page_id, URL::URL url) : m_page_id(move(page_id)), m_url(move(url)) {}

    virtual ~DidChangeUrl() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidChangeUrl; }
    static i32 static_message_id() { return (int)MessageID::DidChangeUrl; }
    virtual const char* message_name() const override { return "WebContentClient::DidChangeUrl"; }

    static ErrorOr<NonnullOwnPtr<DidChangeUrl>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto url = TRY((decoder.decode<URL::URL>()));
        return make<DidChangeUrl>(move(page_id), move(url));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, URL::URL const& url)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidChangeUrl));
        TRY(stream.encode(page_id));
        TRY(stream.encode(url));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_url);
    }

    u64 page_id() const { return m_page_id; }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

private:
    u64 m_page_id;
    URL::URL m_url;
};

class DidRequestTooltipOverride final : public IPC::Message {
public:
    DidRequestTooltipOverride(DidRequestTooltipOverride const&) = default;
    DidRequestTooltipOverride(DidRequestTooltipOverride&&) = default;
    DidRequestTooltipOverride& operator=(DidRequestTooltipOverride const&) = default;
    DidRequestTooltipOverride(u64 page_id, Gfx::IntPoint position, ByteString title) : m_page_id(move(page_id)), m_position(move(position)), m_title(move(title)) {}

    virtual ~DidRequestTooltipOverride() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestTooltipOverride; }
    static i32 static_message_id() { return (int)MessageID::DidRequestTooltipOverride; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestTooltipOverride"; }

    static ErrorOr<NonnullOwnPtr<DidRequestTooltipOverride>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto position = TRY((decoder.decode<Gfx::IntPoint>()));
        auto title = TRY((decoder.decode<ByteString>()));
        return make<DidRequestTooltipOverride>(move(page_id), move(position), move(title));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Gfx::IntPoint position, StringView title)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestTooltipOverride));
        TRY(stream.encode(page_id));
        TRY(stream.encode(position));
        TRY(stream.encode(title));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_position, m_title);
    }

    u64 page_id() const { return m_page_id; }

    Gfx::IntPoint position() const { return m_position; }

    const ByteString& title() const { return m_title; }
    ByteString take_title() { return move(m_title); }

private:
    u64 m_page_id;
    Gfx::IntPoint m_position;
    ByteString m_title;
};

class DidStopTooltipOverride final : public IPC::Message {
public:
    DidStopTooltipOverride(DidStopTooltipOverride const&) = default;
    DidStopTooltipOverride(DidStopTooltipOverride&&) = default;
    DidStopTooltipOverride& operator=(DidStopTooltipOverride const&) = default;
    DidStopTooltipOverride(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidStopTooltipOverride(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidStopTooltipOverride() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidStopTooltipOverride; }
    static i32 static_message_id() { return (int)MessageID::DidStopTooltipOverride; }
    virtual const char* message_name() const override { return "WebContentClient::DidStopTooltipOverride"; }

    static ErrorOr<NonnullOwnPtr<DidStopTooltipOverride>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        return make<DidStopTooltipOverride>(move(page_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidStopTooltipOverride));
        TRY(stream.encode(page_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id);
    }

    u64 page_id() const { return m_page_id; }

private:
    u64 m_page_id;
};

class DidEnterTooltipArea final : public IPC::Message {
public:
    DidEnterTooltipArea(DidEnterTooltipArea const&) = default;
    DidEnterTooltipArea(DidEnterTooltipArea&&) = default;
    DidEnterTooltipArea& operator=(DidEnterTooltipArea const&) = default;
    DidEnterTooltipArea(u64 page_id, ByteString title) : m_page_id(move(page_id)), m_title(move(title)) {}

    virtual ~DidEnterTooltipArea() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidEnterTooltipArea; }
    static i32 static_message_id() { return (int)MessageID::DidEnterTooltipArea; }
    virtual const char* message_name() const override { return "WebContentClient::DidEnterTooltipArea"; }

    static ErrorOr<NonnullOwnPtr<DidEnterTooltipArea>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto title = TRY((decoder.decode<ByteString>()));
        return make<DidEnterTooltipArea>(move(page_id), move(title));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, StringView title)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidEnterTooltipArea));
        TRY(stream.encode(page_id));
        TRY(stream.encode(title));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_title);
    }

    u64 page_id() const { return m_page_id; }

    const ByteString& title() const { return m_title; }
    ByteString take_title() { return move(m_title); }

private:
    u64 m_page_id;
    ByteString m_title;
};

class DidLeaveTooltipArea final : public IPC::Message {
public:
    DidLeaveTooltipArea(DidLeaveTooltipArea const&) = default;
    DidLeaveTooltipArea(DidLeaveTooltipArea&&) = default;
    DidLeaveTooltipArea& operator=(DidLeaveTooltipArea const&) = default;
    DidLeaveTooltipArea(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidLeaveTooltipArea(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidLeaveTooltipArea() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidLeaveTooltipArea; }
    static i32 static_message_id() { return (int)MessageID::DidLeaveTooltipArea; }
    virtual const char* message_name() const override { return "WebContentClient::DidLeaveTooltipArea"; }

    static ErrorOr<NonnullOwnPtr<DidLeaveTooltipArea>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        return make<DidLeaveTooltipArea>(move(page_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidLeaveTooltipArea));
        TRY(stream.encode(page_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id);
    }

    u64 page_id() const { return m_page_id; }

private:
    u64 m_page_id;
};

class DidHoverLink final : public IPC::Message {
public:
    DidHoverLink(DidHoverLink const&) = default;
    DidHoverLink(DidHoverLink&&) = default;
    DidHoverLink& operator=(DidHoverLink const&) = default;
    DidHoverLink(u64 page_id, URL::URL url) : m_page_id(move(page_id)), m_url(move(url)) {}

    virtual ~DidHoverLink() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidHoverLink; }
    static i32 static_message_id() { return (int)MessageID::DidHoverLink; }
    virtual const char* message_name() const override { return "WebContentClient::DidHoverLink"; }

    static ErrorOr<NonnullOwnPtr<DidHoverLink>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto url = TRY((decoder.decode<URL::URL>()));
        return make<DidHoverLink>(move(page_id), move(url));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, URL::URL const& url)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidHoverLink));
        TRY(stream.encode(page_id));
        TRY(stream.encode(url));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_url);
    }

    u64 page_id() const { return m_page_id; }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

private:
    u64 m_page_id;
    URL::URL m_url;
};

class DidUnhoverLink final : public IPC::Message {
public:
    DidUnhoverLink(DidUnhoverLink const&) = default;
    DidUnhoverLink(DidUnhoverLink&&) = default;
    DidUnhoverLink& operator=(DidUnhoverLink const&) = default;
    DidUnhoverLink(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidUnhoverLink(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidUnhoverLink() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidUnhoverLink; }
    static i32 static_message_id() { return (int)MessageID::DidUnhoverLink; }
    virtual const char* message_name() const override { return "WebContentClient::DidUnhoverLink"; }

    static ErrorOr<NonnullOwnPtr<DidUnhoverLink>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        return make<DidUnhoverLink>(move(page_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidUnhoverLink));
        TRY(stream.encode(page_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id);
    }

    u64 page_id() const { return m_page_id; }

private:
    u64 m_page_id;
};

class DidClickLink final : public IPC::Message {
public:
    DidClickLink(DidClickLink const&) = default;
    DidClickLink(DidClickLink&&) = default;
    DidClickLink& operator=(DidClickLink const&) = default;
    DidClickLink(u64 page_id, URL::URL url, ByteString target, unsigned modifiers) : m_page_id(move(page_id)), m_url(move(url)), m_target(move(target)), m_modifiers(move(modifiers)) {}

    virtual ~DidClickLink() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidClickLink; }
    static i32 static_message_id() { return (int)MessageID::DidClickLink; }
    virtual const char* message_name() const override { return "WebContentClient::DidClickLink"; }

    static ErrorOr<NonnullOwnPtr<DidClickLink>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto url = TRY((decoder.decode<URL::URL>()));
        auto target = TRY((decoder.decode<ByteString>()));
        auto modifiers = TRY((decoder.decode<unsigned>()));
        return make<DidClickLink>(move(page_id), move(url), move(target), move(modifiers));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, URL::URL const& url, StringView target, unsigned modifiers)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidClickLink));
        TRY(stream.encode(page_id));
        TRY(stream.encode(url));
        TRY(stream.encode(target));
        TRY(stream.encode(modifiers));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_url, m_target, m_modifiers);
    }

    u64 page_id() const { return m_page_id; }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const ByteString& target() const { return m_target; }
    ByteString take_target() { return move(m_target); }

    unsigned modifiers() const { return m_modifiers; }

private:
    u64 m_page_id;
    URL::URL m_url;
    ByteString m_target;
    unsigned m_modifiers;
};

class DidMiddleClickLink final : public IPC::Message {
public:
    DidMiddleClickLink(DidMiddleClickLink const&) = default;
    DidMiddleClickLink(DidMiddleClickLink&&) = default;
    DidMiddleClickLink& operator=(DidMiddleClickLink const&) = default;
    DidMiddleClickLink(u64 page_id, URL::URL url, ByteString target, unsigned modifiers) : m_page_id(move(page_id)), m_url(move(url)), m_target(move(target)), m_modifiers(move(modifiers)) {}

    virtual ~DidMiddleClickLink() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidMiddleClickLink; }
    static i32 static_message_id() { return (int)MessageID::DidMiddleClickLink; }
    virtual const char* message_name() const override { return "WebContentClient::DidMiddleClickLink"; }

    static ErrorOr<NonnullOwnPtr<DidMiddleClickLink>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto url = TRY((decoder.decode<URL::URL>()));
        auto target = TRY((decoder.decode<ByteString>()));
        auto modifiers = TRY((decoder.decode<unsigned>()));
        return make<DidMiddleClickLink>(move(page_id), move(url), move(target), move(modifiers));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, URL::URL const& url, StringView target, unsigned modifiers)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidMiddleClickLink));
        TRY(stream.encode(page_id));
        TRY(stream.encode(url));
        TRY(stream.encode(target));
        TRY(stream.encode(modifiers));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_url, m_target, m_modifiers);
    }

    u64 page_id() const { return m_page_id; }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const ByteString& target() const { return m_target; }
    ByteString take_target() { return move(m_target); }

    unsigned modifiers() const { return m_modifiers; }

private:
    u64 m_page_id;
    URL::URL m_url;
    ByteString m_target;
    unsigned m_modifiers;
};

class DidRequestContextMenu final : public IPC::Message {
public:
    DidRequestContextMenu(DidRequestContextMenu const&) = default;
    DidRequestContextMenu(DidRequestContextMenu&&) = default;
    DidRequestContextMenu& operator=(DidRequestContextMenu const&) = default;
    DidRequestContextMenu(u64 page_id, Gfx::IntPoint content_position) : m_page_id(move(page_id)), m_content_position(move(content_position)) {}

    virtual ~DidRequestContextMenu() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestContextMenu; }
    static i32 static_message_id() { return (int)MessageID::DidRequestContextMenu; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestContextMenu"; }

    static ErrorOr<NonnullOwnPtr<DidRequestContextMenu>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto content_position = TRY((decoder.decode<Gfx::IntPoint>()));
        return make<DidRequestContextMenu>(move(page_id), move(content_position));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Gfx::IntPoint content_position)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestContextMenu));
        TRY(stream.encode(page_id));
        TRY(stream.encode(content_position));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_content_position);
    }

    u64 page_id() const { return m_page_id; }

    Gfx::IntPoint content_position() const { return m_content_position; }

private:
    u64 m_page_id;
    Gfx::IntPoint m_content_position;
};

class DidRequestLinkContextMenu final : public IPC::Message {
public:
    DidRequestLinkContextMenu(DidRequestLinkContextMenu const&) = default;
    DidRequestLinkContextMenu(DidRequestLinkContextMenu&&) = default;
    DidRequestLinkContextMenu& operator=(DidRequestLinkContextMenu const&) = default;
    DidRequestLinkContextMenu(u64 page_id, Gfx::IntPoint content_position, URL::URL url, ByteString target, unsigned modifiers) : m_page_id(move(page_id)), m_content_position(move(content_position)), m_url(move(url)), m_target(move(target)), m_modifiers(move(modifiers)) {}

    virtual ~DidRequestLinkContextMenu() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestLinkContextMenu; }
    static i32 static_message_id() { return (int)MessageID::DidRequestLinkContextMenu; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestLinkContextMenu"; }

    static ErrorOr<NonnullOwnPtr<DidRequestLinkContextMenu>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto content_position = TRY((decoder.decode<Gfx::IntPoint>()));
        auto url = TRY((decoder.decode<URL::URL>()));
        auto target = TRY((decoder.decode<ByteString>()));
        auto modifiers = TRY((decoder.decode<unsigned>()));
        return make<DidRequestLinkContextMenu>(move(page_id), move(content_position), move(url), move(target), move(modifiers));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Gfx::IntPoint content_position, URL::URL const& url, StringView target, unsigned modifiers)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestLinkContextMenu));
        TRY(stream.encode(page_id));
        TRY(stream.encode(content_position));
        TRY(stream.encode(url));
        TRY(stream.encode(target));
        TRY(stream.encode(modifiers));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_content_position, m_url, m_target, m_modifiers);
    }

    u64 page_id() const { return m_page_id; }

    Gfx::IntPoint content_position() const { return m_content_position; }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const ByteString& target() const { return m_target; }
    ByteString take_target() { return move(m_target); }

    unsigned modifiers() const { return m_modifiers; }

private:
    u64 m_page_id;
    Gfx::IntPoint m_content_position;
    URL::URL m_url;
    ByteString m_target;
    unsigned m_modifiers;
};

class DidRequestImageContextMenu final : public IPC::Message {
public:
    DidRequestImageContextMenu(DidRequestImageContextMenu const&) = default;
    DidRequestImageContextMenu(DidRequestImageContextMenu&&) = default;
    DidRequestImageContextMenu& operator=(DidRequestImageContextMenu const&) = default;
    DidRequestImageContextMenu(u64 page_id, Gfx::IntPoint content_position, URL::URL url, ByteString target, unsigned modifiers, Optional<Gfx::ShareableBitmap> bitmap) : m_page_id(move(page_id)), m_content_position(move(content_position)), m_url(move(url)), m_target(move(target)), m_modifiers(move(modifiers)), m_bitmap(move(bitmap)) {}

    virtual ~DidRequestImageContextMenu() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestImageContextMenu; }
    static i32 static_message_id() { return (int)MessageID::DidRequestImageContextMenu; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestImageContextMenu"; }

    static ErrorOr<NonnullOwnPtr<DidRequestImageContextMenu>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto content_position = TRY((decoder.decode<Gfx::IntPoint>()));
        auto url = TRY((decoder.decode<URL::URL>()));
        auto target = TRY((decoder.decode<ByteString>()));
        auto modifiers = TRY((decoder.decode<unsigned>()));
        auto bitmap = TRY((decoder.decode<Optional<Gfx::ShareableBitmap>>()));
        return make<DidRequestImageContextMenu>(move(page_id), move(content_position), move(url), move(target), move(modifiers), move(bitmap));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Gfx::IntPoint content_position, URL::URL const& url, StringView target, unsigned modifiers, Optional<Gfx::ShareableBitmap> const& bitmap)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestImageContextMenu));
        TRY(stream.encode(page_id));
        TRY(stream.encode(content_position));
        TRY(stream.encode(url));
        TRY(stream.encode(target));
        TRY(stream.encode(modifiers));
        TRY(stream.encode(bitmap));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_content_position, m_url, m_target, m_modifiers, m_bitmap);
    }

    u64 page_id() const { return m_page_id; }

    Gfx::IntPoint content_position() const { return m_content_position; }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const ByteString& target() const { return m_target; }
    ByteString take_target() { return move(m_target); }

    unsigned modifiers() const { return m_modifiers; }

    const Optional<Gfx::ShareableBitmap>& bitmap() const { return m_bitmap; }
    Optional<Gfx::ShareableBitmap> take_bitmap() { return move(m_bitmap); }

private:
    u64 m_page_id;
    Gfx::IntPoint m_content_position;
    URL::URL m_url;
    ByteString m_target;
    unsigned m_modifiers;
    Optional<Gfx::ShareableBitmap> m_bitmap;
};

class DidRequestMediaContextMenu final : public IPC::Message {
public:
    DidRequestMediaContextMenu(DidRequestMediaContextMenu const&) = default;
    DidRequestMediaContextMenu(DidRequestMediaContextMenu&&) = default;
    DidRequestMediaContextMenu& operator=(DidRequestMediaContextMenu const&) = default;
    DidRequestMediaContextMenu(u64 page_id, Gfx::IntPoint content_position, ByteString target, unsigned modifiers, Web::Page::MediaContextMenu menu) : m_page_id(move(page_id)), m_content_position(move(content_position)), m_target(move(target)), m_modifiers(move(modifiers)), m_menu(move(menu)) {}

    virtual ~DidRequestMediaContextMenu() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestMediaContextMenu; }
    static i32 static_message_id() { return (int)MessageID::DidRequestMediaContextMenu; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestMediaContextMenu"; }

    static ErrorOr<NonnullOwnPtr<DidRequestMediaContextMenu>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto content_position = TRY((decoder.decode<Gfx::IntPoint>()));
        auto target = TRY((decoder.decode<ByteString>()));
        auto modifiers = TRY((decoder.decode<unsigned>()));
        auto menu = TRY((decoder.decode<Web::Page::MediaContextMenu>()));
        return make<DidRequestMediaContextMenu>(move(page_id), move(content_position), move(target), move(modifiers), move(menu));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Gfx::IntPoint content_position, StringView target, unsigned modifiers, Web::Page::MediaContextMenu const& menu)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestMediaContextMenu));
        TRY(stream.encode(page_id));
        TRY(stream.encode(content_position));
        TRY(stream.encode(target));
        TRY(stream.encode(modifiers));
        TRY(stream.encode(menu));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_content_position, m_target, m_modifiers, m_menu);
    }

    u64 page_id() const { return m_page_id; }

    Gfx::IntPoint content_position() const { return m_content_position; }

    const ByteString& target() const { return m_target; }
    ByteString take_target() { return move(m_target); }

    unsigned modifiers() const { return m_modifiers; }

    const Web::Page::MediaContextMenu& menu() const { return m_menu; }
    Web::Page::MediaContextMenu take_menu() { return move(m_menu); }

private:
    u64 m_page_id;
    Gfx::IntPoint m_content_position;
    ByteString m_target;
    unsigned m_modifiers;
    Web::Page::MediaContextMenu m_menu;
};

class DidRequestAlert final : public IPC::Message {
public:
    DidRequestAlert(DidRequestAlert const&) = default;
    DidRequestAlert(DidRequestAlert&&) = default;
    DidRequestAlert& operator=(DidRequestAlert const&) = default;
    DidRequestAlert(u64 page_id, String message) : m_page_id(move(page_id)), m_message(move(message)) {}

    virtual ~DidRequestAlert() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestAlert; }
    static i32 static_message_id() { return (int)MessageID::DidRequestAlert; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestAlert"; }

    static ErrorOr<NonnullOwnPtr<DidRequestAlert>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto message = TRY((decoder.decode<String>()));
        return make<DidRequestAlert>(move(page_id), move(message));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, StringView message)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestAlert));
        TRY(stream.encode(page_id));
        TRY(stream.encode(message));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_message);
    }

    u64 page_id() const { return m_page_id; }

    const String& message() const { return m_message; }
    String take_message() { return move(m_message); }

private:
    u64 m_page_id;
    String m_message;
};

class DidRequestConfirm final : public IPC::Message {
public:
    DidRequestConfirm(DidRequestConfirm const&) = default;
    DidRequestConfirm(DidRequestConfirm&&) = default;
    DidRequestConfirm& operator=(DidRequestConfirm const&) = default;
    DidRequestConfirm(u64 page_id, String message) : m_page_id(move(page_id)), m_message(move(message)) {}

    virtual ~DidRequestConfirm() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestConfirm; }
    static i32 static_message_id() { return (int)MessageID::DidRequestConfirm; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestConfirm"; }

    static ErrorOr<NonnullOwnPtr<DidRequestConfirm>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto message = TRY((decoder.decode<String>()));
        return make<DidRequestConfirm>(move(page_id), move(message));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, StringView message)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestConfirm));
        TRY(stream.encode(page_id));
        TRY(stream.encode(message));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_message);
    }

    u64 page_id() const { return m_page_id; }

    const String& message() const { return m_message; }
    String take_message() { return move(m_message); }

private:
    u64 m_page_id;
    String m_message;
};

class DidRequestPrompt final : public IPC::Message {
public:
    DidRequestPrompt(DidRequestPrompt const&) = default;
    DidRequestPrompt(DidRequestPrompt&&) = default;
    DidRequestPrompt& operator=(DidRequestPrompt const&) = default;
    DidRequestPrompt(u64 page_id, String message, String default_) : m_page_id(move(page_id)), m_message(move(message)), m_default_(move(default_)) {}

    virtual ~DidRequestPrompt() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestPrompt; }
    static i32 static_message_id() { return (int)MessageID::DidRequestPrompt; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestPrompt"; }

    static ErrorOr<NonnullOwnPtr<DidRequestPrompt>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto message = TRY((decoder.decode<String>()));
        auto default_ = TRY((decoder.decode<String>()));
        return make<DidRequestPrompt>(move(page_id), move(message), move(default_));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, StringView message, StringView default_)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestPrompt));
        TRY(stream.encode(page_id));
        TRY(stream.encode(message));
        TRY(stream.encode(default_));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_message, m_default_);
    }

    u64 page_id() const { return m_page_id; }

    const String& message() const { return m_message; }
    String take_message() { return move(m_message); }

    const String& default_() const { return m_default_; }
    String take_default_() { return move(m_default_); }

private:
    u64 m_page_id;
    String m_message;
    String m_default_;
};

class DidRequestSetPromptText final : public IPC::Message {
public:
    DidRequestSetPromptText(DidRequestSetPromptText const&) = default;
    DidRequestSetPromptText(DidRequestSetPromptText&&) = default;
    DidRequestSetPromptText& operator=(DidRequestSetPromptText const&) = default;
    DidRequestSetPromptText(u64 page_id, String message) : m_page_id(move(page_id)), m_message(move(message)) {}

    virtual ~DidRequestSetPromptText() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestSetPromptText; }
    static i32 static_message_id() { return (int)MessageID::DidRequestSetPromptText; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestSetPromptText"; }

    static ErrorOr<NonnullOwnPtr<DidRequestSetPromptText>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto message = TRY((decoder.decode<String>()));
        return make<DidRequestSetPromptText>(move(page_id), move(message));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, StringView message)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestSetPromptText));
        TRY(stream.encode(page_id));
        TRY(stream.encode(message));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_message);
    }

    u64 page_id() const { return m_page_id; }

    const String& message() const { return m_message; }
    String take_message() { return move(m_message); }

private:
    u64 m_page_id;
    String m_message;
};

class DidRequestAcceptDialog final : public IPC::Message {
public:
    DidRequestAcceptDialog(DidRequestAcceptDialog const&) = default;
    DidRequestAcceptDialog(DidRequestAcceptDialog&&) = default;
    DidRequestAcceptDialog& operator=(DidRequestAcceptDialog const&) = default;
    DidRequestAcceptDialog(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestAcceptDialog(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestAcceptDialog() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestAcceptDialog; }
    static i32 static_message_id() { return (int)MessageID::DidRequestAcceptDialog; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestAcceptDialog"; }

    static ErrorOr<NonnullOwnPtr<DidRequestAcceptDialog>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        return make<DidRequestAcceptDialog>(move(page_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestAcceptDialog));
        TRY(stream.encode(page_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id);
    }

    u64 page_id() const { return m_page_id; }

private:
    u64 m_page_id;
};

class DidRequestDismissDialog final : public IPC::Message {
public:
    DidRequestDismissDialog(DidRequestDismissDialog const&) = default;
    DidRequestDismissDialog(DidRequestDismissDialog&&) = default;
    DidRequestDismissDialog& operator=(DidRequestDismissDialog const&) = default;
    DidRequestDismissDialog(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestDismissDialog(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestDismissDialog() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestDismissDialog; }
    static i32 static_message_id() { return (int)MessageID::DidRequestDismissDialog; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestDismissDialog"; }

    static ErrorOr<NonnullOwnPtr<DidRequestDismissDialog>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        return make<DidRequestDismissDialog>(move(page_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestDismissDialog));
        TRY(stream.encode(page_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id);
    }

    u64 page_id() const { return m_page_id; }

private:
    u64 m_page_id;
};

class DidGetSource final : public IPC::Message {
public:
    DidGetSource(DidGetSource const&) = default;
    DidGetSource(DidGetSource&&) = default;
    DidGetSource& operator=(DidGetSource const&) = default;
    DidGetSource(u64 page_id, URL::URL url, URL::URL base_url, String source) : m_page_id(move(page_id)), m_url(move(url)), m_base_url(move(base_url)), m_source(move(source)) {}

    virtual ~DidGetSource() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidGetSource; }
    static i32 static_message_id() { return (int)MessageID::DidGetSource; }
    virtual const char* message_name() const override { return "WebContentClient::DidGetSource"; }

    static ErrorOr<NonnullOwnPtr<DidGetSource>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto url = TRY((decoder.decode<URL::URL>()));
        auto base_url = TRY((decoder.decode<URL::URL>()));
        auto source = TRY((decoder.decode<String>()));
        return make<DidGetSource>(move(page_id), move(url), move(base_url), move(source));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, URL::URL const& url, URL::URL const& base_url, StringView source)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidGetSource));
        TRY(stream.encode(page_id));
        TRY(stream.encode(url));
        TRY(stream.encode(base_url));
        TRY(stream.encode(source));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_url, m_base_url, m_source);
    }

    u64 page_id() const { return m_page_id; }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const URL::URL& base_url() const { return m_base_url; }
    URL::URL take_base_url() { return move(m_base_url); }

    const String& source() const { return m_source; }
    String take_source() { return move(m_source); }

private:
    u64 m_page_id;
    URL::URL m_url;
    URL::URL m_base_url;
    String m_source;
};

class DidInspectDomTree final : public IPC::Message {
public:
    DidInspectDomTree(DidInspectDomTree const&) = default;
    DidInspectDomTree(DidInspectDomTree&&) = default;
    DidInspectDomTree& operator=(DidInspectDomTree const&) = default;
    DidInspectDomTree(u64 page_id, String dom_tree) : m_page_id(move(page_id)), m_dom_tree(move(dom_tree)) {}

    virtual ~DidInspectDomTree() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidInspectDomTree; }
    static i32 static_message_id() { return (int)MessageID::DidInspectDomTree; }
    virtual const char* message_name() const override { return "WebContentClient::DidInspectDomTree"; }

    static ErrorOr<NonnullOwnPtr<DidInspectDomTree>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto dom_tree = TRY((decoder.decode<String>()));
        return make<DidInspectDomTree>(move(page_id), move(dom_tree));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, StringView dom_tree)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidInspectDomTree));
        TRY(stream.encode(page_id));
        TRY(stream.encode(dom_tree));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_dom_tree);
    }

    u64 page_id() const { return m_page_id; }

    const String& dom_tree() const { return m_dom_tree; }
    String take_dom_tree() { return move(m_dom_tree); }

private:
    u64 m_page_id;
    String m_dom_tree;
};

class DidInspectDomNode final : public IPC::Message {
public:
    DidInspectDomNode(DidInspectDomNode const&) = default;
    DidInspectDomNode(DidInspectDomNode&&) = default;
    DidInspectDomNode& operator=(DidInspectDomNode const&) = default;
    DidInspectDomNode(u64 page_id, WebView::DOMNodeProperties properties) : m_page_id(move(page_id)), m_properties(move(properties)) {}

    virtual ~DidInspectDomNode() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidInspectDomNode; }
    static i32 static_message_id() { return (int)MessageID::DidInspectDomNode; }
    virtual const char* message_name() const override { return "WebContentClient::DidInspectDomNode"; }

    static ErrorOr<NonnullOwnPtr<DidInspectDomNode>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto properties = TRY((decoder.decode<WebView::DOMNodeProperties>()));
        return make<DidInspectDomNode>(move(page_id), move(properties));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, WebView::DOMNodeProperties const& properties)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidInspectDomNode));
        TRY(stream.encode(page_id));
        TRY(stream.encode(properties));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_properties);
    }

    u64 page_id() const { return m_page_id; }

    const WebView::DOMNodeProperties& properties() const { return m_properties; }
    WebView::DOMNodeProperties take_properties() { return move(m_properties); }

private:
    u64 m_page_id;
    WebView::DOMNodeProperties m_properties;
};

class DidInspectAccessibilityTree final : public IPC::Message {
public:
    DidInspectAccessibilityTree(DidInspectAccessibilityTree const&) = default;
    DidInspectAccessibilityTree(DidInspectAccessibilityTree&&) = default;
    DidInspectAccessibilityTree& operator=(DidInspectAccessibilityTree const&) = default;
    DidInspectAccessibilityTree(u64 page_id, String accessibility_tree) : m_page_id(move(page_id)), m_accessibility_tree(move(accessibility_tree)) {}

    virtual ~DidInspectAccessibilityTree() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidInspectAccessibilityTree; }
    static i32 static_message_id() { return (int)MessageID::DidInspectAccessibilityTree; }
    virtual const char* message_name() const override { return "WebContentClient::DidInspectAccessibilityTree"; }

    static ErrorOr<NonnullOwnPtr<DidInspectAccessibilityTree>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto accessibility_tree = TRY((decoder.decode<String>()));
        return make<DidInspectAccessibilityTree>(move(page_id), move(accessibility_tree));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, StringView accessibility_tree)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidInspectAccessibilityTree));
        TRY(stream.encode(page_id));
        TRY(stream.encode(accessibility_tree));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_accessibility_tree);
    }

    u64 page_id() const { return m_page_id; }

    const String& accessibility_tree() const { return m_accessibility_tree; }
    String take_accessibility_tree() { return move(m_accessibility_tree); }

private:
    u64 m_page_id;
    String m_accessibility_tree;
};

class DidGetHoveredNodeId final : public IPC::Message {
public:
    DidGetHoveredNodeId(DidGetHoveredNodeId const&) = default;
    DidGetHoveredNodeId(DidGetHoveredNodeId&&) = default;
    DidGetHoveredNodeId& operator=(DidGetHoveredNodeId const&) = default;
    DidGetHoveredNodeId(u64 page_id, Web::UniqueNodeID node_id) : m_page_id(move(page_id)), m_node_id(move(node_id)) {}

    virtual ~DidGetHoveredNodeId() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidGetHoveredNodeId; }
    static i32 static_message_id() { return (int)MessageID::DidGetHoveredNodeId; }
    virtual const char* message_name() const override { return "WebContentClient::DidGetHoveredNodeId"; }

    static ErrorOr<NonnullOwnPtr<DidGetHoveredNodeId>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));
        return make<DidGetHoveredNodeId>(move(page_id), move(node_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Web::UniqueNodeID const& node_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidGetHoveredNodeId));
        TRY(stream.encode(page_id));
        TRY(stream.encode(node_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_node_id);
    }

    u64 page_id() const { return m_page_id; }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

private:
    u64 m_page_id;
    Web::UniqueNodeID m_node_id;
};

class DidFinishEditingDomNode final : public IPC::Message {
public:
    DidFinishEditingDomNode(DidFinishEditingDomNode const&) = default;
    DidFinishEditingDomNode(DidFinishEditingDomNode&&) = default;
    DidFinishEditingDomNode& operator=(DidFinishEditingDomNode const&) = default;
    DidFinishEditingDomNode(u64 page_id, Optional<Web::UniqueNodeID> node_id) : m_page_id(move(page_id)), m_node_id(move(node_id)) {}

    virtual ~DidFinishEditingDomNode() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidFinishEditingDomNode; }
    static i32 static_message_id() { return (int)MessageID::DidFinishEditingDomNode; }
    virtual const char* message_name() const override { return "WebContentClient::DidFinishEditingDomNode"; }

    static ErrorOr<NonnullOwnPtr<DidFinishEditingDomNode>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto node_id = TRY((decoder.decode<Optional<Web::UniqueNodeID>>()));
        return make<DidFinishEditingDomNode>(move(page_id), move(node_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Optional<Web::UniqueNodeID> const& node_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidFinishEditingDomNode));
        TRY(stream.encode(page_id));
        TRY(stream.encode(node_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_node_id);
    }

    u64 page_id() const { return m_page_id; }

    const Optional<Web::UniqueNodeID>& node_id() const { return m_node_id; }
    Optional<Web::UniqueNodeID> take_node_id() { return move(m_node_id); }

private:
    u64 m_page_id;
    Optional<Web::UniqueNodeID> m_node_id;
};

class DidMutateDom final : public IPC::Message {
public:
    DidMutateDom(DidMutateDom const&) = default;
    DidMutateDom(DidMutateDom&&) = default;
    DidMutateDom& operator=(DidMutateDom const&) = default;
    DidMutateDom(u64 page_id, WebView::Mutation mutation) : m_page_id(move(page_id)), m_mutation(move(mutation)) {}

    virtual ~DidMutateDom() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidMutateDom; }
    static i32 static_message_id() { return (int)MessageID::DidMutateDom; }
    virtual const char* message_name() const override { return "WebContentClient::DidMutateDom"; }

    static ErrorOr<NonnullOwnPtr<DidMutateDom>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto mutation = TRY((decoder.decode<WebView::Mutation>()));
        return make<DidMutateDom>(move(page_id), move(mutation));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, WebView::Mutation const& mutation)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidMutateDom));
        TRY(stream.encode(page_id));
        TRY(stream.encode(mutation));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_mutation);
    }

    u64 page_id() const { return m_page_id; }

    const WebView::Mutation& mutation() const { return m_mutation; }
    WebView::Mutation take_mutation() { return move(m_mutation); }

private:
    u64 m_page_id;
    WebView::Mutation m_mutation;
};

class DidGetDomNodeHtml final : public IPC::Message {
public:
    DidGetDomNodeHtml(DidGetDomNodeHtml const&) = default;
    DidGetDomNodeHtml(DidGetDomNodeHtml&&) = default;
    DidGetDomNodeHtml& operator=(DidGetDomNodeHtml const&) = default;
    DidGetDomNodeHtml(u64 page_id, String html) : m_page_id(move(page_id)), m_html(move(html)) {}

    virtual ~DidGetDomNodeHtml() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidGetDomNodeHtml; }
    static i32 static_message_id() { return (int)MessageID::DidGetDomNodeHtml; }
    virtual const char* message_name() const override { return "WebContentClient::DidGetDomNodeHtml"; }

    static ErrorOr<NonnullOwnPtr<DidGetDomNodeHtml>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto html = TRY((decoder.decode<String>()));
        return make<DidGetDomNodeHtml>(move(page_id), move(html));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, StringView html)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidGetDomNodeHtml));
        TRY(stream.encode(page_id));
        TRY(stream.encode(html));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_html);
    }

    u64 page_id() const { return m_page_id; }

    const String& html() const { return m_html; }
    String take_html() { return move(m_html); }

private:
    u64 m_page_id;
    String m_html;
};

class DidListStyleSheets final : public IPC::Message {
public:
    DidListStyleSheets(DidListStyleSheets const&) = default;
    DidListStyleSheets(DidListStyleSheets&&) = default;
    DidListStyleSheets& operator=(DidListStyleSheets const&) = default;
    DidListStyleSheets(u64 page_id, Vector<Web::CSS::StyleSheetIdentifier> style_sheets) : m_page_id(move(page_id)), m_style_sheets(move(style_sheets)) {}

    virtual ~DidListStyleSheets() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidListStyleSheets; }
    static i32 static_message_id() { return (int)MessageID::DidListStyleSheets; }
    virtual const char* message_name() const override { return "WebContentClient::DidListStyleSheets"; }

    static ErrorOr<NonnullOwnPtr<DidListStyleSheets>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto style_sheets = TRY((decoder.decode<Vector<Web::CSS::StyleSheetIdentifier>>()));
        return make<DidListStyleSheets>(move(page_id), move(style_sheets));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, ReadonlySpan<Web::CSS::StyleSheetIdentifier> style_sheets)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidListStyleSheets));
        TRY(stream.encode(page_id));
        TRY(stream.encode(style_sheets));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_style_sheets);
    }

    u64 page_id() const { return m_page_id; }

    const Vector<Web::CSS::StyleSheetIdentifier>& style_sheets() const { return m_style_sheets; }
    Vector<Web::CSS::StyleSheetIdentifier> take_style_sheets() { return move(m_style_sheets); }

private:
    u64 m_page_id;
    Vector<Web::CSS::StyleSheetIdentifier> m_style_sheets;
};

class DidGetStyleSheetSource final : public IPC::Message {
public:
    DidGetStyleSheetSource(DidGetStyleSheetSource const&) = default;
    DidGetStyleSheetSource(DidGetStyleSheetSource&&) = default;
    DidGetStyleSheetSource& operator=(DidGetStyleSheetSource const&) = default;
    DidGetStyleSheetSource(u64 page_id, Web::CSS::StyleSheetIdentifier identifier, URL::URL base_url, String source) : m_page_id(move(page_id)), m_identifier(move(identifier)), m_base_url(move(base_url)), m_source(move(source)) {}

    virtual ~DidGetStyleSheetSource() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidGetStyleSheetSource; }
    static i32 static_message_id() { return (int)MessageID::DidGetStyleSheetSource; }
    virtual const char* message_name() const override { return "WebContentClient::DidGetStyleSheetSource"; }

    static ErrorOr<NonnullOwnPtr<DidGetStyleSheetSource>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto identifier = TRY((decoder.decode<Web::CSS::StyleSheetIdentifier>()));
        auto base_url = TRY((decoder.decode<URL::URL>()));
        auto source = TRY((decoder.decode<String>()));
        return make<DidGetStyleSheetSource>(move(page_id), move(identifier), move(base_url), move(source));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Web::CSS::StyleSheetIdentifier const& identifier, URL::URL const& base_url, StringView source)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidGetStyleSheetSource));
        TRY(stream.encode(page_id));
        TRY(stream.encode(identifier));
        TRY(stream.encode(base_url));
        TRY(stream.encode(source));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_identifier, m_base_url, m_source);
    }

    u64 page_id() const { return m_page_id; }

    const Web::CSS::StyleSheetIdentifier& identifier() const { return m_identifier; }
    Web::CSS::StyleSheetIdentifier take_identifier() { return move(m_identifier); }

    const URL::URL& base_url() const { return m_base_url; }
    URL::URL take_base_url() { return move(m_base_url); }

    const String& source() const { return m_source; }
    String take_source() { return move(m_source); }

private:
    u64 m_page_id;
    Web::CSS::StyleSheetIdentifier m_identifier;
    URL::URL m_base_url;
    String m_source;
};

class DidTakeScreenshot final : public IPC::Message {
public:
    DidTakeScreenshot(DidTakeScreenshot const&) = default;
    DidTakeScreenshot(DidTakeScreenshot&&) = default;
    DidTakeScreenshot& operator=(DidTakeScreenshot const&) = default;
    DidTakeScreenshot(u64 page_id, Gfx::ShareableBitmap screenshot) : m_page_id(move(page_id)), m_screenshot(move(screenshot)) {}

    virtual ~DidTakeScreenshot() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidTakeScreenshot; }
    static i32 static_message_id() { return (int)MessageID::DidTakeScreenshot; }
    virtual const char* message_name() const override { return "WebContentClient::DidTakeScreenshot"; }

    static ErrorOr<NonnullOwnPtr<DidTakeScreenshot>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto screenshot = TRY((decoder.decode<Gfx::ShareableBitmap>()));
        return make<DidTakeScreenshot>(move(page_id), move(screenshot));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Gfx::ShareableBitmap const& screenshot)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidTakeScreenshot));
        TRY(stream.encode(page_id));
        TRY(stream.encode(screenshot));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_screenshot);
    }

    u64 page_id() const { return m_page_id; }

    const Gfx::ShareableBitmap& screenshot() const { return m_screenshot; }
    Gfx::ShareableBitmap take_screenshot() { return move(m_screenshot); }

private:
    u64 m_page_id;
    Gfx::ShareableBitmap m_screenshot;
};

class DidGetInternalPageInfo final : public IPC::Message {
public:
    DidGetInternalPageInfo(DidGetInternalPageInfo const&) = default;
    DidGetInternalPageInfo(DidGetInternalPageInfo&&) = default;
    DidGetInternalPageInfo& operator=(DidGetInternalPageInfo const&) = default;
    DidGetInternalPageInfo(u64 page_id, WebView::PageInfoType type, String info) : m_page_id(move(page_id)), m_type(move(type)), m_info(move(info)) {}

    virtual ~DidGetInternalPageInfo() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidGetInternalPageInfo; }
    static i32 static_message_id() { return (int)MessageID::DidGetInternalPageInfo; }
    virtual const char* message_name() const override { return "WebContentClient::DidGetInternalPageInfo"; }

    static ErrorOr<NonnullOwnPtr<DidGetInternalPageInfo>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto type = TRY((decoder.decode<WebView::PageInfoType>()));
        auto info = TRY((decoder.decode<String>()));
        return make<DidGetInternalPageInfo>(move(page_id), move(type), move(info));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, WebView::PageInfoType type, StringView info)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidGetInternalPageInfo));
        TRY(stream.encode(page_id));
        TRY(stream.encode(type));
        TRY(stream.encode(info));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_type, m_info);
    }

    u64 page_id() const { return m_page_id; }

    WebView::PageInfoType type() const { return m_type; }

    const String& info() const { return m_info; }
    String take_info() { return move(m_info); }

private:
    u64 m_page_id;
    WebView::PageInfoType m_type;
    String m_info;
};

class DidChangeFavicon final : public IPC::Message {
public:
    DidChangeFavicon(DidChangeFavicon const&) = default;
    DidChangeFavicon(DidChangeFavicon&&) = default;
    DidChangeFavicon& operator=(DidChangeFavicon const&) = default;
    DidChangeFavicon(u64 page_id, Gfx::ShareableBitmap favicon) : m_page_id(move(page_id)), m_favicon(move(favicon)) {}

    virtual ~DidChangeFavicon() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidChangeFavicon; }
    static i32 static_message_id() { return (int)MessageID::DidChangeFavicon; }
    virtual const char* message_name() const override { return "WebContentClient::DidChangeFavicon"; }

    static ErrorOr<NonnullOwnPtr<DidChangeFavicon>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto favicon = TRY((decoder.decode<Gfx::ShareableBitmap>()));
        return make<DidChangeFavicon>(move(page_id), move(favicon));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Gfx::ShareableBitmap const& favicon)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidChangeFavicon));
        TRY(stream.encode(page_id));
        TRY(stream.encode(favicon));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_favicon);
    }

    u64 page_id() const { return m_page_id; }

    const Gfx::ShareableBitmap& favicon() const { return m_favicon; }
    Gfx::ShareableBitmap take_favicon() { return move(m_favicon); }

private:
    u64 m_page_id;
    Gfx::ShareableBitmap m_favicon;
};

class DidRequestAllCookiesResponse final : public IPC::Message {
public:
    DidRequestAllCookiesResponse(DidRequestAllCookiesResponse const&) = default;
    DidRequestAllCookiesResponse(DidRequestAllCookiesResponse&&) = default;
    DidRequestAllCookiesResponse& operator=(DidRequestAllCookiesResponse const&) = default;
    DidRequestAllCookiesResponse(Vector<Web::Cookie::Cookie> cookies) : m_cookies(move(cookies)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, Vector<Web::Cookie::Cookie>>)
    DidRequestAllCookiesResponse(WrappedReturnType&& value)
        : m_cookies(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestAllCookiesResponse() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestAllCookiesResponse; }
    static i32 static_message_id() { return (int)MessageID::DidRequestAllCookiesResponse; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestAllCookiesResponse"; }

    static ErrorOr<NonnullOwnPtr<DidRequestAllCookiesResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto cookies = TRY((decoder.decode<Vector<Web::Cookie::Cookie>>()));
        return make<DidRequestAllCookiesResponse>(move(cookies));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(ReadonlySpan<Web::Cookie::Cookie> cookies)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestAllCookiesResponse));
        TRY(stream.encode(cookies));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_cookies);
    }

    const Vector<Web::Cookie::Cookie>& cookies() const { return m_cookies; }
    Vector<Web::Cookie::Cookie> take_cookies() { return move(m_cookies); }

private:
    Vector<Web::Cookie::Cookie> m_cookies;
};

class DidRequestAllCookies final : public IPC::Message {
public:
   typedef class DidRequestAllCookiesResponse ResponseType;

    DidRequestAllCookies(DidRequestAllCookies const&) = default;
    DidRequestAllCookies(DidRequestAllCookies&&) = default;
    DidRequestAllCookies& operator=(DidRequestAllCookies const&) = default;
    DidRequestAllCookies(URL::URL url) : m_url(move(url)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, URL::URL>)
    DidRequestAllCookies(WrappedReturnType&& value)
        : m_url(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestAllCookies() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestAllCookies; }
    static i32 static_message_id() { return (int)MessageID::DidRequestAllCookies; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestAllCookies"; }

    static ErrorOr<NonnullOwnPtr<DidRequestAllCookies>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto url = TRY((decoder.decode<URL::URL>()));
        return make<DidRequestAllCookies>(move(url));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(URL::URL const& url)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestAllCookies));
        TRY(stream.encode(url));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_url);
    }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

private:
    URL::URL m_url;
};

class DidRequestNamedCookieResponse final : public IPC::Message {
public:
    DidRequestNamedCookieResponse(DidRequestNamedCookieResponse const&) = default;
    DidRequestNamedCookieResponse(DidRequestNamedCookieResponse&&) = default;
    DidRequestNamedCookieResponse& operator=(DidRequestNamedCookieResponse const&) = default;
    DidRequestNamedCookieResponse(Optional<Web::Cookie::Cookie> cookie) : m_cookie(move(cookie)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, Optional<Web::Cookie::Cookie>>)
    DidRequestNamedCookieResponse(WrappedReturnType&& value)
        : m_cookie(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestNamedCookieResponse() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestNamedCookieResponse; }
    static i32 static_message_id() { return (int)MessageID::DidRequestNamedCookieResponse; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestNamedCookieResponse"; }

    static ErrorOr<NonnullOwnPtr<DidRequestNamedCookieResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto cookie = TRY((decoder.decode<Optional<Web::Cookie::Cookie>>()));
        return make<DidRequestNamedCookieResponse>(move(cookie));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(Optional<Web::Cookie::Cookie> const& cookie)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestNamedCookieResponse));
        TRY(stream.encode(cookie));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_cookie);
    }

    const Optional<Web::Cookie::Cookie>& cookie() const { return m_cookie; }
    Optional<Web::Cookie::Cookie> take_cookie() { return move(m_cookie); }

private:
    Optional<Web::Cookie::Cookie> m_cookie;
};

class DidRequestNamedCookie final : public IPC::Message {
public:
   typedef class DidRequestNamedCookieResponse ResponseType;

    DidRequestNamedCookie(DidRequestNamedCookie const&) = default;
    DidRequestNamedCookie(DidRequestNamedCookie&&) = default;
    DidRequestNamedCookie& operator=(DidRequestNamedCookie const&) = default;
    DidRequestNamedCookie(URL::URL url, String name) : m_url(move(url)), m_name(move(name)) {}

    virtual ~DidRequestNamedCookie() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestNamedCookie; }
    static i32 static_message_id() { return (int)MessageID::DidRequestNamedCookie; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestNamedCookie"; }

    static ErrorOr<NonnullOwnPtr<DidRequestNamedCookie>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto url = TRY((decoder.decode<URL::URL>()));
        auto name = TRY((decoder.decode<String>()));
        return make<DidRequestNamedCookie>(move(url), move(name));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(URL::URL const& url, StringView name)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestNamedCookie));
        TRY(stream.encode(url));
        TRY(stream.encode(name));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_url, m_name);
    }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const String& name() const { return m_name; }
    String take_name() { return move(m_name); }

private:
    URL::URL m_url;
    String m_name;
};

class DidRequestCookieResponse final : public IPC::Message {
public:
    DidRequestCookieResponse(DidRequestCookieResponse const&) = default;
    DidRequestCookieResponse(DidRequestCookieResponse&&) = default;
    DidRequestCookieResponse& operator=(DidRequestCookieResponse const&) = default;
    DidRequestCookieResponse(String cookie) : m_cookie(move(cookie)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, String>)
    DidRequestCookieResponse(WrappedReturnType&& value)
        : m_cookie(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestCookieResponse() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestCookieResponse; }
    static i32 static_message_id() { return (int)MessageID::DidRequestCookieResponse; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestCookieResponse"; }

    static ErrorOr<NonnullOwnPtr<DidRequestCookieResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto cookie = TRY((decoder.decode<String>()));
        return make<DidRequestCookieResponse>(move(cookie));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(StringView cookie)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestCookieResponse));
        TRY(stream.encode(cookie));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_cookie);
    }

    const String& cookie() const { return m_cookie; }
    String take_cookie() { return move(m_cookie); }

private:
    String m_cookie;
};

class DidRequestCookie final : public IPC::Message {
public:
   typedef class DidRequestCookieResponse ResponseType;

    DidRequestCookie(DidRequestCookie const&) = default;
    DidRequestCookie(DidRequestCookie&&) = default;
    DidRequestCookie& operator=(DidRequestCookie const&) = default;
    DidRequestCookie(URL::URL url, Web::Cookie::Source source) : m_url(move(url)), m_source(move(source)) {}

    virtual ~DidRequestCookie() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestCookie; }
    static i32 static_message_id() { return (int)MessageID::DidRequestCookie; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestCookie"; }

    static ErrorOr<NonnullOwnPtr<DidRequestCookie>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto url = TRY((decoder.decode<URL::URL>()));
        auto source = TRY((decoder.decode<Web::Cookie::Source>()));
        return make<DidRequestCookie>(move(url), move(source));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(URL::URL const& url, Web::Cookie::Source source)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestCookie));
        TRY(stream.encode(url));
        TRY(stream.encode(source));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_url, m_source);
    }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    Web::Cookie::Source source() const { return m_source; }

private:
    URL::URL m_url;
    Web::Cookie::Source m_source;
};

class DidSetCookieResponse final : public IPC::Message {
public:
    DidSetCookieResponse(DidSetCookieResponse const&) = default;
    DidSetCookieResponse(DidSetCookieResponse&&) = default;
    DidSetCookieResponse& operator=(DidSetCookieResponse const&) = default;
    DidSetCookieResponse() {}

    virtual ~DidSetCookieResponse() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidSetCookieResponse; }
    static i32 static_message_id() { return (int)MessageID::DidSetCookieResponse; }
    virtual const char* message_name() const override { return "WebContentClient::DidSetCookieResponse"; }

    static ErrorOr<NonnullOwnPtr<DidSetCookieResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        return make<DidSetCookieResponse>();
    }

    static ErrorOr<IPC::MessageBuffer> static_encode()
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidSetCookieResponse));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode();
    }

private:
};

class DidSetCookie final : public IPC::Message {
public:
   typedef class DidSetCookieResponse ResponseType;

    DidSetCookie(DidSetCookie const&) = default;
    DidSetCookie(DidSetCookie&&) = default;
    DidSetCookie& operator=(DidSetCookie const&) = default;
    DidSetCookie(URL::URL url, Web::Cookie::ParsedCookie cookie, Web::Cookie::Source source) : m_url(move(url)), m_cookie(move(cookie)), m_source(move(source)) {}

    virtual ~DidSetCookie() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidSetCookie; }
    static i32 static_message_id() { return (int)MessageID::DidSetCookie; }
    virtual const char* message_name() const override { return "WebContentClient::DidSetCookie"; }

    static ErrorOr<NonnullOwnPtr<DidSetCookie>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto url = TRY((decoder.decode<URL::URL>()));
        auto cookie = TRY((decoder.decode<Web::Cookie::ParsedCookie>()));
        auto source = TRY((decoder.decode<Web::Cookie::Source>()));
        return make<DidSetCookie>(move(url), move(cookie), move(source));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(URL::URL const& url, Web::Cookie::ParsedCookie const& cookie, Web::Cookie::Source source)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidSetCookie));
        TRY(stream.encode(url));
        TRY(stream.encode(cookie));
        TRY(stream.encode(source));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_url, m_cookie, m_source);
    }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const Web::Cookie::ParsedCookie& cookie() const { return m_cookie; }
    Web::Cookie::ParsedCookie take_cookie() { return move(m_cookie); }

    Web::Cookie::Source source() const { return m_source; }

private:
    URL::URL m_url;
    Web::Cookie::ParsedCookie m_cookie;
    Web::Cookie::Source m_source;
};

class DidUpdateCookie final : public IPC::Message {
public:
    DidUpdateCookie(DidUpdateCookie const&) = default;
    DidUpdateCookie(DidUpdateCookie&&) = default;
    DidUpdateCookie& operator=(DidUpdateCookie const&) = default;
    DidUpdateCookie(Web::Cookie::Cookie cookie) : m_cookie(move(cookie)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, Web::Cookie::Cookie>)
    DidUpdateCookie(WrappedReturnType&& value)
        : m_cookie(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidUpdateCookie() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidUpdateCookie; }
    static i32 static_message_id() { return (int)MessageID::DidUpdateCookie; }
    virtual const char* message_name() const override { return "WebContentClient::DidUpdateCookie"; }

    static ErrorOr<NonnullOwnPtr<DidUpdateCookie>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto cookie = TRY((decoder.decode<Web::Cookie::Cookie>()));
        return make<DidUpdateCookie>(move(cookie));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(Web::Cookie::Cookie const& cookie)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidUpdateCookie));
        TRY(stream.encode(cookie));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_cookie);
    }

    const Web::Cookie::Cookie& cookie() const { return m_cookie; }
    Web::Cookie::Cookie take_cookie() { return move(m_cookie); }

private:
    Web::Cookie::Cookie m_cookie;
};

class DidExpireCookiesWithTimeOffset final : public IPC::Message {
public:
    DidExpireCookiesWithTimeOffset(DidExpireCookiesWithTimeOffset const&) = default;
    DidExpireCookiesWithTimeOffset(DidExpireCookiesWithTimeOffset&&) = default;
    DidExpireCookiesWithTimeOffset& operator=(DidExpireCookiesWithTimeOffset const&) = default;
    DidExpireCookiesWithTimeOffset(AK::Duration offset) : m_offset(move(offset)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, AK::Duration>)
    DidExpireCookiesWithTimeOffset(WrappedReturnType&& value)
        : m_offset(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidExpireCookiesWithTimeOffset() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidExpireCookiesWithTimeOffset; }
    static i32 static_message_id() { return (int)MessageID::DidExpireCookiesWithTimeOffset; }
    virtual const char* message_name() const override { return "WebContentClient::DidExpireCookiesWithTimeOffset"; }

    static ErrorOr<NonnullOwnPtr<DidExpireCookiesWithTimeOffset>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto offset = TRY((decoder.decode<AK::Duration>()));
        return make<DidExpireCookiesWithTimeOffset>(move(offset));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(AK::Duration offset)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidExpireCookiesWithTimeOffset));
        TRY(stream.encode(offset));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_offset);
    }

    AK::Duration offset() const { return m_offset; }

private:
    AK::Duration m_offset;
};

class DidUpdateResourceCount final : public IPC::Message {
public:
    DidUpdateResourceCount(DidUpdateResourceCount const&) = default;
    DidUpdateResourceCount(DidUpdateResourceCount&&) = default;
    DidUpdateResourceCount& operator=(DidUpdateResourceCount const&) = default;
    DidUpdateResourceCount(u64 page_id, i32 count_waiting) : m_page_id(move(page_id)), m_count_waiting(move(count_waiting)) {}

    virtual ~DidUpdateResourceCount() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidUpdateResourceCount; }
    static i32 static_message_id() { return (int)MessageID::DidUpdateResourceCount; }
    virtual const char* message_name() const override { return "WebContentClient::DidUpdateResourceCount"; }

    static ErrorOr<NonnullOwnPtr<DidUpdateResourceCount>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto count_waiting = TRY((decoder.decode<i32>()));
        return make<DidUpdateResourceCount>(move(page_id), move(count_waiting));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, i32 count_waiting)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidUpdateResourceCount));
        TRY(stream.encode(page_id));
        TRY(stream.encode(count_waiting));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_count_waiting);
    }

    u64 page_id() const { return m_page_id; }

    i32 count_waiting() const { return m_count_waiting; }

private:
    u64 m_page_id;
    i32 m_count_waiting;
};

class DidRequestNewWebViewResponse final : public IPC::Message {
public:
    DidRequestNewWebViewResponse(DidRequestNewWebViewResponse const&) = default;
    DidRequestNewWebViewResponse(DidRequestNewWebViewResponse&&) = default;
    DidRequestNewWebViewResponse& operator=(DidRequestNewWebViewResponse const&) = default;
    DidRequestNewWebViewResponse(String handle) : m_handle(move(handle)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, String>)
    DidRequestNewWebViewResponse(WrappedReturnType&& value)
        : m_handle(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestNewWebViewResponse() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestNewWebViewResponse; }
    static i32 static_message_id() { return (int)MessageID::DidRequestNewWebViewResponse; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestNewWebViewResponse"; }

    static ErrorOr<NonnullOwnPtr<DidRequestNewWebViewResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto handle = TRY((decoder.decode<String>()));
        return make<DidRequestNewWebViewResponse>(move(handle));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(StringView handle)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestNewWebViewResponse));
        TRY(stream.encode(handle));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_handle);
    }

    const String& handle() const { return m_handle; }
    String take_handle() { return move(m_handle); }

private:
    String m_handle;
};

class DidRequestNewWebView final : public IPC::Message {
public:
   typedef class DidRequestNewWebViewResponse ResponseType;

    DidRequestNewWebView(DidRequestNewWebView const&) = default;
    DidRequestNewWebView(DidRequestNewWebView&&) = default;
    DidRequestNewWebView& operator=(DidRequestNewWebView const&) = default;
    DidRequestNewWebView(u64 page_id, Web::HTML::ActivateTab activate_tab, Web::HTML::WebViewHints hints, Optional<u64> page_index) : m_page_id(move(page_id)), m_activate_tab(move(activate_tab)), m_hints(move(hints)), m_page_index(move(page_index)) {}

    virtual ~DidRequestNewWebView() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestNewWebView; }
    static i32 static_message_id() { return (int)MessageID::DidRequestNewWebView; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestNewWebView"; }

    static ErrorOr<NonnullOwnPtr<DidRequestNewWebView>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto activate_tab = TRY((decoder.decode<Web::HTML::ActivateTab>()));
        auto hints = TRY((decoder.decode<Web::HTML::WebViewHints>()));
        auto page_index = TRY((decoder.decode<Optional<u64>>()));
        return make<DidRequestNewWebView>(move(page_id), move(activate_tab), move(hints), move(page_index));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Web::HTML::ActivateTab const& activate_tab, Web::HTML::WebViewHints const& hints, Optional<u64> const& page_index)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestNewWebView));
        TRY(stream.encode(page_id));
        TRY(stream.encode(activate_tab));
        TRY(stream.encode(hints));
        TRY(stream.encode(page_index));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_activate_tab, m_hints, m_page_index);
    }

    u64 page_id() const { return m_page_id; }

    const Web::HTML::ActivateTab& activate_tab() const { return m_activate_tab; }
    Web::HTML::ActivateTab take_activate_tab() { return move(m_activate_tab); }

    const Web::HTML::WebViewHints& hints() const { return m_hints; }
    Web::HTML::WebViewHints take_hints() { return move(m_hints); }

    const Optional<u64>& page_index() const { return m_page_index; }
    Optional<u64> take_page_index() { return move(m_page_index); }

private:
    u64 m_page_id;
    Web::HTML::ActivateTab m_activate_tab;
    Web::HTML::WebViewHints m_hints;
    Optional<u64> m_page_index;
};

class DidRequestActivateTab final : public IPC::Message {
public:
    DidRequestActivateTab(DidRequestActivateTab const&) = default;
    DidRequestActivateTab(DidRequestActivateTab&&) = default;
    DidRequestActivateTab& operator=(DidRequestActivateTab const&) = default;
    DidRequestActivateTab(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestActivateTab(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestActivateTab() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestActivateTab; }
    static i32 static_message_id() { return (int)MessageID::DidRequestActivateTab; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestActivateTab"; }

    static ErrorOr<NonnullOwnPtr<DidRequestActivateTab>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        return make<DidRequestActivateTab>(move(page_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestActivateTab));
        TRY(stream.encode(page_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id);
    }

    u64 page_id() const { return m_page_id; }

private:
    u64 m_page_id;
};

class DidCloseBrowsingContext final : public IPC::Message {
public:
    DidCloseBrowsingContext(DidCloseBrowsingContext const&) = default;
    DidCloseBrowsingContext(DidCloseBrowsingContext&&) = default;
    DidCloseBrowsingContext& operator=(DidCloseBrowsingContext const&) = default;
    DidCloseBrowsingContext(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidCloseBrowsingContext(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidCloseBrowsingContext() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidCloseBrowsingContext; }
    static i32 static_message_id() { return (int)MessageID::DidCloseBrowsingContext; }
    virtual const char* message_name() const override { return "WebContentClient::DidCloseBrowsingContext"; }

    static ErrorOr<NonnullOwnPtr<DidCloseBrowsingContext>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        return make<DidCloseBrowsingContext>(move(page_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidCloseBrowsingContext));
        TRY(stream.encode(page_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id);
    }

    u64 page_id() const { return m_page_id; }

private:
    u64 m_page_id;
};

class DidRequestRestoreWindow final : public IPC::Message {
public:
    DidRequestRestoreWindow(DidRequestRestoreWindow const&) = default;
    DidRequestRestoreWindow(DidRequestRestoreWindow&&) = default;
    DidRequestRestoreWindow& operator=(DidRequestRestoreWindow const&) = default;
    DidRequestRestoreWindow(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestRestoreWindow(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestRestoreWindow() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestRestoreWindow; }
    static i32 static_message_id() { return (int)MessageID::DidRequestRestoreWindow; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestRestoreWindow"; }

    static ErrorOr<NonnullOwnPtr<DidRequestRestoreWindow>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        return make<DidRequestRestoreWindow>(move(page_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestRestoreWindow));
        TRY(stream.encode(page_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id);
    }

    u64 page_id() const { return m_page_id; }

private:
    u64 m_page_id;
};

class DidRequestRepositionWindow final : public IPC::Message {
public:
    DidRequestRepositionWindow(DidRequestRepositionWindow const&) = default;
    DidRequestRepositionWindow(DidRequestRepositionWindow&&) = default;
    DidRequestRepositionWindow& operator=(DidRequestRepositionWindow const&) = default;
    DidRequestRepositionWindow(u64 page_id, Gfx::IntPoint position) : m_page_id(move(page_id)), m_position(move(position)) {}

    virtual ~DidRequestRepositionWindow() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestRepositionWindow; }
    static i32 static_message_id() { return (int)MessageID::DidRequestRepositionWindow; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestRepositionWindow"; }

    static ErrorOr<NonnullOwnPtr<DidRequestRepositionWindow>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto position = TRY((decoder.decode<Gfx::IntPoint>()));
        return make<DidRequestRepositionWindow>(move(page_id), move(position));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Gfx::IntPoint position)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestRepositionWindow));
        TRY(stream.encode(page_id));
        TRY(stream.encode(position));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_position);
    }

    u64 page_id() const { return m_page_id; }

    Gfx::IntPoint position() const { return m_position; }

private:
    u64 m_page_id;
    Gfx::IntPoint m_position;
};

class DidRequestResizeWindow final : public IPC::Message {
public:
    DidRequestResizeWindow(DidRequestResizeWindow const&) = default;
    DidRequestResizeWindow(DidRequestResizeWindow&&) = default;
    DidRequestResizeWindow& operator=(DidRequestResizeWindow const&) = default;
    DidRequestResizeWindow(u64 page_id, Gfx::IntSize size) : m_page_id(move(page_id)), m_size(move(size)) {}

    virtual ~DidRequestResizeWindow() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestResizeWindow; }
    static i32 static_message_id() { return (int)MessageID::DidRequestResizeWindow; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestResizeWindow"; }

    static ErrorOr<NonnullOwnPtr<DidRequestResizeWindow>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto size = TRY((decoder.decode<Gfx::IntSize>()));
        return make<DidRequestResizeWindow>(move(page_id), move(size));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Gfx::IntSize size)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestResizeWindow));
        TRY(stream.encode(page_id));
        TRY(stream.encode(size));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_size);
    }

    u64 page_id() const { return m_page_id; }

    Gfx::IntSize size() const { return m_size; }

private:
    u64 m_page_id;
    Gfx::IntSize m_size;
};

class DidRequestMaximizeWindow final : public IPC::Message {
public:
    DidRequestMaximizeWindow(DidRequestMaximizeWindow const&) = default;
    DidRequestMaximizeWindow(DidRequestMaximizeWindow&&) = default;
    DidRequestMaximizeWindow& operator=(DidRequestMaximizeWindow const&) = default;
    DidRequestMaximizeWindow(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestMaximizeWindow(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestMaximizeWindow() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestMaximizeWindow; }
    static i32 static_message_id() { return (int)MessageID::DidRequestMaximizeWindow; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestMaximizeWindow"; }

    static ErrorOr<NonnullOwnPtr<DidRequestMaximizeWindow>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        return make<DidRequestMaximizeWindow>(move(page_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestMaximizeWindow));
        TRY(stream.encode(page_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id);
    }

    u64 page_id() const { return m_page_id; }

private:
    u64 m_page_id;
};

class DidRequestMinimizeWindow final : public IPC::Message {
public:
    DidRequestMinimizeWindow(DidRequestMinimizeWindow const&) = default;
    DidRequestMinimizeWindow(DidRequestMinimizeWindow&&) = default;
    DidRequestMinimizeWindow& operator=(DidRequestMinimizeWindow const&) = default;
    DidRequestMinimizeWindow(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestMinimizeWindow(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestMinimizeWindow() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestMinimizeWindow; }
    static i32 static_message_id() { return (int)MessageID::DidRequestMinimizeWindow; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestMinimizeWindow"; }

    static ErrorOr<NonnullOwnPtr<DidRequestMinimizeWindow>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        return make<DidRequestMinimizeWindow>(move(page_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestMinimizeWindow));
        TRY(stream.encode(page_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id);
    }

    u64 page_id() const { return m_page_id; }

private:
    u64 m_page_id;
};

class DidRequestFullscreenWindow final : public IPC::Message {
public:
    DidRequestFullscreenWindow(DidRequestFullscreenWindow const&) = default;
    DidRequestFullscreenWindow(DidRequestFullscreenWindow&&) = default;
    DidRequestFullscreenWindow& operator=(DidRequestFullscreenWindow const&) = default;
    DidRequestFullscreenWindow(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestFullscreenWindow(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestFullscreenWindow() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestFullscreenWindow; }
    static i32 static_message_id() { return (int)MessageID::DidRequestFullscreenWindow; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestFullscreenWindow"; }

    static ErrorOr<NonnullOwnPtr<DidRequestFullscreenWindow>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        return make<DidRequestFullscreenWindow>(move(page_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestFullscreenWindow));
        TRY(stream.encode(page_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id);
    }

    u64 page_id() const { return m_page_id; }

private:
    u64 m_page_id;
};

class DidRequestFile final : public IPC::Message {
public:
    DidRequestFile(DidRequestFile const&) = default;
    DidRequestFile(DidRequestFile&&) = default;
    DidRequestFile& operator=(DidRequestFile const&) = default;
    DidRequestFile(u64 page_id, ByteString path, i32 request_id) : m_page_id(move(page_id)), m_path(move(path)), m_request_id(move(request_id)) {}

    virtual ~DidRequestFile() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestFile; }
    static i32 static_message_id() { return (int)MessageID::DidRequestFile; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestFile"; }

    static ErrorOr<NonnullOwnPtr<DidRequestFile>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto path = TRY((decoder.decode<ByteString>()));
        auto request_id = TRY((decoder.decode<i32>()));
        return make<DidRequestFile>(move(page_id), move(path), move(request_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, StringView path, i32 request_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestFile));
        TRY(stream.encode(page_id));
        TRY(stream.encode(path));
        TRY(stream.encode(request_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_path, m_request_id);
    }

    u64 page_id() const { return m_page_id; }

    const ByteString& path() const { return m_path; }
    ByteString take_path() { return move(m_path); }

    i32 request_id() const { return m_request_id; }

private:
    u64 m_page_id;
    ByteString m_path;
    i32 m_request_id;
};

class DidRequestColorPicker final : public IPC::Message {
public:
    DidRequestColorPicker(DidRequestColorPicker const&) = default;
    DidRequestColorPicker(DidRequestColorPicker&&) = default;
    DidRequestColorPicker& operator=(DidRequestColorPicker const&) = default;
    DidRequestColorPicker(u64 page_id, Color current_color) : m_page_id(move(page_id)), m_current_color(move(current_color)) {}

    virtual ~DidRequestColorPicker() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestColorPicker; }
    static i32 static_message_id() { return (int)MessageID::DidRequestColorPicker; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestColorPicker"; }

    static ErrorOr<NonnullOwnPtr<DidRequestColorPicker>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto current_color = TRY((decoder.decode<Color>()));
        return make<DidRequestColorPicker>(move(page_id), move(current_color));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Color const& current_color)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestColorPicker));
        TRY(stream.encode(page_id));
        TRY(stream.encode(current_color));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_current_color);
    }

    u64 page_id() const { return m_page_id; }

    const Color& current_color() const { return m_current_color; }
    Color take_current_color() { return move(m_current_color); }

private:
    u64 m_page_id;
    Color m_current_color;
};

class DidRequestFilePicker final : public IPC::Message {
public:
    DidRequestFilePicker(DidRequestFilePicker const&) = default;
    DidRequestFilePicker(DidRequestFilePicker&&) = default;
    DidRequestFilePicker& operator=(DidRequestFilePicker const&) = default;
    DidRequestFilePicker(u64 page_id, Web::HTML::FileFilter accepted_file_types, Web::HTML::AllowMultipleFiles allow_multiple_files) : m_page_id(move(page_id)), m_accepted_file_types(move(accepted_file_types)), m_allow_multiple_files(move(allow_multiple_files)) {}

    virtual ~DidRequestFilePicker() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestFilePicker; }
    static i32 static_message_id() { return (int)MessageID::DidRequestFilePicker; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestFilePicker"; }

    static ErrorOr<NonnullOwnPtr<DidRequestFilePicker>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto accepted_file_types = TRY((decoder.decode<Web::HTML::FileFilter>()));
        auto allow_multiple_files = TRY((decoder.decode<Web::HTML::AllowMultipleFiles>()));
        return make<DidRequestFilePicker>(move(page_id), move(accepted_file_types), move(allow_multiple_files));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Web::HTML::FileFilter const& accepted_file_types, Web::HTML::AllowMultipleFiles allow_multiple_files)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestFilePicker));
        TRY(stream.encode(page_id));
        TRY(stream.encode(accepted_file_types));
        TRY(stream.encode(allow_multiple_files));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_accepted_file_types, m_allow_multiple_files);
    }

    u64 page_id() const { return m_page_id; }

    const Web::HTML::FileFilter& accepted_file_types() const { return m_accepted_file_types; }
    Web::HTML::FileFilter take_accepted_file_types() { return move(m_accepted_file_types); }

    Web::HTML::AllowMultipleFiles allow_multiple_files() const { return m_allow_multiple_files; }

private:
    u64 m_page_id;
    Web::HTML::FileFilter m_accepted_file_types;
    Web::HTML::AllowMultipleFiles m_allow_multiple_files;
};

class DidRequestSelectDropdown final : public IPC::Message {
public:
    DidRequestSelectDropdown(DidRequestSelectDropdown const&) = default;
    DidRequestSelectDropdown(DidRequestSelectDropdown&&) = default;
    DidRequestSelectDropdown& operator=(DidRequestSelectDropdown const&) = default;
    DidRequestSelectDropdown(u64 page_id, Gfx::IntPoint content_position, i32 minimum_width, Vector<Web::HTML::SelectItem> items) : m_page_id(move(page_id)), m_content_position(move(content_position)), m_minimum_width(move(minimum_width)), m_items(move(items)) {}

    virtual ~DidRequestSelectDropdown() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestSelectDropdown; }
    static i32 static_message_id() { return (int)MessageID::DidRequestSelectDropdown; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestSelectDropdown"; }

    static ErrorOr<NonnullOwnPtr<DidRequestSelectDropdown>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto content_position = TRY((decoder.decode<Gfx::IntPoint>()));
        auto minimum_width = TRY((decoder.decode<i32>()));
        auto items = TRY((decoder.decode<Vector<Web::HTML::SelectItem>>()));
        return make<DidRequestSelectDropdown>(move(page_id), move(content_position), move(minimum_width), move(items));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Gfx::IntPoint content_position, i32 minimum_width, ReadonlySpan<Web::HTML::SelectItem> items)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidRequestSelectDropdown));
        TRY(stream.encode(page_id));
        TRY(stream.encode(content_position));
        TRY(stream.encode(minimum_width));
        TRY(stream.encode(items));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_content_position, m_minimum_width, m_items);
    }

    u64 page_id() const { return m_page_id; }

    Gfx::IntPoint content_position() const { return m_content_position; }

    i32 minimum_width() const { return m_minimum_width; }

    const Vector<Web::HTML::SelectItem>& items() const { return m_items; }
    Vector<Web::HTML::SelectItem> take_items() { return move(m_items); }

private:
    u64 m_page_id;
    Gfx::IntPoint m_content_position;
    i32 m_minimum_width;
    Vector<Web::HTML::SelectItem> m_items;
};

class DidFinishHandlingInputEvent final : public IPC::Message {
public:
    DidFinishHandlingInputEvent(DidFinishHandlingInputEvent const&) = default;
    DidFinishHandlingInputEvent(DidFinishHandlingInputEvent&&) = default;
    DidFinishHandlingInputEvent& operator=(DidFinishHandlingInputEvent const&) = default;
    DidFinishHandlingInputEvent(u64 page_id, Web::EventResult event_result) : m_page_id(move(page_id)), m_event_result(move(event_result)) {}

    virtual ~DidFinishHandlingInputEvent() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidFinishHandlingInputEvent; }
    static i32 static_message_id() { return (int)MessageID::DidFinishHandlingInputEvent; }
    virtual const char* message_name() const override { return "WebContentClient::DidFinishHandlingInputEvent"; }

    static ErrorOr<NonnullOwnPtr<DidFinishHandlingInputEvent>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto event_result = TRY((decoder.decode<Web::EventResult>()));
        return make<DidFinishHandlingInputEvent>(move(page_id), move(event_result));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Web::EventResult event_result)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidFinishHandlingInputEvent));
        TRY(stream.encode(page_id));
        TRY(stream.encode(event_result));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_event_result);
    }

    u64 page_id() const { return m_page_id; }

    Web::EventResult event_result() const { return m_event_result; }

private:
    u64 m_page_id;
    Web::EventResult m_event_result;
};

class DidChangeThemeColor final : public IPC::Message {
public:
    DidChangeThemeColor(DidChangeThemeColor const&) = default;
    DidChangeThemeColor(DidChangeThemeColor&&) = default;
    DidChangeThemeColor& operator=(DidChangeThemeColor const&) = default;
    DidChangeThemeColor(u64 page_id, Gfx::Color color) : m_page_id(move(page_id)), m_color(move(color)) {}

    virtual ~DidChangeThemeColor() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidChangeThemeColor; }
    static i32 static_message_id() { return (int)MessageID::DidChangeThemeColor; }
    virtual const char* message_name() const override { return "WebContentClient::DidChangeThemeColor"; }

    static ErrorOr<NonnullOwnPtr<DidChangeThemeColor>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto color = TRY((decoder.decode<Gfx::Color>()));
        return make<DidChangeThemeColor>(move(page_id), move(color));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Gfx::Color color)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidChangeThemeColor));
        TRY(stream.encode(page_id));
        TRY(stream.encode(color));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_color);
    }

    u64 page_id() const { return m_page_id; }

    Gfx::Color color() const { return m_color; }

private:
    u64 m_page_id;
    Gfx::Color m_color;
};

class DidInsertClipboardEntry final : public IPC::Message {
public:
    DidInsertClipboardEntry(DidInsertClipboardEntry const&) = default;
    DidInsertClipboardEntry(DidInsertClipboardEntry&&) = default;
    DidInsertClipboardEntry& operator=(DidInsertClipboardEntry const&) = default;
    DidInsertClipboardEntry(u64 page_id, String data, String presentation_style, String mime_type) : m_page_id(move(page_id)), m_data(move(data)), m_presentation_style(move(presentation_style)), m_mime_type(move(mime_type)) {}

    virtual ~DidInsertClipboardEntry() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidInsertClipboardEntry; }
    static i32 static_message_id() { return (int)MessageID::DidInsertClipboardEntry; }
    virtual const char* message_name() const override { return "WebContentClient::DidInsertClipboardEntry"; }

    static ErrorOr<NonnullOwnPtr<DidInsertClipboardEntry>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto data = TRY((decoder.decode<String>()));
        auto presentation_style = TRY((decoder.decode<String>()));
        auto mime_type = TRY((decoder.decode<String>()));
        return make<DidInsertClipboardEntry>(move(page_id), move(data), move(presentation_style), move(mime_type));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, StringView data, StringView presentation_style, StringView mime_type)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidInsertClipboardEntry));
        TRY(stream.encode(page_id));
        TRY(stream.encode(data));
        TRY(stream.encode(presentation_style));
        TRY(stream.encode(mime_type));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_data, m_presentation_style, m_mime_type);
    }

    u64 page_id() const { return m_page_id; }

    const String& data() const { return m_data; }
    String take_data() { return move(m_data); }

    const String& presentation_style() const { return m_presentation_style; }
    String take_presentation_style() { return move(m_presentation_style); }

    const String& mime_type() const { return m_mime_type; }
    String take_mime_type() { return move(m_mime_type); }

private:
    u64 m_page_id;
    String m_data;
    String m_presentation_style;
    String m_mime_type;
};

class DidUpdateNavigationButtonsState final : public IPC::Message {
public:
    DidUpdateNavigationButtonsState(DidUpdateNavigationButtonsState const&) = default;
    DidUpdateNavigationButtonsState(DidUpdateNavigationButtonsState&&) = default;
    DidUpdateNavigationButtonsState& operator=(DidUpdateNavigationButtonsState const&) = default;
    DidUpdateNavigationButtonsState(u64 page_id, bool back_enabled, bool forward_enabled) : m_page_id(move(page_id)), m_back_enabled(move(back_enabled)), m_forward_enabled(move(forward_enabled)) {}

    virtual ~DidUpdateNavigationButtonsState() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidUpdateNavigationButtonsState; }
    static i32 static_message_id() { return (int)MessageID::DidUpdateNavigationButtonsState; }
    virtual const char* message_name() const override { return "WebContentClient::DidUpdateNavigationButtonsState"; }

    static ErrorOr<NonnullOwnPtr<DidUpdateNavigationButtonsState>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto back_enabled = TRY((decoder.decode<bool>()));
        auto forward_enabled = TRY((decoder.decode<bool>()));
        return make<DidUpdateNavigationButtonsState>(move(page_id), move(back_enabled), move(forward_enabled));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, bool back_enabled, bool forward_enabled)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidUpdateNavigationButtonsState));
        TRY(stream.encode(page_id));
        TRY(stream.encode(back_enabled));
        TRY(stream.encode(forward_enabled));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_back_enabled, m_forward_enabled);
    }

    u64 page_id() const { return m_page_id; }

    bool back_enabled() const { return m_back_enabled; }

    bool forward_enabled() const { return m_forward_enabled; }

private:
    u64 m_page_id;
    bool m_back_enabled;
    bool m_forward_enabled;
};

class DidAllocateBackingStores final : public IPC::Message {
public:
    DidAllocateBackingStores(DidAllocateBackingStores const&) = default;
    DidAllocateBackingStores(DidAllocateBackingStores&&) = default;
    DidAllocateBackingStores& operator=(DidAllocateBackingStores const&) = default;
    DidAllocateBackingStores(u64 page_id, i32 front_bitmap_id, Gfx::ShareableBitmap front_bitmap, i32 back_bitmap_id, Gfx::ShareableBitmap back_bitmap) : m_page_id(move(page_id)), m_front_bitmap_id(move(front_bitmap_id)), m_front_bitmap(move(front_bitmap)), m_back_bitmap_id(move(back_bitmap_id)), m_back_bitmap(move(back_bitmap)) {}

    virtual ~DidAllocateBackingStores() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidAllocateBackingStores; }
    static i32 static_message_id() { return (int)MessageID::DidAllocateBackingStores; }
    virtual const char* message_name() const override { return "WebContentClient::DidAllocateBackingStores"; }

    static ErrorOr<NonnullOwnPtr<DidAllocateBackingStores>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto front_bitmap_id = TRY((decoder.decode<i32>()));
        auto front_bitmap = TRY((decoder.decode<Gfx::ShareableBitmap>()));
        auto back_bitmap_id = TRY((decoder.decode<i32>()));
        auto back_bitmap = TRY((decoder.decode<Gfx::ShareableBitmap>()));
        return make<DidAllocateBackingStores>(move(page_id), move(front_bitmap_id), move(front_bitmap), move(back_bitmap_id), move(back_bitmap));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, i32 front_bitmap_id, Gfx::ShareableBitmap const& front_bitmap, i32 back_bitmap_id, Gfx::ShareableBitmap const& back_bitmap)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidAllocateBackingStores));
        TRY(stream.encode(page_id));
        TRY(stream.encode(front_bitmap_id));
        TRY(stream.encode(front_bitmap));
        TRY(stream.encode(back_bitmap_id));
        TRY(stream.encode(back_bitmap));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_front_bitmap_id, m_front_bitmap, m_back_bitmap_id, m_back_bitmap);
    }

    u64 page_id() const { return m_page_id; }

    i32 front_bitmap_id() const { return m_front_bitmap_id; }

    const Gfx::ShareableBitmap& front_bitmap() const { return m_front_bitmap; }
    Gfx::ShareableBitmap take_front_bitmap() { return move(m_front_bitmap); }

    i32 back_bitmap_id() const { return m_back_bitmap_id; }

    const Gfx::ShareableBitmap& back_bitmap() const { return m_back_bitmap; }
    Gfx::ShareableBitmap take_back_bitmap() { return move(m_back_bitmap); }

private:
    u64 m_page_id;
    i32 m_front_bitmap_id;
    Gfx::ShareableBitmap m_front_bitmap;
    i32 m_back_bitmap_id;
    Gfx::ShareableBitmap m_back_bitmap;
};

class DidChangeAudioPlayState final : public IPC::Message {
public:
    DidChangeAudioPlayState(DidChangeAudioPlayState const&) = default;
    DidChangeAudioPlayState(DidChangeAudioPlayState&&) = default;
    DidChangeAudioPlayState& operator=(DidChangeAudioPlayState const&) = default;
    DidChangeAudioPlayState(u64 page_id, Web::HTML::AudioPlayState play_state) : m_page_id(move(page_id)), m_play_state(move(play_state)) {}

    virtual ~DidChangeAudioPlayState() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidChangeAudioPlayState; }
    static i32 static_message_id() { return (int)MessageID::DidChangeAudioPlayState; }
    virtual const char* message_name() const override { return "WebContentClient::DidChangeAudioPlayState"; }

    static ErrorOr<NonnullOwnPtr<DidChangeAudioPlayState>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto play_state = TRY((decoder.decode<Web::HTML::AudioPlayState>()));
        return make<DidChangeAudioPlayState>(move(page_id), move(play_state));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, Web::HTML::AudioPlayState play_state)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidChangeAudioPlayState));
        TRY(stream.encode(page_id));
        TRY(stream.encode(play_state));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_play_state);
    }

    u64 page_id() const { return m_page_id; }

    Web::HTML::AudioPlayState play_state() const { return m_play_state; }

private:
    u64 m_page_id;
    Web::HTML::AudioPlayState m_play_state;
};

class DidExecuteJsConsoleInput final : public IPC::Message {
public:
    DidExecuteJsConsoleInput(DidExecuteJsConsoleInput const&) = default;
    DidExecuteJsConsoleInput(DidExecuteJsConsoleInput&&) = default;
    DidExecuteJsConsoleInput& operator=(DidExecuteJsConsoleInput const&) = default;
    DidExecuteJsConsoleInput(u64 page_id, JsonValue result) : m_page_id(move(page_id)), m_result(move(result)) {}

    virtual ~DidExecuteJsConsoleInput() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidExecuteJsConsoleInput; }
    static i32 static_message_id() { return (int)MessageID::DidExecuteJsConsoleInput; }
    virtual const char* message_name() const override { return "WebContentClient::DidExecuteJsConsoleInput"; }

    static ErrorOr<NonnullOwnPtr<DidExecuteJsConsoleInput>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto result = TRY((decoder.decode<JsonValue>()));
        return make<DidExecuteJsConsoleInput>(move(page_id), move(result));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, JsonValue const& result)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidExecuteJsConsoleInput));
        TRY(stream.encode(page_id));
        TRY(stream.encode(result));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_result);
    }

    u64 page_id() const { return m_page_id; }

    const JsonValue& result() const { return m_result; }
    JsonValue take_result() { return move(m_result); }

private:
    u64 m_page_id;
    JsonValue m_result;
};

class DidOutputJsConsoleMessage final : public IPC::Message {
public:
    DidOutputJsConsoleMessage(DidOutputJsConsoleMessage const&) = default;
    DidOutputJsConsoleMessage(DidOutputJsConsoleMessage&&) = default;
    DidOutputJsConsoleMessage& operator=(DidOutputJsConsoleMessage const&) = default;
    DidOutputJsConsoleMessage(u64 page_id, i32 message_index) : m_page_id(move(page_id)), m_message_index(move(message_index)) {}

    virtual ~DidOutputJsConsoleMessage() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidOutputJsConsoleMessage; }
    static i32 static_message_id() { return (int)MessageID::DidOutputJsConsoleMessage; }
    virtual const char* message_name() const override { return "WebContentClient::DidOutputJsConsoleMessage"; }

    static ErrorOr<NonnullOwnPtr<DidOutputJsConsoleMessage>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto message_index = TRY((decoder.decode<i32>()));
        return make<DidOutputJsConsoleMessage>(move(page_id), move(message_index));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, i32 message_index)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidOutputJsConsoleMessage));
        TRY(stream.encode(page_id));
        TRY(stream.encode(message_index));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_message_index);
    }

    u64 page_id() const { return m_page_id; }

    i32 message_index() const { return m_message_index; }

private:
    u64 m_page_id;
    i32 m_message_index;
};

class DidGetJsConsoleMessages final : public IPC::Message {
public:
    DidGetJsConsoleMessages(DidGetJsConsoleMessages const&) = default;
    DidGetJsConsoleMessages(DidGetJsConsoleMessages&&) = default;
    DidGetJsConsoleMessages& operator=(DidGetJsConsoleMessages const&) = default;
    DidGetJsConsoleMessages(u64 page_id, i32 start_index, Vector<WebView::ConsoleOutput> console_output) : m_page_id(move(page_id)), m_start_index(move(start_index)), m_console_output(move(console_output)) {}

    virtual ~DidGetJsConsoleMessages() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidGetJsConsoleMessages; }
    static i32 static_message_id() { return (int)MessageID::DidGetJsConsoleMessages; }
    virtual const char* message_name() const override { return "WebContentClient::DidGetJsConsoleMessages"; }

    static ErrorOr<NonnullOwnPtr<DidGetJsConsoleMessages>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto start_index = TRY((decoder.decode<i32>()));
        auto console_output = TRY((decoder.decode<Vector<WebView::ConsoleOutput>>()));
        return make<DidGetJsConsoleMessages>(move(page_id), move(start_index), move(console_output));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, i32 start_index, ReadonlySpan<WebView::ConsoleOutput> console_output)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidGetJsConsoleMessages));
        TRY(stream.encode(page_id));
        TRY(stream.encode(start_index));
        TRY(stream.encode(console_output));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_start_index, m_console_output);
    }

    u64 page_id() const { return m_page_id; }

    i32 start_index() const { return m_start_index; }

    const Vector<WebView::ConsoleOutput>& console_output() const { return m_console_output; }
    Vector<WebView::ConsoleOutput> take_console_output() { return move(m_console_output); }

private:
    u64 m_page_id;
    i32 m_start_index;
    Vector<WebView::ConsoleOutput> m_console_output;
};

class DidFinishTest final : public IPC::Message {
public:
    DidFinishTest(DidFinishTest const&) = default;
    DidFinishTest(DidFinishTest&&) = default;
    DidFinishTest& operator=(DidFinishTest const&) = default;
    DidFinishTest(u64 page_id, String text) : m_page_id(move(page_id)), m_text(move(text)) {}

    virtual ~DidFinishTest() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidFinishTest; }
    static i32 static_message_id() { return (int)MessageID::DidFinishTest; }
    virtual const char* message_name() const override { return "WebContentClient::DidFinishTest"; }

    static ErrorOr<NonnullOwnPtr<DidFinishTest>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto text = TRY((decoder.decode<String>()));
        return make<DidFinishTest>(move(page_id), move(text));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, StringView text)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidFinishTest));
        TRY(stream.encode(page_id));
        TRY(stream.encode(text));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_text);
    }

    u64 page_id() const { return m_page_id; }

    const String& text() const { return m_text; }
    String take_text() { return move(m_text); }

private:
    u64 m_page_id;
    String m_text;
};

class DidSetTestTimeout final : public IPC::Message {
public:
    DidSetTestTimeout(DidSetTestTimeout const&) = default;
    DidSetTestTimeout(DidSetTestTimeout&&) = default;
    DidSetTestTimeout& operator=(DidSetTestTimeout const&) = default;
    DidSetTestTimeout(u64 page_id, double milliseconds) : m_page_id(move(page_id)), m_milliseconds(move(milliseconds)) {}

    virtual ~DidSetTestTimeout() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidSetTestTimeout; }
    static i32 static_message_id() { return (int)MessageID::DidSetTestTimeout; }
    virtual const char* message_name() const override { return "WebContentClient::DidSetTestTimeout"; }

    static ErrorOr<NonnullOwnPtr<DidSetTestTimeout>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto milliseconds = TRY((decoder.decode<double>()));
        return make<DidSetTestTimeout>(move(page_id), move(milliseconds));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, double milliseconds)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidSetTestTimeout));
        TRY(stream.encode(page_id));
        TRY(stream.encode(milliseconds));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_milliseconds);
    }

    u64 page_id() const { return m_page_id; }

    double milliseconds() const { return m_milliseconds; }

private:
    u64 m_page_id;
    double m_milliseconds;
};

class DidSetBrowserZoom final : public IPC::Message {
public:
    DidSetBrowserZoom(DidSetBrowserZoom const&) = default;
    DidSetBrowserZoom(DidSetBrowserZoom&&) = default;
    DidSetBrowserZoom& operator=(DidSetBrowserZoom const&) = default;
    DidSetBrowserZoom(u64 page_id, double factor) : m_page_id(move(page_id)), m_factor(move(factor)) {}

    virtual ~DidSetBrowserZoom() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidSetBrowserZoom; }
    static i32 static_message_id() { return (int)MessageID::DidSetBrowserZoom; }
    virtual const char* message_name() const override { return "WebContentClient::DidSetBrowserZoom"; }

    static ErrorOr<NonnullOwnPtr<DidSetBrowserZoom>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto factor = TRY((decoder.decode<double>()));
        return make<DidSetBrowserZoom>(move(page_id), move(factor));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, double factor)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidSetBrowserZoom));
        TRY(stream.encode(page_id));
        TRY(stream.encode(factor));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_factor);
    }

    u64 page_id() const { return m_page_id; }

    double factor() const { return m_factor; }

private:
    u64 m_page_id;
    double m_factor;
};

class DidFindInPage final : public IPC::Message {
public:
    DidFindInPage(DidFindInPage const&) = default;
    DidFindInPage(DidFindInPage&&) = default;
    DidFindInPage& operator=(DidFindInPage const&) = default;
    DidFindInPage(u64 page_id, size_t current_match_index, Optional<size_t> total_match_count) : m_page_id(move(page_id)), m_current_match_index(move(current_match_index)), m_total_match_count(move(total_match_count)) {}

    virtual ~DidFindInPage() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::DidFindInPage; }
    static i32 static_message_id() { return (int)MessageID::DidFindInPage; }
    virtual const char* message_name() const override { return "WebContentClient::DidFindInPage"; }

    static ErrorOr<NonnullOwnPtr<DidFindInPage>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        auto current_match_index = TRY((decoder.decode<size_t>()));
        auto total_match_count = TRY((decoder.decode<Optional<size_t>>()));
        return make<DidFindInPage>(move(page_id), move(current_match_index), move(total_match_count));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id, size_t current_match_index, Optional<size_t> const& total_match_count)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::DidFindInPage));
        TRY(stream.encode(page_id));
        TRY(stream.encode(current_match_index));
        TRY(stream.encode(total_match_count));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id, m_current_match_index, m_total_match_count);
    }

    u64 page_id() const { return m_page_id; }

    size_t current_match_index() const { return m_current_match_index; }

    const Optional<size_t>& total_match_count() const { return m_total_match_count; }
    Optional<size_t> take_total_match_count() { return move(m_total_match_count); }

private:
    u64 m_page_id;
    size_t m_current_match_index;
    Optional<size_t> m_total_match_count;
};

class RequestWorkerAgentResponse final : public IPC::Message {
public:
    RequestWorkerAgentResponse(RequestWorkerAgentResponse const&) = default;
    RequestWorkerAgentResponse(RequestWorkerAgentResponse&&) = default;
    RequestWorkerAgentResponse& operator=(RequestWorkerAgentResponse const&) = default;
    RequestWorkerAgentResponse(IPC::File socket) : m_socket(move(socket)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, IPC::File>)
    RequestWorkerAgentResponse(WrappedReturnType&& value)
        : m_socket(forward<WrappedReturnType>(value))
    {
    }

    virtual ~RequestWorkerAgentResponse() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::RequestWorkerAgentResponse; }
    static i32 static_message_id() { return (int)MessageID::RequestWorkerAgentResponse; }
    virtual const char* message_name() const override { return "WebContentClient::RequestWorkerAgentResponse"; }

    static ErrorOr<NonnullOwnPtr<RequestWorkerAgentResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto socket = TRY((decoder.decode<IPC::File>()));
        return make<RequestWorkerAgentResponse>(move(socket));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(IPC::File const& socket)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::RequestWorkerAgentResponse));
        TRY(stream.encode(socket));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_socket);
    }

    const IPC::File& socket() const { return m_socket; }
    IPC::File take_socket() { return move(m_socket); }

private:
    IPC::File m_socket;
};

class RequestWorkerAgent final : public IPC::Message {
public:
   typedef class RequestWorkerAgentResponse ResponseType;

    RequestWorkerAgent(RequestWorkerAgent const&) = default;
    RequestWorkerAgent(RequestWorkerAgent&&) = default;
    RequestWorkerAgent& operator=(RequestWorkerAgent const&) = default;
    RequestWorkerAgent(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    RequestWorkerAgent(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~RequestWorkerAgent() override = default;

    static constexpr u32 ENDPOINT_MAGIC = 2565717604;

    virtual u32 endpoint_magic() const override { return ENDPOINT_MAGIC; }
    virtual i32 message_id() const override { return (int)MessageID::RequestWorkerAgent; }
    static i32 static_message_id() { return (int)MessageID::RequestWorkerAgent; }
    virtual const char* message_name() const override { return "WebContentClient::RequestWorkerAgent"; }

    static ErrorOr<NonnullOwnPtr<RequestWorkerAgent>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };
        auto page_id = TRY((decoder.decode<u64>()));
        return make<RequestWorkerAgent>(move(page_id));
    }

    static ErrorOr<IPC::MessageBuffer> static_encode(u64 page_id)
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(ENDPOINT_MAGIC));
        TRY(stream.encode((int)MessageID::RequestWorkerAgent));
        TRY(stream.encode(page_id));
        return buffer;
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        return static_encode(m_page_id);
    }

    u64 page_id() const { return m_page_id; }

private:
    u64 m_page_id;
};

} // namespace Messages::WebContentClient

template<typename LocalEndpoint, typename PeerEndpoint>
class WebContentClientProxy {
public:
    // Used to disambiguate the constructor call.
    struct Tag { };

    WebContentClientProxy(IPC::Connection<LocalEndpoint, PeerEndpoint>& connection, Tag)
        : m_connection(connection)
    { }

    void async_did_request_new_process_for_navigation(u64 page_id, URL::URL const& url) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestNewProcessForNavigation::static_encode(page_id, move(url)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_start_loading(u64 page_id, URL::URL const& url, bool is_redirect) {
        auto message_buffer = MUST(Messages::WebContentClient::DidStartLoading::static_encode(page_id, move(url), is_redirect));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_finish_loading(u64 page_id, URL::URL const& url) {
        auto message_buffer = MUST(Messages::WebContentClient::DidFinishLoading::static_encode(page_id, move(url)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_refresh(u64 page_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestRefresh::static_encode(page_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_paint(u64 page_id, Gfx::IntRect const& content_rect, i32 bitmap_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidPaint::static_encode(page_id, move(content_rect), bitmap_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_cursor_change(u64 page_id, Gfx::Cursor const& cursor) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestCursorChange::static_encode(page_id, move(cursor)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_change_title(u64 page_id, StringView title) {
        auto message_buffer = MUST(Messages::WebContentClient::DidChangeTitle::static_encode(page_id, title));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_change_url(u64 page_id, URL::URL const& url) {
        auto message_buffer = MUST(Messages::WebContentClient::DidChangeUrl::static_encode(page_id, move(url)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_tooltip_override(u64 page_id, Gfx::IntPoint position, StringView title) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestTooltipOverride::static_encode(page_id, position, title));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_stop_tooltip_override(u64 page_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidStopTooltipOverride::static_encode(page_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_enter_tooltip_area(u64 page_id, StringView title) {
        auto message_buffer = MUST(Messages::WebContentClient::DidEnterTooltipArea::static_encode(page_id, title));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_leave_tooltip_area(u64 page_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidLeaveTooltipArea::static_encode(page_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_hover_link(u64 page_id, URL::URL const& url) {
        auto message_buffer = MUST(Messages::WebContentClient::DidHoverLink::static_encode(page_id, move(url)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_unhover_link(u64 page_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidUnhoverLink::static_encode(page_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_click_link(u64 page_id, URL::URL const& url, StringView target, unsigned modifiers) {
        auto message_buffer = MUST(Messages::WebContentClient::DidClickLink::static_encode(page_id, move(url), target, modifiers));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_middle_click_link(u64 page_id, URL::URL const& url, StringView target, unsigned modifiers) {
        auto message_buffer = MUST(Messages::WebContentClient::DidMiddleClickLink::static_encode(page_id, move(url), target, modifiers));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_context_menu(u64 page_id, Gfx::IntPoint content_position) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestContextMenu::static_encode(page_id, content_position));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_link_context_menu(u64 page_id, Gfx::IntPoint content_position, URL::URL const& url, StringView target, unsigned modifiers) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestLinkContextMenu::static_encode(page_id, content_position, move(url), target, modifiers));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_image_context_menu(u64 page_id, Gfx::IntPoint content_position, URL::URL const& url, StringView target, unsigned modifiers, Optional<Gfx::ShareableBitmap> const& bitmap) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestImageContextMenu::static_encode(page_id, content_position, move(url), target, modifiers, move(bitmap)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_media_context_menu(u64 page_id, Gfx::IntPoint content_position, StringView target, unsigned modifiers, Web::Page::MediaContextMenu const& menu) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestMediaContextMenu::static_encode(page_id, content_position, target, modifiers, move(menu)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_alert(u64 page_id, StringView message) {
        VERIFY(Utf8View { message }.validate());
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestAlert::static_encode(page_id, message));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_alert(u64 page_id, String const& message) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestAlert::static_encode(page_id, message));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_confirm(u64 page_id, StringView message) {
        VERIFY(Utf8View { message }.validate());
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestConfirm::static_encode(page_id, message));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_confirm(u64 page_id, String const& message) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestConfirm::static_encode(page_id, message));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_prompt(u64 page_id, StringView message, StringView default_) {
        VERIFY(Utf8View { message }.validate());
        VERIFY(Utf8View { default_ }.validate());
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestPrompt::static_encode(page_id, message, default_));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_prompt(u64 page_id, String const& message, String const& default_) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestPrompt::static_encode(page_id, message, default_));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_set_prompt_text(u64 page_id, StringView message) {
        VERIFY(Utf8View { message }.validate());
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestSetPromptText::static_encode(page_id, message));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_set_prompt_text(u64 page_id, String const& message) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestSetPromptText::static_encode(page_id, message));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_accept_dialog(u64 page_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestAcceptDialog::static_encode(page_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_dismiss_dialog(u64 page_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestDismissDialog::static_encode(page_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_get_source(u64 page_id, URL::URL const& url, URL::URL const& base_url, StringView source) {
        VERIFY(Utf8View { source }.validate());
        auto message_buffer = MUST(Messages::WebContentClient::DidGetSource::static_encode(page_id, move(url), move(base_url), source));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_get_source(u64 page_id, URL::URL const& url, URL::URL const& base_url, String const& source) {
        auto message_buffer = MUST(Messages::WebContentClient::DidGetSource::static_encode(page_id, move(url), move(base_url), source));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_inspect_dom_tree(u64 page_id, StringView dom_tree) {
        VERIFY(Utf8View { dom_tree }.validate());
        auto message_buffer = MUST(Messages::WebContentClient::DidInspectDomTree::static_encode(page_id, dom_tree));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_inspect_dom_tree(u64 page_id, String const& dom_tree) {
        auto message_buffer = MUST(Messages::WebContentClient::DidInspectDomTree::static_encode(page_id, dom_tree));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_inspect_dom_node(u64 page_id, WebView::DOMNodeProperties const& properties) {
        auto message_buffer = MUST(Messages::WebContentClient::DidInspectDomNode::static_encode(page_id, move(properties)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_inspect_accessibility_tree(u64 page_id, StringView accessibility_tree) {
        VERIFY(Utf8View { accessibility_tree }.validate());
        auto message_buffer = MUST(Messages::WebContentClient::DidInspectAccessibilityTree::static_encode(page_id, accessibility_tree));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_inspect_accessibility_tree(u64 page_id, String const& accessibility_tree) {
        auto message_buffer = MUST(Messages::WebContentClient::DidInspectAccessibilityTree::static_encode(page_id, accessibility_tree));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_get_hovered_node_id(u64 page_id, Web::UniqueNodeID const& node_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidGetHoveredNodeId::static_encode(page_id, move(node_id)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_finish_editing_dom_node(u64 page_id, Optional<Web::UniqueNodeID> const& node_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidFinishEditingDomNode::static_encode(page_id, move(node_id)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_mutate_dom(u64 page_id, WebView::Mutation const& mutation) {
        auto message_buffer = MUST(Messages::WebContentClient::DidMutateDom::static_encode(page_id, move(mutation)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_get_dom_node_html(u64 page_id, StringView html) {
        VERIFY(Utf8View { html }.validate());
        auto message_buffer = MUST(Messages::WebContentClient::DidGetDomNodeHtml::static_encode(page_id, html));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_get_dom_node_html(u64 page_id, String const& html) {
        auto message_buffer = MUST(Messages::WebContentClient::DidGetDomNodeHtml::static_encode(page_id, html));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_list_style_sheets(u64 page_id, ReadonlySpan<Web::CSS::StyleSheetIdentifier> style_sheets) {
        auto message_buffer = MUST(Messages::WebContentClient::DidListStyleSheets::static_encode(page_id, style_sheets));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_get_style_sheet_source(u64 page_id, Web::CSS::StyleSheetIdentifier const& identifier, URL::URL const& base_url, StringView source) {
        VERIFY(Utf8View { source }.validate());
        auto message_buffer = MUST(Messages::WebContentClient::DidGetStyleSheetSource::static_encode(page_id, move(identifier), move(base_url), source));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_get_style_sheet_source(u64 page_id, Web::CSS::StyleSheetIdentifier const& identifier, URL::URL const& base_url, String const& source) {
        auto message_buffer = MUST(Messages::WebContentClient::DidGetStyleSheetSource::static_encode(page_id, move(identifier), move(base_url), source));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_take_screenshot(u64 page_id, Gfx::ShareableBitmap const& screenshot) {
        auto message_buffer = MUST(Messages::WebContentClient::DidTakeScreenshot::static_encode(page_id, move(screenshot)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_get_internal_page_info(u64 page_id, WebView::PageInfoType type, StringView info) {
        VERIFY(Utf8View { info }.validate());
        auto message_buffer = MUST(Messages::WebContentClient::DidGetInternalPageInfo::static_encode(page_id, type, info));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_get_internal_page_info(u64 page_id, WebView::PageInfoType type, String const& info) {
        auto message_buffer = MUST(Messages::WebContentClient::DidGetInternalPageInfo::static_encode(page_id, type, info));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_change_favicon(u64 page_id, Gfx::ShareableBitmap const& favicon) {
        auto message_buffer = MUST(Messages::WebContentClient::DidChangeFavicon::static_encode(page_id, move(favicon)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    Vector<Web::Cookie::Cookie> did_request_all_cookies(URL::URL url) {
        return m_connection.template send_sync<Messages::WebContentClient::DidRequestAllCookies>(move(url))->take_cookies();
    }

    void async_did_request_all_cookies(URL::URL const& url) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestAllCookies::static_encode(move(url)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    IPC::IPCErrorOr<Vector<Web::Cookie::Cookie>> try_did_request_all_cookies(URL::URL url) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentClient::DidRequestAllCookies>(move(url));
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    Optional<Web::Cookie::Cookie> did_request_named_cookie(URL::URL url, String name) {
        return m_connection.template send_sync<Messages::WebContentClient::DidRequestNamedCookie>(move(url), move(name))->take_cookie();
    }

    void async_did_request_named_cookie(URL::URL const& url, StringView name) {
        VERIFY(Utf8View { name }.validate());
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestNamedCookie::static_encode(move(url), name));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_named_cookie(URL::URL const& url, String const& name) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestNamedCookie::static_encode(move(url), name));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    IPC::IPCErrorOr<Optional<Web::Cookie::Cookie>> try_did_request_named_cookie(URL::URL url, String name) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentClient::DidRequestNamedCookie>(move(url), move(name));
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    String did_request_cookie(URL::URL url, Web::Cookie::Source source) {
        return m_connection.template send_sync<Messages::WebContentClient::DidRequestCookie>(move(url), source)->take_cookie();
    }

    void async_did_request_cookie(URL::URL const& url, Web::Cookie::Source source) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestCookie::static_encode(move(url), source));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    IPC::IPCErrorOr<String> try_did_request_cookie(URL::URL url, Web::Cookie::Source source) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentClient::DidRequestCookie>(move(url), source);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    void did_set_cookie(URL::URL url, Web::Cookie::ParsedCookie cookie, Web::Cookie::Source source) {
        (void) m_connection.template send_sync<Messages::WebContentClient::DidSetCookie>(move(url), move(cookie), source);
    }

    void async_did_set_cookie(URL::URL const& url, Web::Cookie::ParsedCookie const& cookie, Web::Cookie::Source source) {
        auto message_buffer = MUST(Messages::WebContentClient::DidSetCookie::static_encode(move(url), move(cookie), source));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    IPC::IPCErrorOr<void> try_did_set_cookie(URL::URL url, Web::Cookie::ParsedCookie cookie, Web::Cookie::Source source) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentClient::DidSetCookie>(move(url), move(cookie), source);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return { };

    }

    void async_did_update_cookie(Web::Cookie::Cookie const& cookie) {
        auto message_buffer = MUST(Messages::WebContentClient::DidUpdateCookie::static_encode(move(cookie)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_expire_cookies_with_time_offset(AK::Duration offset) {
        auto message_buffer = MUST(Messages::WebContentClient::DidExpireCookiesWithTimeOffset::static_encode(offset));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_update_resource_count(u64 page_id, i32 count_waiting) {
        auto message_buffer = MUST(Messages::WebContentClient::DidUpdateResourceCount::static_encode(page_id, count_waiting));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    String did_request_new_web_view(u64 page_id, Web::HTML::ActivateTab activate_tab, Web::HTML::WebViewHints hints, Optional<u64> page_index) {
        return m_connection.template send_sync<Messages::WebContentClient::DidRequestNewWebView>(page_id, move(activate_tab), move(hints), move(page_index))->take_handle();
    }

    void async_did_request_new_web_view(u64 page_id, Web::HTML::ActivateTab const& activate_tab, Web::HTML::WebViewHints const& hints, Optional<u64> const& page_index) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestNewWebView::static_encode(page_id, move(activate_tab), move(hints), move(page_index)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    IPC::IPCErrorOr<String> try_did_request_new_web_view(u64 page_id, Web::HTML::ActivateTab activate_tab, Web::HTML::WebViewHints hints, Optional<u64> page_index) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentClient::DidRequestNewWebView>(page_id, move(activate_tab), move(hints), move(page_index));
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    void async_did_request_activate_tab(u64 page_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestActivateTab::static_encode(page_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_close_browsing_context(u64 page_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidCloseBrowsingContext::static_encode(page_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_restore_window(u64 page_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestRestoreWindow::static_encode(page_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_reposition_window(u64 page_id, Gfx::IntPoint position) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestRepositionWindow::static_encode(page_id, position));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_resize_window(u64 page_id, Gfx::IntSize size) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestResizeWindow::static_encode(page_id, size));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_maximize_window(u64 page_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestMaximizeWindow::static_encode(page_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_minimize_window(u64 page_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestMinimizeWindow::static_encode(page_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_fullscreen_window(u64 page_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestFullscreenWindow::static_encode(page_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_file(u64 page_id, StringView path, i32 request_id) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestFile::static_encode(page_id, path, request_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_color_picker(u64 page_id, Color const& current_color) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestColorPicker::static_encode(page_id, move(current_color)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_file_picker(u64 page_id, Web::HTML::FileFilter const& accepted_file_types, Web::HTML::AllowMultipleFiles allow_multiple_files) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestFilePicker::static_encode(page_id, move(accepted_file_types), allow_multiple_files));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_request_select_dropdown(u64 page_id, Gfx::IntPoint content_position, i32 minimum_width, ReadonlySpan<Web::HTML::SelectItem> items) {
        auto message_buffer = MUST(Messages::WebContentClient::DidRequestSelectDropdown::static_encode(page_id, content_position, minimum_width, items));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_finish_handling_input_event(u64 page_id, Web::EventResult event_result) {
        auto message_buffer = MUST(Messages::WebContentClient::DidFinishHandlingInputEvent::static_encode(page_id, event_result));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_change_theme_color(u64 page_id, Gfx::Color color) {
        auto message_buffer = MUST(Messages::WebContentClient::DidChangeThemeColor::static_encode(page_id, color));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_insert_clipboard_entry(u64 page_id, StringView data, StringView presentation_style, StringView mime_type) {
        VERIFY(Utf8View { data }.validate());
        VERIFY(Utf8View { presentation_style }.validate());
        VERIFY(Utf8View { mime_type }.validate());
        auto message_buffer = MUST(Messages::WebContentClient::DidInsertClipboardEntry::static_encode(page_id, data, presentation_style, mime_type));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_insert_clipboard_entry(u64 page_id, String const& data, String const& presentation_style, String const& mime_type) {
        auto message_buffer = MUST(Messages::WebContentClient::DidInsertClipboardEntry::static_encode(page_id, data, presentation_style, mime_type));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_update_navigation_buttons_state(u64 page_id, bool back_enabled, bool forward_enabled) {
        auto message_buffer = MUST(Messages::WebContentClient::DidUpdateNavigationButtonsState::static_encode(page_id, back_enabled, forward_enabled));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_allocate_backing_stores(u64 page_id, i32 front_bitmap_id, Gfx::ShareableBitmap const& front_bitmap, i32 back_bitmap_id, Gfx::ShareableBitmap const& back_bitmap) {
        auto message_buffer = MUST(Messages::WebContentClient::DidAllocateBackingStores::static_encode(page_id, front_bitmap_id, move(front_bitmap), back_bitmap_id, move(back_bitmap)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_change_audio_play_state(u64 page_id, Web::HTML::AudioPlayState play_state) {
        auto message_buffer = MUST(Messages::WebContentClient::DidChangeAudioPlayState::static_encode(page_id, play_state));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_execute_js_console_input(u64 page_id, JsonValue const& result) {
        auto message_buffer = MUST(Messages::WebContentClient::DidExecuteJsConsoleInput::static_encode(page_id, move(result)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_output_js_console_message(u64 page_id, i32 message_index) {
        auto message_buffer = MUST(Messages::WebContentClient::DidOutputJsConsoleMessage::static_encode(page_id, message_index));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_get_js_console_messages(u64 page_id, i32 start_index, ReadonlySpan<WebView::ConsoleOutput> console_output) {
        auto message_buffer = MUST(Messages::WebContentClient::DidGetJsConsoleMessages::static_encode(page_id, start_index, console_output));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_finish_test(u64 page_id, StringView text) {
        VERIFY(Utf8View { text }.validate());
        auto message_buffer = MUST(Messages::WebContentClient::DidFinishTest::static_encode(page_id, text));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_finish_test(u64 page_id, String const& text) {
        auto message_buffer = MUST(Messages::WebContentClient::DidFinishTest::static_encode(page_id, text));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_set_test_timeout(u64 page_id, double milliseconds) {
        auto message_buffer = MUST(Messages::WebContentClient::DidSetTestTimeout::static_encode(page_id, milliseconds));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_set_browser_zoom(u64 page_id, double factor) {
        auto message_buffer = MUST(Messages::WebContentClient::DidSetBrowserZoom::static_encode(page_id, factor));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    void async_did_find_in_page(u64 page_id, size_t current_match_index, Optional<size_t> const& total_match_count) {
        auto message_buffer = MUST(Messages::WebContentClient::DidFindInPage::static_encode(page_id, current_match_index, move(total_match_count)));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    IPC::File request_worker_agent(u64 page_id) {
        return m_connection.template send_sync<Messages::WebContentClient::RequestWorkerAgent>(page_id)->take_socket();
    }

    void async_request_worker_agent(u64 page_id) {
        auto message_buffer = MUST(Messages::WebContentClient::RequestWorkerAgent::static_encode(page_id));
        MUST(m_connection.post_message(move(message_buffer))); 
    }

    IPC::IPCErrorOr<IPC::File> try_request_worker_agent(u64 page_id) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentClient::RequestWorkerAgent>(page_id);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

private:
    IPC::Connection<LocalEndpoint, PeerEndpoint>& m_connection;
};

template<typename LocalEndpoint, typename PeerEndpoint>
class WebContentClientProxy;
class WebContentClientStub;

class WebContentClientEndpoint {
public:
    template<typename LocalEndpoint>
    using Proxy = WebContentClientProxy<LocalEndpoint, WebContentClientEndpoint>;
    using Stub = WebContentClientStub;

    static u32 static_magic() { return 2565717604; }

    static ErrorOr<NonnullOwnPtr<IPC::Message>> decode_message(ReadonlyBytes buffer, [[maybe_unused]] Queue<IPC::File>& files)
    {
        FixedMemoryStream stream { buffer };
        auto message_endpoint_magic = TRY(stream.read_value<u32>());

        if (message_endpoint_magic != 2565717604) {
            return Error::from_string_literal("Endpoint magic number mismatch, not my message!");
        }

        auto message_id = TRY(stream.read_value<i32>());


        switch (message_id) {

        case (int)Messages::WebContentClient::MessageID::DidRequestNewProcessForNavigation:
            return TRY(Messages::WebContentClient::DidRequestNewProcessForNavigation::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidStartLoading:
            return TRY(Messages::WebContentClient::DidStartLoading::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidFinishLoading:
            return TRY(Messages::WebContentClient::DidFinishLoading::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestRefresh:
            return TRY(Messages::WebContentClient::DidRequestRefresh::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidPaint:
            return TRY(Messages::WebContentClient::DidPaint::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestCursorChange:
            return TRY(Messages::WebContentClient::DidRequestCursorChange::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidChangeTitle:
            return TRY(Messages::WebContentClient::DidChangeTitle::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidChangeUrl:
            return TRY(Messages::WebContentClient::DidChangeUrl::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestTooltipOverride:
            return TRY(Messages::WebContentClient::DidRequestTooltipOverride::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidStopTooltipOverride:
            return TRY(Messages::WebContentClient::DidStopTooltipOverride::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidEnterTooltipArea:
            return TRY(Messages::WebContentClient::DidEnterTooltipArea::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidLeaveTooltipArea:
            return TRY(Messages::WebContentClient::DidLeaveTooltipArea::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidHoverLink:
            return TRY(Messages::WebContentClient::DidHoverLink::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidUnhoverLink:
            return TRY(Messages::WebContentClient::DidUnhoverLink::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidClickLink:
            return TRY(Messages::WebContentClient::DidClickLink::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidMiddleClickLink:
            return TRY(Messages::WebContentClient::DidMiddleClickLink::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestContextMenu:
            return TRY(Messages::WebContentClient::DidRequestContextMenu::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestLinkContextMenu:
            return TRY(Messages::WebContentClient::DidRequestLinkContextMenu::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestImageContextMenu:
            return TRY(Messages::WebContentClient::DidRequestImageContextMenu::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestMediaContextMenu:
            return TRY(Messages::WebContentClient::DidRequestMediaContextMenu::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestAlert:
            return TRY(Messages::WebContentClient::DidRequestAlert::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestConfirm:
            return TRY(Messages::WebContentClient::DidRequestConfirm::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestPrompt:
            return TRY(Messages::WebContentClient::DidRequestPrompt::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestSetPromptText:
            return TRY(Messages::WebContentClient::DidRequestSetPromptText::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestAcceptDialog:
            return TRY(Messages::WebContentClient::DidRequestAcceptDialog::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestDismissDialog:
            return TRY(Messages::WebContentClient::DidRequestDismissDialog::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidGetSource:
            return TRY(Messages::WebContentClient::DidGetSource::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidInspectDomTree:
            return TRY(Messages::WebContentClient::DidInspectDomTree::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidInspectDomNode:
            return TRY(Messages::WebContentClient::DidInspectDomNode::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidInspectAccessibilityTree:
            return TRY(Messages::WebContentClient::DidInspectAccessibilityTree::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidGetHoveredNodeId:
            return TRY(Messages::WebContentClient::DidGetHoveredNodeId::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidFinishEditingDomNode:
            return TRY(Messages::WebContentClient::DidFinishEditingDomNode::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidMutateDom:
            return TRY(Messages::WebContentClient::DidMutateDom::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidGetDomNodeHtml:
            return TRY(Messages::WebContentClient::DidGetDomNodeHtml::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidListStyleSheets:
            return TRY(Messages::WebContentClient::DidListStyleSheets::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidGetStyleSheetSource:
            return TRY(Messages::WebContentClient::DidGetStyleSheetSource::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidTakeScreenshot:
            return TRY(Messages::WebContentClient::DidTakeScreenshot::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidGetInternalPageInfo:
            return TRY(Messages::WebContentClient::DidGetInternalPageInfo::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidChangeFavicon:
            return TRY(Messages::WebContentClient::DidChangeFavicon::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestAllCookies:
            return TRY(Messages::WebContentClient::DidRequestAllCookies::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestAllCookiesResponse:
            return TRY(Messages::WebContentClient::DidRequestAllCookiesResponse::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestNamedCookie:
            return TRY(Messages::WebContentClient::DidRequestNamedCookie::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestNamedCookieResponse:
            return TRY(Messages::WebContentClient::DidRequestNamedCookieResponse::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestCookie:
            return TRY(Messages::WebContentClient::DidRequestCookie::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestCookieResponse:
            return TRY(Messages::WebContentClient::DidRequestCookieResponse::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidSetCookie:
            return TRY(Messages::WebContentClient::DidSetCookie::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidSetCookieResponse:
            return TRY(Messages::WebContentClient::DidSetCookieResponse::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidUpdateCookie:
            return TRY(Messages::WebContentClient::DidUpdateCookie::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidExpireCookiesWithTimeOffset:
            return TRY(Messages::WebContentClient::DidExpireCookiesWithTimeOffset::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidUpdateResourceCount:
            return TRY(Messages::WebContentClient::DidUpdateResourceCount::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestNewWebView:
            return TRY(Messages::WebContentClient::DidRequestNewWebView::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestNewWebViewResponse:
            return TRY(Messages::WebContentClient::DidRequestNewWebViewResponse::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestActivateTab:
            return TRY(Messages::WebContentClient::DidRequestActivateTab::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidCloseBrowsingContext:
            return TRY(Messages::WebContentClient::DidCloseBrowsingContext::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestRestoreWindow:
            return TRY(Messages::WebContentClient::DidRequestRestoreWindow::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestRepositionWindow:
            return TRY(Messages::WebContentClient::DidRequestRepositionWindow::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestResizeWindow:
            return TRY(Messages::WebContentClient::DidRequestResizeWindow::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestMaximizeWindow:
            return TRY(Messages::WebContentClient::DidRequestMaximizeWindow::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestMinimizeWindow:
            return TRY(Messages::WebContentClient::DidRequestMinimizeWindow::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestFullscreenWindow:
            return TRY(Messages::WebContentClient::DidRequestFullscreenWindow::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestFile:
            return TRY(Messages::WebContentClient::DidRequestFile::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestColorPicker:
            return TRY(Messages::WebContentClient::DidRequestColorPicker::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestFilePicker:
            return TRY(Messages::WebContentClient::DidRequestFilePicker::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestSelectDropdown:
            return TRY(Messages::WebContentClient::DidRequestSelectDropdown::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidFinishHandlingInputEvent:
            return TRY(Messages::WebContentClient::DidFinishHandlingInputEvent::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidChangeThemeColor:
            return TRY(Messages::WebContentClient::DidChangeThemeColor::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidInsertClipboardEntry:
            return TRY(Messages::WebContentClient::DidInsertClipboardEntry::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidUpdateNavigationButtonsState:
            return TRY(Messages::WebContentClient::DidUpdateNavigationButtonsState::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidAllocateBackingStores:
            return TRY(Messages::WebContentClient::DidAllocateBackingStores::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidChangeAudioPlayState:
            return TRY(Messages::WebContentClient::DidChangeAudioPlayState::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidExecuteJsConsoleInput:
            return TRY(Messages::WebContentClient::DidExecuteJsConsoleInput::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidOutputJsConsoleMessage:
            return TRY(Messages::WebContentClient::DidOutputJsConsoleMessage::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidGetJsConsoleMessages:
            return TRY(Messages::WebContentClient::DidGetJsConsoleMessages::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidFinishTest:
            return TRY(Messages::WebContentClient::DidFinishTest::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidSetTestTimeout:
            return TRY(Messages::WebContentClient::DidSetTestTimeout::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidSetBrowserZoom:
            return TRY(Messages::WebContentClient::DidSetBrowserZoom::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidFindInPage:
            return TRY(Messages::WebContentClient::DidFindInPage::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::RequestWorkerAgent:
            return TRY(Messages::WebContentClient::RequestWorkerAgent::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::RequestWorkerAgentResponse:
            return TRY(Messages::WebContentClient::RequestWorkerAgentResponse::decode(stream, files));
        default:
            return Error::from_string_literal("Failed to decode WebContentClient message");
        }

        VERIFY_NOT_REACHED();
    }

};

class WebContentClientStub : public IPC::Stub {
public:
    WebContentClientStub() { }
    virtual ~WebContentClientStub() override { }

    virtual u32 magic() const override { return 2565717604; }
    virtual ByteString name() const override { return "WebContentClient"; }

    virtual ErrorOr<OwnPtr<IPC::MessageBuffer>> handle(NonnullOwnPtr<IPC::Message> message) override
    {
        switch (message->message_id()) {

        case (int)Messages::WebContentClient::MessageID::DidRequestNewProcessForNavigation: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestNewProcessForNavigation&>(*message);
            did_request_new_process_for_navigation(request.page_id(), request.take_url());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidStartLoading: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidStartLoading&>(*message);
            did_start_loading(request.page_id(), request.take_url(), request.is_redirect());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidFinishLoading: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidFinishLoading&>(*message);
            did_finish_loading(request.page_id(), request.take_url());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestRefresh: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestRefresh&>(*message);
            did_request_refresh(request.page_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidPaint: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidPaint&>(*message);
            did_paint(request.page_id(), request.take_content_rect(), request.bitmap_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestCursorChange: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestCursorChange&>(*message);
            did_request_cursor_change(request.page_id(), request.take_cursor());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidChangeTitle: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidChangeTitle&>(*message);
            did_change_title(request.page_id(), request.take_title());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidChangeUrl: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidChangeUrl&>(*message);
            did_change_url(request.page_id(), request.take_url());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestTooltipOverride: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestTooltipOverride&>(*message);
            did_request_tooltip_override(request.page_id(), request.position(), request.take_title());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidStopTooltipOverride: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidStopTooltipOverride&>(*message);
            did_stop_tooltip_override(request.page_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidEnterTooltipArea: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidEnterTooltipArea&>(*message);
            did_enter_tooltip_area(request.page_id(), request.take_title());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidLeaveTooltipArea: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidLeaveTooltipArea&>(*message);
            did_leave_tooltip_area(request.page_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidHoverLink: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidHoverLink&>(*message);
            did_hover_link(request.page_id(), request.take_url());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidUnhoverLink: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidUnhoverLink&>(*message);
            did_unhover_link(request.page_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidClickLink: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidClickLink&>(*message);
            did_click_link(request.page_id(), request.take_url(), request.take_target(), request.modifiers());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidMiddleClickLink: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidMiddleClickLink&>(*message);
            did_middle_click_link(request.page_id(), request.take_url(), request.take_target(), request.modifiers());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestContextMenu: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestContextMenu&>(*message);
            did_request_context_menu(request.page_id(), request.content_position());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestLinkContextMenu: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestLinkContextMenu&>(*message);
            did_request_link_context_menu(request.page_id(), request.content_position(), request.take_url(), request.take_target(), request.modifiers());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestImageContextMenu: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestImageContextMenu&>(*message);
            did_request_image_context_menu(request.page_id(), request.content_position(), request.take_url(), request.take_target(), request.modifiers(), request.take_bitmap());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestMediaContextMenu: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestMediaContextMenu&>(*message);
            did_request_media_context_menu(request.page_id(), request.content_position(), request.take_target(), request.modifiers(), request.take_menu());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestAlert: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestAlert&>(*message);
            did_request_alert(request.page_id(), request.take_message());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestConfirm: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestConfirm&>(*message);
            did_request_confirm(request.page_id(), request.take_message());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestPrompt: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestPrompt&>(*message);
            did_request_prompt(request.page_id(), request.take_message(), request.take_default_());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestSetPromptText: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestSetPromptText&>(*message);
            did_request_set_prompt_text(request.page_id(), request.take_message());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestAcceptDialog: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestAcceptDialog&>(*message);
            did_request_accept_dialog(request.page_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestDismissDialog: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestDismissDialog&>(*message);
            did_request_dismiss_dialog(request.page_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidGetSource: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidGetSource&>(*message);
            did_get_source(request.page_id(), request.take_url(), request.take_base_url(), request.take_source());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidInspectDomTree: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidInspectDomTree&>(*message);
            did_inspect_dom_tree(request.page_id(), request.take_dom_tree());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidInspectDomNode: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidInspectDomNode&>(*message);
            did_inspect_dom_node(request.page_id(), request.take_properties());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidInspectAccessibilityTree: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidInspectAccessibilityTree&>(*message);
            did_inspect_accessibility_tree(request.page_id(), request.take_accessibility_tree());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidGetHoveredNodeId: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidGetHoveredNodeId&>(*message);
            did_get_hovered_node_id(request.page_id(), request.take_node_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidFinishEditingDomNode: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidFinishEditingDomNode&>(*message);
            did_finish_editing_dom_node(request.page_id(), request.take_node_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidMutateDom: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidMutateDom&>(*message);
            did_mutate_dom(request.page_id(), request.take_mutation());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidGetDomNodeHtml: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidGetDomNodeHtml&>(*message);
            did_get_dom_node_html(request.page_id(), request.take_html());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidListStyleSheets: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidListStyleSheets&>(*message);
            did_list_style_sheets(request.page_id(), request.take_style_sheets());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidGetStyleSheetSource: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidGetStyleSheetSource&>(*message);
            did_get_style_sheet_source(request.page_id(), request.take_identifier(), request.take_base_url(), request.take_source());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidTakeScreenshot: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidTakeScreenshot&>(*message);
            did_take_screenshot(request.page_id(), request.take_screenshot());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidGetInternalPageInfo: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidGetInternalPageInfo&>(*message);
            did_get_internal_page_info(request.page_id(), request.type(), request.take_info());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidChangeFavicon: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidChangeFavicon&>(*message);
            did_change_favicon(request.page_id(), request.take_favicon());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestAllCookies: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestAllCookies&>(*message);
            auto response = did_request_all_cookies(request.take_url());
            return make<IPC::MessageBuffer>(TRY(response.encode()));
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestNamedCookie: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestNamedCookie&>(*message);
            auto response = did_request_named_cookie(request.take_url(), request.take_name());
            return make<IPC::MessageBuffer>(TRY(response.encode()));
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestCookie: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestCookie&>(*message);
            auto response = did_request_cookie(request.take_url(), request.source());
            return make<IPC::MessageBuffer>(TRY(response.encode()));
        }
        case (int)Messages::WebContentClient::MessageID::DidSetCookie: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidSetCookie&>(*message);
            did_set_cookie(request.take_url(), request.take_cookie(), request.source());
            auto response = Messages::WebContentClient::DidSetCookieResponse { };
            return make<IPC::MessageBuffer>(TRY(response.encode()));
        }
        case (int)Messages::WebContentClient::MessageID::DidUpdateCookie: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidUpdateCookie&>(*message);
            did_update_cookie(request.take_cookie());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidExpireCookiesWithTimeOffset: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidExpireCookiesWithTimeOffset&>(*message);
            did_expire_cookies_with_time_offset(request.offset());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidUpdateResourceCount: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidUpdateResourceCount&>(*message);
            did_update_resource_count(request.page_id(), request.count_waiting());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestNewWebView: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestNewWebView&>(*message);
            auto response = did_request_new_web_view(request.page_id(), request.take_activate_tab(), request.take_hints(), request.take_page_index());
            return make<IPC::MessageBuffer>(TRY(response.encode()));
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestActivateTab: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestActivateTab&>(*message);
            did_request_activate_tab(request.page_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidCloseBrowsingContext: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidCloseBrowsingContext&>(*message);
            did_close_browsing_context(request.page_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestRestoreWindow: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestRestoreWindow&>(*message);
            did_request_restore_window(request.page_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestRepositionWindow: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestRepositionWindow&>(*message);
            did_request_reposition_window(request.page_id(), request.position());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestResizeWindow: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestResizeWindow&>(*message);
            did_request_resize_window(request.page_id(), request.size());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestMaximizeWindow: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestMaximizeWindow&>(*message);
            did_request_maximize_window(request.page_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestMinimizeWindow: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestMinimizeWindow&>(*message);
            did_request_minimize_window(request.page_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestFullscreenWindow: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestFullscreenWindow&>(*message);
            did_request_fullscreen_window(request.page_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestFile: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestFile&>(*message);
            did_request_file(request.page_id(), request.take_path(), request.request_id());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestColorPicker: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestColorPicker&>(*message);
            did_request_color_picker(request.page_id(), request.take_current_color());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestFilePicker: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestFilePicker&>(*message);
            did_request_file_picker(request.page_id(), request.take_accepted_file_types(), request.allow_multiple_files());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidRequestSelectDropdown: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidRequestSelectDropdown&>(*message);
            did_request_select_dropdown(request.page_id(), request.content_position(), request.minimum_width(), request.take_items());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidFinishHandlingInputEvent: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidFinishHandlingInputEvent&>(*message);
            did_finish_handling_input_event(request.page_id(), request.event_result());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidChangeThemeColor: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidChangeThemeColor&>(*message);
            did_change_theme_color(request.page_id(), request.color());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidInsertClipboardEntry: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidInsertClipboardEntry&>(*message);
            did_insert_clipboard_entry(request.page_id(), request.take_data(), request.take_presentation_style(), request.take_mime_type());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidUpdateNavigationButtonsState: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidUpdateNavigationButtonsState&>(*message);
            did_update_navigation_buttons_state(request.page_id(), request.back_enabled(), request.forward_enabled());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidAllocateBackingStores: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidAllocateBackingStores&>(*message);
            did_allocate_backing_stores(request.page_id(), request.front_bitmap_id(), request.take_front_bitmap(), request.back_bitmap_id(), request.take_back_bitmap());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidChangeAudioPlayState: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidChangeAudioPlayState&>(*message);
            did_change_audio_play_state(request.page_id(), request.play_state());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidExecuteJsConsoleInput: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidExecuteJsConsoleInput&>(*message);
            did_execute_js_console_input(request.page_id(), request.take_result());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidOutputJsConsoleMessage: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidOutputJsConsoleMessage&>(*message);
            did_output_js_console_message(request.page_id(), request.message_index());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidGetJsConsoleMessages: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidGetJsConsoleMessages&>(*message);
            did_get_js_console_messages(request.page_id(), request.start_index(), request.take_console_output());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidFinishTest: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidFinishTest&>(*message);
            did_finish_test(request.page_id(), request.take_text());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidSetTestTimeout: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidSetTestTimeout&>(*message);
            did_set_test_timeout(request.page_id(), request.milliseconds());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidSetBrowserZoom: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidSetBrowserZoom&>(*message);
            did_set_browser_zoom(request.page_id(), request.factor());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::DidFindInPage: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::DidFindInPage&>(*message);
            did_find_in_page(request.page_id(), request.current_match_index(), request.take_total_match_count());
            return nullptr;
        }
        case (int)Messages::WebContentClient::MessageID::RequestWorkerAgent: {
            [[maybe_unused]] auto& request = static_cast<Messages::WebContentClient::RequestWorkerAgent&>(*message);
            auto response = request_worker_agent(request.page_id());
            return make<IPC::MessageBuffer>(TRY(response.encode()));
        }
        default:
            return Error::from_string_literal("Unknown message ID for WebContentClient endpoint");
        }
    }

    virtual void did_request_new_process_for_navigation(u64 page_id, URL::URL url) = 0;
    virtual void did_start_loading(u64 page_id, URL::URL url, bool is_redirect) = 0;
    virtual void did_finish_loading(u64 page_id, URL::URL url) = 0;
    virtual void did_request_refresh(u64 page_id) = 0;
    virtual void did_paint(u64 page_id, Gfx::IntRect content_rect, i32 bitmap_id) = 0;
    virtual void did_request_cursor_change(u64 page_id, Gfx::Cursor cursor) = 0;
    virtual void did_change_title(u64 page_id, ByteString title) = 0;
    virtual void did_change_url(u64 page_id, URL::URL url) = 0;
    virtual void did_request_tooltip_override(u64 page_id, Gfx::IntPoint position, ByteString title) = 0;
    virtual void did_stop_tooltip_override(u64 page_id) = 0;
    virtual void did_enter_tooltip_area(u64 page_id, ByteString title) = 0;
    virtual void did_leave_tooltip_area(u64 page_id) = 0;
    virtual void did_hover_link(u64 page_id, URL::URL url) = 0;
    virtual void did_unhover_link(u64 page_id) = 0;
    virtual void did_click_link(u64 page_id, URL::URL url, ByteString target, unsigned modifiers) = 0;
    virtual void did_middle_click_link(u64 page_id, URL::URL url, ByteString target, unsigned modifiers) = 0;
    virtual void did_request_context_menu(u64 page_id, Gfx::IntPoint content_position) = 0;
    virtual void did_request_link_context_menu(u64 page_id, Gfx::IntPoint content_position, URL::URL url, ByteString target, unsigned modifiers) = 0;
    virtual void did_request_image_context_menu(u64 page_id, Gfx::IntPoint content_position, URL::URL url, ByteString target, unsigned modifiers, Optional<Gfx::ShareableBitmap> bitmap) = 0;
    virtual void did_request_media_context_menu(u64 page_id, Gfx::IntPoint content_position, ByteString target, unsigned modifiers, Web::Page::MediaContextMenu menu) = 0;
    virtual void did_request_alert(u64 page_id, String message) = 0;
    virtual void did_request_confirm(u64 page_id, String message) = 0;
    virtual void did_request_prompt(u64 page_id, String message, String default_) = 0;
    virtual void did_request_set_prompt_text(u64 page_id, String message) = 0;
    virtual void did_request_accept_dialog(u64 page_id) = 0;
    virtual void did_request_dismiss_dialog(u64 page_id) = 0;
    virtual void did_get_source(u64 page_id, URL::URL url, URL::URL base_url, String source) = 0;
    virtual void did_inspect_dom_tree(u64 page_id, String dom_tree) = 0;
    virtual void did_inspect_dom_node(u64 page_id, WebView::DOMNodeProperties properties) = 0;
    virtual void did_inspect_accessibility_tree(u64 page_id, String accessibility_tree) = 0;
    virtual void did_get_hovered_node_id(u64 page_id, Web::UniqueNodeID node_id) = 0;
    virtual void did_finish_editing_dom_node(u64 page_id, Optional<Web::UniqueNodeID> node_id) = 0;
    virtual void did_mutate_dom(u64 page_id, WebView::Mutation mutation) = 0;
    virtual void did_get_dom_node_html(u64 page_id, String html) = 0;
    virtual void did_list_style_sheets(u64 page_id, Vector<Web::CSS::StyleSheetIdentifier> style_sheets) = 0;
    virtual void did_get_style_sheet_source(u64 page_id, Web::CSS::StyleSheetIdentifier identifier, URL::URL base_url, String source) = 0;
    virtual void did_take_screenshot(u64 page_id, Gfx::ShareableBitmap screenshot) = 0;
    virtual void did_get_internal_page_info(u64 page_id, WebView::PageInfoType type, String info) = 0;
    virtual void did_change_favicon(u64 page_id, Gfx::ShareableBitmap favicon) = 0;
    virtual Messages::WebContentClient::DidRequestAllCookiesResponse did_request_all_cookies(URL::URL url) = 0;
    virtual Messages::WebContentClient::DidRequestNamedCookieResponse did_request_named_cookie(URL::URL url, String name) = 0;
    virtual Messages::WebContentClient::DidRequestCookieResponse did_request_cookie(URL::URL url, Web::Cookie::Source source) = 0;
    virtual void did_set_cookie(URL::URL url, Web::Cookie::ParsedCookie cookie, Web::Cookie::Source source) = 0;
    virtual void did_update_cookie(Web::Cookie::Cookie cookie) = 0;
    virtual void did_expire_cookies_with_time_offset(AK::Duration offset) = 0;
    virtual void did_update_resource_count(u64 page_id, i32 count_waiting) = 0;
    virtual Messages::WebContentClient::DidRequestNewWebViewResponse did_request_new_web_view(u64 page_id, Web::HTML::ActivateTab activate_tab, Web::HTML::WebViewHints hints, Optional<u64> page_index) = 0;
    virtual void did_request_activate_tab(u64 page_id) = 0;
    virtual void did_close_browsing_context(u64 page_id) = 0;
    virtual void did_request_restore_window(u64 page_id) = 0;
    virtual void did_request_reposition_window(u64 page_id, Gfx::IntPoint position) = 0;
    virtual void did_request_resize_window(u64 page_id, Gfx::IntSize size) = 0;
    virtual void did_request_maximize_window(u64 page_id) = 0;
    virtual void did_request_minimize_window(u64 page_id) = 0;
    virtual void did_request_fullscreen_window(u64 page_id) = 0;
    virtual void did_request_file(u64 page_id, ByteString path, i32 request_id) = 0;
    virtual void did_request_color_picker(u64 page_id, Color current_color) = 0;
    virtual void did_request_file_picker(u64 page_id, Web::HTML::FileFilter accepted_file_types, Web::HTML::AllowMultipleFiles allow_multiple_files) = 0;
    virtual void did_request_select_dropdown(u64 page_id, Gfx::IntPoint content_position, i32 minimum_width, Vector<Web::HTML::SelectItem> items) = 0;
    virtual void did_finish_handling_input_event(u64 page_id, Web::EventResult event_result) = 0;
    virtual void did_change_theme_color(u64 page_id, Gfx::Color color) = 0;
    virtual void did_insert_clipboard_entry(u64 page_id, String data, String presentation_style, String mime_type) = 0;
    virtual void did_update_navigation_buttons_state(u64 page_id, bool back_enabled, bool forward_enabled) = 0;
    virtual void did_allocate_backing_stores(u64 page_id, i32 front_bitmap_id, Gfx::ShareableBitmap front_bitmap, i32 back_bitmap_id, Gfx::ShareableBitmap back_bitmap) = 0;
    virtual void did_change_audio_play_state(u64 page_id, Web::HTML::AudioPlayState play_state) = 0;
    virtual void did_execute_js_console_input(u64 page_id, JsonValue result) = 0;
    virtual void did_output_js_console_message(u64 page_id, i32 message_index) = 0;
    virtual void did_get_js_console_messages(u64 page_id, i32 start_index, Vector<WebView::ConsoleOutput> console_output) = 0;
    virtual void did_finish_test(u64 page_id, String text) = 0;
    virtual void did_set_test_timeout(u64 page_id, double milliseconds) = 0;
    virtual void did_set_browser_zoom(u64 page_id, double factor) = 0;
    virtual void did_find_in_page(u64 page_id, size_t current_match_index, Optional<size_t> total_match_count) = 0;
    virtual Messages::WebContentClient::RequestWorkerAgentResponse request_worker_agent(u64 page_id) = 0;
};

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic pop
#endif
